#ifndef bqtNewRadixMapHH
#define bqtNewRadixMapHH

#include "RadixTreeGMP-Vector.hh"
#include "NewRadixUtils.hh"

template<size_t RealSize, typename DataType, unsigned TokenBits>
class NewRadixMapGMP: private RadixTreeGMP<RealSize,sizeof(DataType),TokenBits>
{
private:
    typedef RadixTreeGMP<RealSize,sizeof(DataType),TokenBits> parent;
public:
    using parent::size;
    using parent::clear;
    inline bool find(BigInt& Key, DataType& value) const
    {
        return parent::find(Key, (unsigned char*)&value);
    }
    inline bool add(BigInt& Key, const DataType& value)
    {
        return parent::add(Key, (const unsigned char*)&value);
    }
    inline bool add_or_find(BigInt& Key, DataType& value)
    {
        return parent::add_or_find(Key, (unsigned char*)&value);
    }
    inline void ensure(BigInt& Key, const DataType& value)
    {
        parent::ensure(Key, (const unsigned char*)&value);
    }
};

/* Specializations for cases when the data fits in addressing space */
#if 0
#include <ext/pb_ds/assoc_container.hpp>
template<size_t RealSize, typename DataType>
class NewRadixMap_HashMap
{
private:
    struct KeyHolder
    {
        unsigned char KeySeq[RealSize];
        
        void Encode(BigInt& key_enc)
        {
            static BigInt rem_enc;
            size_t a=0;
            for(; a<RealSize; ++a)
            {
                unsigned remainder = key_enc.DivMod2Exp(key_enc, rem_enc, 8);
                KeySeq[RealSize-a-1] = remainder;
            }
            if(key_enc)
            {
                abort();
            }
        }
        
        bool operator==(const KeyHolder& b) const
        {
            return std::memcmp(KeySeq, b.KeySeq, RealSize) == 0;
        }
    } __attribute__((packed));
    
    struct Hash
    {
        size_t operator() (const KeyHolder& h) const
        {
            unsigned char res[sizeof(size_t)] = { 0 };
            const size_t copysize = std::min(sizeof(res), RealSize);
            const size_t excess   = RealSize - copysize;
            std::memcpy(res, h.KeySeq + excess, copysize);
            return *(const size_t*)res;
        }
    };
    

    typedef pb_ds::cc_hash_table<KeyHolder, DataType, Hash> data_t;
    data_t data;
public:
    size_t size() const { return data.size(); }
    void clear() { data.clear(); }

    inline bool find(BigInt& Key, DataType& value) const
    {
        KeyHolder key; key.Encode(Key);
        typename data_t::const_point_iterator i = data.find(key);
        if(i == data.end()) return false;
        value = i->second;
        return true;
    }
    inline bool add(BigInt& Key, const DataType& value)
    {
        KeyHolder key; key.Encode(Key);
        
        std::pair<typename data_t::point_iterator, bool>
            res = data.insert(std::pair<KeyHolder,DataType> (key, value) );
        
        return res.second;
    }
    inline bool add_or_find(BigInt& Key, DataType& value)
    {
        KeyHolder key; key.Encode(Key);
        
        std::pair<typename data_t::point_iterator, bool>
            res = data.insert(std::pair<KeyHolder,DataType> (key, value) );
        
        if(!res.second)
            value = res.first->second;
        
        return res.second;
    }
    inline void ensure(BigInt& Key, const DataType& value)
    {
        KeyHolder key; key.Encode(Key);
        
        std::pair<typename data_t::point_iterator, bool>
            res = data.insert(std::pair<KeyHolder,DataType> (key, value) );
        
        if(!res.second)
        {
            res.first->second = value;
        }
    }
};

template<typename DataType>
class NewRadixMapGMP<1,DataType,unsigned char,8>: public NewRadixMap_HashMap<1,DataType> { };
template<typename DataType>
class NewRadixMapGMP<2,DataType,unsigned char,8>: public NewRadixMap_HashMap<2,DataType> { };
template<typename DataType>
class NewRadixMapGMP<3,DataType,unsigned char,8>: public NewRadixMap_HashMap<3,DataType> { };
template<typename DataType>
class NewRadixMapGMP<4,DataType,unsigned char,8>: public NewRadixMap_HashMap<4,DataType> { };
#ifdef __x86_64
template<typename DataType>
class NewRadixMapGMP<5,DataType,unsigned char,8>: public NewRadixMap_HashMap<5,DataType> { };
template<typename DataType>
class NewRadixMapGMP<6,DataType,unsigned char,8>: public NewRadixMap_HashMap<6,DataType> { };
template<typename DataType>
class NewRadixMapGMP<7,DataType,unsigned char,8>: public NewRadixMap_HashMap<7,DataType> { };
template<typename DataType>
class NewRadixMapGMP<8,DataType,unsigned char,8>: public NewRadixMap_HashMap<8,DataType> { };
#endif
#endif

enum { RadixMapTokenBits = CompactCoordinateBitsPerUnit };

/* Note: BytesUsedBy won't work if MaxValue-MinValue == (size_t)~0ULL */
template<typename KeyType, size_t MaxValue, size_t KeyLength,
         typename DataType, size_t MinValue=0>
class NewRadixMap:
    private NewRadixMapGMP<
        UnitsUsedBy<MaxValue-MinValue+1,KeyLength, RadixMapTokenBits>::MaxBytes,
        DataType, RadixMapTokenBits>
{
private:
    typedef NewRadixMapGMP<
        UnitsUsedBy<MaxValue-MinValue+1,KeyLength, RadixMapTokenBits>::MaxBytes,
        DataType, RadixMapTokenBits> parent;

    static inline BigInt& Encode(const KeyType*const Key)
    {
        return NewRadixEncoder::Encode<KeyType,MaxValue,MinValue> (Key, KeyLength);
    }
public:
    using parent::size;
    using parent::clear;
    inline bool find(const KeyType*const Key, DataType& value) const
    {
        return parent::find(Encode(Key), value);
    }
    inline bool add(const KeyType*const Key, const DataType& value)
    {
        return parent::add(Encode(Key), value);
    }
    inline bool add_or_find(const KeyType*const Key, DataType& value)
    {
        return parent::add_or_find(Encode(Key), value);
    }
    inline void ensure(const KeyType*const Key, const DataType& value)
    {
        parent::ensure(Encode(Key), value);
    }
};

#endif
