#include "bigint.hh"

/* This hack avoids a linker error due to buggy pb_ds/exceptions.hpp */
#define __throw_container_error inline __throw_container_error
#define __throw_insert_error inline __throw_insert_error
#define __throw_join_error inline __throw_join_error
#define __throw_resize_error inline __throw_resize_error

#include <ext/pb_ds/assoc_container.hpp>

#undef __throw_container_error
#undef __throw_insert_error
#undef __throw_join_error
#undef __throw_resize_error


template<size_t RealSize>
struct RadixTreeGMPbase
{
    struct KeyType
    {
    public:
        typedef unsigned char value_type;
        
        static const value_type MinValue = 0x00;
        static const value_type MaxValue = 0xFF;
    
        void GenSeq(BigInt& key_enc)
        {
            static BigInt rem_enc;
            size_t a=0;
            for(; a<RealSize; ++a)
            {
                unsigned remainder = key_enc.DivMod2Exp(key_enc, rem_enc, 8);
                KeySeq[RealSize-a-1] = remainder;
            }
            if(key_enc)
            {
                abort();
            }
        }
        
        /* The value_type, and these four are required
         * by pb_ds::string_trie_e_access_traits
         */
        typedef const unsigned char* const_reverse_iterator;
        typedef const unsigned char* const_iterator;
        
        const_iterator end()   const { return KeySeq+RealSize; }
        const_iterator begin() const { return KeySeq; }
        
    private:
        unsigned char KeySeq[RealSize];
    } __attribute__((packed));
};

template<size_t RealSize, size_t PayloadSize>
struct RadixTreeGMP: public RadixTreeGMPbase<RealSize>
{
private:
    typedef typename RadixTreeGMPbase<RealSize>::KeyType KeyType;
    
    struct PayloadType
    {
        unsigned char data[PayloadSize];
    } __attribute__((packed));

    typedef pb_ds::trie<
        KeyType,
        PayloadType,
        pb_ds::string_trie_e_access_traits<KeyType, KeyType::MinValue, KeyType::MaxValue>
    > ContainerType;
    
    ContainerType data;

protected:
    inline size_t size() const { return data.size(); }
    void clear() { data.clear(); }
    
    RadixTreeGMP(): data() { }
    
    bool find(BigInt& Key, unsigned char* Data = 0) const
    {
        KeyType k; k.GenSeq(Key);

        typename ContainerType::const_iterator i = data.find(k);
        if(i == data.end()) return false;
        
        std::memcpy(Data, i->second.data, PayloadSize);
        return true;
    }
    
    void ensure(BigInt& Key, const unsigned char* Data)
    {
        KeyType k; k.GenSeq(Key);
        
        PayloadType& d = data[k];
        std::memcpy(d.data, Data, PayloadSize);
    }

    bool add(BigInt& Key, const unsigned char* Data = 0)
    {
        std::pair<KeyType, PayloadType> ins_data;
        ins_data.first.GenSeq(Key);
        std::memcpy(ins_data.second.data, Data, PayloadSize);
        
        std::pair<typename ContainerType::iterator, bool>
            ins_res = data.insert(ins_data);
        
        return (ins_res.second);
    }

    bool add_or_find(BigInt& Key, unsigned char* Data)
    {
        std::pair<KeyType, PayloadType> ins_data;
        ins_data.first.GenSeq(Key);
        std::memcpy(ins_data.second.data, Data, PayloadSize);
        
        std::pair<typename ContainerType::iterator, bool>
            ins_res = data.insert(ins_data);
        
        if(ins_res.second)
        {
            std::memcpy(Data, ins_res.first->second.data, PayloadSize);
            return true;
        }
        
        return false;
    }
};

template<size_t RealSize>
struct RadixTreeGMP<RealSize, 0>: public RadixTreeGMPbase<RealSize>
{
private:
    typedef typename RadixTreeGMPbase<RealSize>::KeyType KeyType;
    
    typedef pb_ds::trie<
        KeyType,
        pb_ds::null_mapped_type,
        pb_ds::string_trie_e_access_traits<KeyType, KeyType::MinValue, KeyType::MaxValue>
    > ContainerType;
    
    ContainerType data;

protected:
    inline size_t size() const { return data.size(); }
    void clear() { data.clear(); }
    
    RadixTreeGMP(): data() { }
    
    bool find(BigInt& Key) const
    {
        KeyType k; k.GenSeq(Key);

        typename ContainerType::const_iterator i = data.find(k);
        if(i == data.end()) return false;
        
        return true;
    }
    
    bool add(BigInt& Key)
    {
        KeyType k; k.GenSeq(Key);
        
        std::pair<typename ContainerType::iterator, bool>
            ins_res = data.insert(k);
        
        return (ins_res.second);
    }
};

#undef bytecode
