#include "vectorstorage.hh"

#undef bytecode
#define bytecode VectorStorage::Buf

static inline unsigned CalcLinkBytes(unsigned long LinkValue)
{
    if(LinkValue & 0xFFFF0000U)
        return (LinkValue & 0xFF000000U) ? 4 : 3;
    return (LinkValue & 0xFF00) ? 2 : 1;
}

template<unsigned RealSize>
struct RadixTreeNode
{
private:
    typedef RadixTreeNode<RealSize> self;
private:
    VectorStorage::IDtype OmaID;
public:
    explicit RadixTreeNode(VectorStorage::IDtype id) : OmaID(id)
    {
        VectorStorage::Read(OmaID);
        //fprintf(stderr, "Node %u is %p\n", id, this);
    }
public:
    void clear(unsigned offset=0)
    {
        for(unsigned a=0; a < bytecode.size() && offset<RealSize; )
        {
            unsigned byte = bytecode[a++];
            if(byte > 0x0F)
            {
                ++offset;
                continue;
            }
            
            TranslateLinkAsObject(ReadLink1(a, byte & 3)).clear(offset);
            VectorStorage::Free1();
            VectorStorage::Read(OmaID);
            
            if(byte & 8)
            {
                TranslateLinkAsObject(ReadLink2(a, byte & 3)).clear(offset);
                VectorStorage::Free1();
                VectorStorage::Read(OmaID);
            }
            
            // bytecode[a]; - ignore this raw byte.
            ++a;
            ++offset;
        }
        bytecode.clear();
    }
    /* MakeLink1.
     *   In input, the buffer contains:
     *    D0 D1 D2 D3...
     *   in other words, plain data.
     *
     *   In output, the buffer contains:
     *
     *    LK link D0 D1 D2 D3...
     *
     *   Where LK is the link code (00..0F) and link
     *   is a 1-4 bytes long link.
     *
     *   The subtree will be empty.
     *
     */
    VectorStorage::IDtype MakeLink1(unsigned& ptr, unsigned char type)
    {
        VectorStorage::IDtype NewID = VectorStorage::Create();
        VectorStorage::Read(OmaID);
        
        if(NewID <= OmaID) abort();
        
        unsigned long LinkValue = NewID - OmaID - 1;
        unsigned LinkBytes = CalcLinkBytes(LinkValue);
        
        unsigned char Buf[64];
        Buf[0] = type | (LinkBytes-1);
        for(unsigned a=0; a<LinkBytes; ++a)
            Buf[a+1] = (LinkValue >> (a*8)) & 0xFF;
        
        bytecode.replace(ptr, 0, Buf, LinkBytes+1);
        
        ptr += LinkBytes+1;
        return NewID;
    }

    /* MakeLink2.
     *   In input, the buffer contains:
     *    LK link D0 D1 D2 D3...
     *   in other words, something that was created by MakeLink1.
     *
     *   In output, the buffer contains:
     *
     *    LK link lonk D0 D1 D2 D3...
     *
     *   Where lonk is the second link.
     *   The first link may have changed in length,
     *   and LK has been updated too.
     *
     *   The subtree will be empty.
     *
     */
    VectorStorage::IDtype MakeLink2(unsigned ptr)
    {
        VectorStorage::IDtype NewID = VectorStorage::Create();
        VectorStorage::Read(OmaID);
        
        if(NewID <= OmaID) abort();
        
        unsigned long LinkValue2 = NewID - OmaID - 1;
        unsigned LinkBytes = CalcLinkBytes(LinkValue2);
        
        unsigned char ctrl = bytecode[ptr];
        
        unsigned OldLinkBytes = (ctrl & 3)+1;
        unsigned long OldLink = 0;
        for(unsigned a=0; a<OldLinkBytes; ++a)
            OldLink |= bytecode[ptr+1+a] << (a*8);
        
        if(OldLinkBytes > LinkBytes) LinkBytes = OldLinkBytes;
        
        unsigned char Buf[64];
        
        Buf[0] = (ctrl&0x04) | 0x08 | (LinkBytes-1);
        
        for(unsigned a=0; a<LinkBytes; ++a)
        {
            Buf[a+1]           = (OldLink    >> (a*8)) & 0xFF;
            Buf[a+LinkBytes+1] = (LinkValue2 >> (a*8)) & 0xFF;
        }
        
        bytecode.replace(ptr, OldLinkBytes+1,
                          Buf, LinkBytes*2+1);

        return NewID;
    }

    /* Reads a link created by MakeLink1.
     */
    unsigned long ReadLink1(unsigned& ptr, unsigned bytes_minus_1) const
    {
        unsigned long result = 0;
        for(unsigned a=0; a<=bytes_minus_1; ++a)
            result |= bytecode[ptr++] << (a*8);
        /*
        fprintf(stderr, "<%u wants to read result=%lu(%u bytes), ID=%lu>",
            OmaID, result, bytes_minus_1+1, OmaID+1+result);
        */
        return result;
    }
    
    /* Reads a link created by MakeLink2.
     *
     * Currently, it's read the same way as Link1.
     */
    unsigned long ReadLink2(unsigned& ptr, unsigned bytes_minus_1) const
    {
        return ReadLink1(ptr, bytes_minus_1);
    }
    
    /* Translates the link into a node id. */
    VectorStorage::IDtype TranslateLink(unsigned long result) const
    {
        return OmaID + 1 + result;
    }
    
    /* Translates the link into a node pointer. */
    self TranslateLinkAsObject(unsigned long result) const
    {
        VectorStorage::IDtype NewID = TranslateLink(result);
        return self(NewID);
    }
};

template<unsigned RealSize, unsigned PayloadSize=0>
struct RadixTreeGMP
{
private:
    typedef RadixTreeNode<RealSize> NodeType;
protected:
    inline unsigned size() const { return Size; }
    void clear()
    {
        DeleteRoot();
        Size=0;
        RootID = VectorStorage::Create();
    }
    
    RadixTreeGMP(): Size(0), RootID(VectorStorage::Create())
    {
    }
    ~RadixTreeGMP()
    {
        DeleteRoot();
    }

    bool find(BigInt& Key, unsigned char* Data = 0) const
    {
        Encode(Key);
        
        VectorStorage::IDtype id = RootID;
        unsigned ptr             = 0;
        if(!Finder(id, ptr)) return false;
        
        if(Data)
        {
            //VectorStorage::Read(id);
            std::memcpy(Data, bytecode.GetReadPtr(ptr,PayloadSize), PayloadSize); 
        }
        return true;
    }
    bool ensure(BigInt& Key, const unsigned char* Data)
    {
        Encode(Key);
        
        DumpKey("Ensuring ", Data);
        
        VectorStorage::IDtype id = RootID;
        unsigned ptr             = 0;
        if(!Adder(id, ptr, Data))
        {
            /* Already existed? */
            /* Then id,ptr should now be pointing to the
             * beginning of the data.
             */
            //VectorStorage::Read(id);
            bytecode.replace(ptr,PayloadSize, Data,PayloadSize);
        }
        return true;
    }

    bool add(BigInt& Key, const unsigned char* Data = 0)
    {
        Encode(Key);

        DumpKey("Adding ", Data);
        
        VectorStorage::IDtype id = RootID;
        unsigned ptr             = 0;
        return Adder(id, ptr, Data);
    }

private:
    RadixTreeGMP(const RadixTreeGMP&);
    void operator=(const RadixTreeGMP&);
    
private:
    unsigned Size;
    VectorStorage::IDtype RootID;
    
    inline void DeleteRoot()
    {
        NodeType(RootID).clear();
        VectorStorage::Free1();
    }

private:
    template<typename StopHandler>
    bool Matcher(VectorStorage::IDtype& id, unsigned& ptr, StopHandler handler) const
    {
        NodeType Cur(id);
        
        unsigned end = bytecode.size();
        
        for(unsigned a=0; a<RealSize; ++a)
        {
            const unsigned char want = KeySeq[a];
        Loop:
            if(ptr >= end)
            {
                /* Add back the rest here */
                handler(id, ptr, a, -1);
                return false;
            }
            
            unsigned linkptr = ptr;
            unsigned char byte = bytecode[ptr];
            if(byte == want)
            {
#if 0
                fprintf(stderr, "** want=%02X, got=%02X, continuing...\n",
                    want,byte);
#endif
                ++ptr;
                continue;
            }
            if(byte > 0x0F)
            {
#if 0
                fprintf(stderr, "** want=%02X, got=%02X, making a branch.\n",
                    want,byte);
#endif
                /* Make a Link1 at this point, replacing the "byte" */
                handler(id, ptr, a, want>byte);
                return false;
            }
            ++ptr;
            
            unsigned char ctrl = byte & 0x0F;
            /*
             * Ctrl values >> 2 are:
             *
             *  0: Continues in forward, link = smaller, bigger = NIL
             *  1: Continues in forward, link = bigger,  smaller = NIL
             *  2: Continues in forward, link1 = smaller, link2 = bigger
             *  3: Continues in forward, link1 = bigger,  link2 = smaller
             *
             * The lower 3 bits are the link length.
             */
            unsigned long link1val = Cur.ReadLink1(ptr, ctrl & 3);
            unsigned long link2val = 0;
            if(ctrl & 8) link2val = Cur.ReadLink2(ptr, ctrl & 3);
            
            byte = bytecode[ptr++];
            
#if 0
            fprintf(stderr, "** want=%02X, ctrl=%02X, compare=%02X\n",
                want,ctrl,byte);
#endif

            switch(ctrl >> 2)
            {
                case 0:
                    if(want == byte) continue;
                    if(want > byte)
                    {
#if 0
                        fprintf(stderr, "** -> making a second tree\n");
#endif
                        /* Make a Link2 at this point */
                        handler(id, linkptr, a, 2);
                        return false;
                    }
#if 0
                    fprintf(stderr, "** -> follow link1\n");
#endif
                    id = Cur.TranslateLink(link1val);
                    goto GotNewID;
                case 1:
                    if(want == byte) continue;
                    if(want < byte)
                    {
#if 0
                        fprintf(stderr, "** -> making a second tree\n");
#endif
                        /* Make a Link2 at this point */
                        handler(id, linkptr, a, 2);
                        return false;
                    }
#if 0
                    fprintf(stderr, "** -> follow link1\n");
#endif
                    id = Cur.TranslateLink(link1val);
                    goto GotNewID;
                case 2:
                    if(want == byte) continue;
#if 0
                    fprintf(stderr, "** -> follow %s\n", want<byte?"link1":"link2");
#endif
                    id = Cur.TranslateLink(want > byte ? link2val : link1val);
                    goto GotNewID;
                case 3:
                    if(want == byte) continue;
#if 0
                    fprintf(stderr, "** -> follow %s\n", want<byte?"link1":"link2");
#endif
                    id = Cur.TranslateLink(want < byte ? link2val : link1val);
                GotNewID:
                    ptr = 0;
                    Cur = NodeType(id);
                    end = bytecode.size();
                    goto Loop;
            }
        }
        return true;
    }
    
    struct DoNothing
    {
        inline void operator() (VectorStorage::IDtype, unsigned, unsigned, int)
        {
        }
    };
    
    inline bool Finder(VectorStorage::IDtype& id, unsigned& ptr) const
    {
        return Matcher(id, ptr, DoNothing());
    }
    
    struct DoAdding
    {
        explicit DoAdding(const unsigned char* d) : Data(d) { }

        /* Modes:
         *   -1  = add back
         *   0,1 = make Link1 (smaller, bigger)
         *   2   = make Link2
         */
        void operator() (VectorStorage::IDtype id, unsigned ptr, unsigned keypos, int mode)
        {
            switch(mode)
            {
                case 2:
                    return (*this)(NodeType(id).MakeLink2(ptr), 0, keypos, -1);
                case 0:
                    return (*this)(NodeType(id).MakeLink1(ptr, 0), 0, keypos, -1);
                case 1:
                    return (*this)(NodeType(id).MakeLink1(ptr, 4), 0, keypos, -1);
                case -1:
                    break;
            }
            
            /* Add the tail */
            unsigned char* KeySeq = RadixTreeGMP<RealSize,PayloadSize>::KeySeq;
            
            const unsigned KeySize = RealSize-keypos;
            const unsigned PaySize = PayloadSize;
            const unsigned CopySize = KeySize + PaySize;
            unsigned char Buf[CopySize];
            
            std::copy(KeySeq+keypos, KeySeq+RealSize, Buf);
            if(Data) std::copy(Data, Data+PayloadSize, Buf+KeySize);
            
            VectorStorage::Read(id);
            bytecode.replace(ptr,0, Buf, CopySize);
        }
    private:
        const unsigned char* Data;
    };
    friend struct DoAdding;
    
    inline bool Adder(VectorStorage::IDtype& id, unsigned& ptr, const unsigned char* Data = 0)
    {
        bool added = !Matcher(id, ptr, DoAdding(Data));
        if(added) { ++Size; Dump("After adding something"); }
        return added;
    }

private:
    static unsigned char KeySeq[RealSize];
    
    static void Encode(BigInt& key_enc)
    {
        static BigInt rem_enc;
        for(unsigned a=0; a<RealSize; ++a)
        {
            unsigned remainder = key_enc.DivMod(key_enc, rem_enc, 240);
            KeySeq[RealSize-a-1] = 0xFF-remainder;
        }
        if(key_enc)
        {
            abort();
        }
    }
    static void DumpKey(const char* msg, const unsigned char* Data)
    {
    return;
        fprintf(stderr, "%s ", msg);
        for(unsigned a=0; a<RealSize; ++a) fprintf(stderr, "%02X ", KeySeq[a]);
        if(Data)
        {
            for(unsigned a=0; a<PayloadSize; ++a) fprintf(stderr, "|%02X", Data[a]);
        }
        fprintf(stderr, "\n");
        std::fflush(stderr);
    }
    void Dump(const char* message)
    {
    return;
        fprintf(stderr, "%s:\n", message);
        Dump(RootID, 2, 0);
    }
    void Dump(VectorStorage::IDtype id, unsigned indent, unsigned byteoffs)
    {
        NodeType Cur(id);
        
        bool newline = true;
        
        unsigned ptr = 0;
        while(ptr < bytecode.size() && byteoffs < RealSize)
        {
            unsigned char byte = bytecode[ptr++];
            if(byte > 0x0F)
            {
                if(newline) { fprintf(stderr, "%*s", indent, ""); newline=false; }
                fprintf(stderr, "%02X ", byte);
                ++byteoffs;
            }
            else
            {
                unsigned long link1val = Cur.ReadLink1(ptr, byte & 3);
                unsigned long link2val = 0;
                if(byte & 8) link2val = Cur.ReadLink2(ptr, byte & 3);
                
                unsigned char nextbyte = bytecode[ptr++];
                //if(!newline) { fprintf(stderr, "\n"); newline=true; }
                if(newline) { fprintf(stderr, "%*s", indent, ""); newline=false; }
                fprintf(stderr, "[%X]%02X\n", byte, nextbyte);

                fprintf(stderr, "%*s%s @ L1: %lu(%lu):\n", indent+1,"",
                    (byte&4)?">":"<",
                    link1val, Cur.TranslateLink(link1val));

                /* "left tree" is the "less branch", so byteoffs += 0 */
                Dump(Cur.TranslateLink(link1val), indent+2, byteoffs);
                VectorStorage::Read(id);
                
                if(byte & 8)
                {
                    /* "right tree" is the "greater branch", so byteoffs += 0 */
                    fprintf(stderr, "%*s%s @ L2: %lu(%lu):\n", indent+1,"",
                        (byte&4)?"<":">",
                        link2val, Cur.TranslateLink(link2val));
                    Dump(Cur.TranslateLink(link2val), indent+2, byteoffs);
                    VectorStorage::Read(id);
                }

                fprintf(stderr, "%*s%s @ cont'd:\n", indent+1,"",
                    "=");
                newline=true;
                /* "continuation" is the "equal branch", so byteoffs += 1 */
                ++byteoffs;
            }
        }
        /* The rest of the bytes are raw */
        unsigned remsiz=0;
        for(; ptr < bytecode.size(); ++remsiz)
        {
            unsigned char byte = bytecode[ptr++];
            if(newline) { fprintf(stderr, "%*s", indent, ""); newline=false; }
            fprintf(stderr, "|%02X", byte);
        }
        if(remsiz != PayloadSize)
        {
            fprintf(stderr, " -- Mismatch in payload size!");
        }
        fprintf(stderr, "\n");
        std::fflush(stderr);
    }
};
template<unsigned RealSize, unsigned PayloadSize>
unsigned char RadixTreeGMP<RealSize,PayloadSize>::KeySeq[];

#undef bytecode
