#include "vectorstorage.hh"
#include "bigint.hh"

#undef bytecode
#define bytecode VectorStorage::Buf

static inline unsigned CalcLinkBytes(unsigned long LinkValue)
{
    if(LinkValue & 0xFFFF0000U)
        return (LinkValue & 0xFF000000U) ? 4 : 3;
    return (LinkValue & 0xFF00) ? 2 : 1;
}

struct RadixTreeNodeBase
{
    unsigned char byte;
} __attribute__((packed));

static bool RadixTreeNodeCompare(const RadixTreeNodeBase& a, unsigned char b)
{
    return a.byte < b;
}

template<unsigned PointerSize = 4>
struct RadixTreeNode: public RadixTreeNodeBase
{
    unsigned delta : (PointerSize*8);
    
    template<unsigned OtherSize>
    inline void operator= (const RadixTreeNode<OtherSize>& b)
    {
        byte  = b.byte;
        delta = b.delta;
    }
} __attribute__((packed));

template<unsigned PayloadSize>
struct RadixTreeLeaf: public RadixTreeNodeBase
{
    unsigned char data[PayloadSize];
} __attribute__((packed));

template<>
struct RadixTreeLeaf<0>: public RadixTreeNodeBase
{
    struct { } __attribute__((packed)) data;
} __attribute__((packed));

template<unsigned RealSize, unsigned PayloadSize=0>
struct RadixTreeGMP
{
private:
protected:
    inline unsigned size() const { return Size; }
    void clear()
    {
        DeleteRoot();
        Size=0;
        RootID = VectorStorage::Create();
    }
    
    RadixTreeGMP(): Size(0), RootID(VectorStorage::Create())
    {
    }
    ~RadixTreeGMP()
    {
        DeleteRoot();
    }
    
    bool find(BigInt& Key, unsigned char* Data = 0) const
    {
        Encode(Key);
        
        //DumpKey("Finding ", Data);
        
        RadixTreeLeaf<PayloadSize>* p = Process(IdleAddFunc());
        if(p && PayloadSize) std::memcpy(Data, &p->data, PayloadSize);
        
        return p;
    }
    
    void ensure(BigInt& Key, const unsigned char* Data)
    {
        Encode(Key);
        
        RadixTreeLeaf<PayloadSize>* p = Process(ActualAddFunc(Data,*this));
        if(p) std::memcpy(&p->data, Data, PayloadSize);
    }

    bool add(BigInt& Key, const unsigned char* Data = 0)
    {
        Encode(Key);
        
        RadixTreeLeaf<PayloadSize>* p = Process(ActualAddFunc(Data,*this));
        return !p;
    }

    bool add_or_find(BigInt& Key, unsigned char* Data)
    {
        Encode(Key);
        
        RadixTreeLeaf<PayloadSize>* p = Process(ActualAddFunc(Data,*this));
        if(p) std::memcpy(Data, &p->data, PayloadSize);
        return !p;
    }

private:
    RadixTreeGMP(const RadixTreeGMP&);
    void operator=(const RadixTreeGMP&);
    
    struct IdleAddFunc
    {
        inline void operator() (VectorStorage::IDtype,unsigned bytepos,unsigned level) const
        {
        }
    };
    
    struct ActualAddFunc
    {
        ActualAddFunc(const unsigned char*d, RadixTreeGMP& o) : Data(d), obj(o) { } 
        
        inline void operator() (VectorStorage::IDtype Cur,unsigned bytepos,unsigned level) const
        {
            obj.AddBack(Cur,bytepos,level,Data);
        }
    private:
        const unsigned char* Data;
        RadixTreeGMP& obj;
    };
    friend struct ActualAddFunc;
    
    template<typename NodeType, typename AddBackFunc>
    static inline NodeType* ProcessHelper
        (VectorStorage::IDtype Cur, unsigned char want,unsigned level,
         unsigned offset, const AddBackFunc& AddBack)
    {
        VectorStorage::Read(Cur);
        
        NodeType* begin = (NodeType*)bytecode.GetReadPtr(offset, bytecode.size()-offset);
        NodeType* end   = begin + ((bytecode.size()-offset) / sizeof(NodeType));
        NodeType* p = std::lower_bound(begin,end, want, RadixTreeNodeCompare);
        if(p == end || p->byte != want)
        {
            unsigned bytepos = (p-begin)*sizeof(NodeType);
            AddBack(Cur,bytepos, level);
            return 0;
        }
        return p;
    }
    
    template<typename AddBackFunc>
    RadixTreeLeaf<PayloadSize>* Process(const AddBackFunc& AddBack) const
    {
        unsigned level = 0;

        VectorStorage::IDtype Cur = RootID;
        for(; level+1 < RealSize; ++level)
        {
            VectorStorage::Read(Cur);
            
            unsigned char want = KeySeq[level];
            
            if(bytecode.empty())
            {
                AddBack(Cur,0, level);
                return 0;
            }
            switch(bytecode[0])
            {
                case 1:
                {
                    typedef RadixTreeNode<1> node;
                    node* p = ProcessHelper<node>(Cur,want,level, 1, AddBack);
                    if(!p) return 0;
                    Cur = Cur + 1 + p->delta;
                    break;
                }
                case 2:
                {
                    typedef RadixTreeNode<2> node;
                    /* Note: The "offset" must be "1" in each call. */
                    node* p = ProcessHelper<node>(Cur,want,level, 1, AddBack);
                    if(!p) return 0;
                    Cur = Cur + 1 + p->delta;
                    break;
                }
                case 3:
                {
                    typedef RadixTreeNode<3> node;
                    node* p = ProcessHelper<node>(Cur,want,level, 1, AddBack);
                    if(!p) return 0;
                    Cur = Cur + 1 + p->delta;
                    break;
                }
                case 4:
                {
                    typedef RadixTreeNode<4> node;
                    node* p = ProcessHelper<node>(Cur,want,level, 1, AddBack);
                    if(!p) return 0;
                    Cur = Cur + 1 + p->delta;
                    break;
                }
            }
        }
        if(RealSize == 0) return 0;
        
        unsigned char want = KeySeq[level];
        
        typedef RadixTreeLeaf<PayloadSize> node;
        return ProcessHelper<node> (Cur,want,level, 0, AddBack);
    }
    
private:
    unsigned Size;
    VectorStorage::IDtype RootID;
    
    inline void DeleteRoot()
    {
        DeleteArray(RootID, 0);
    }
    inline void DeleteArray(VectorStorage::IDtype id, unsigned level)
    {
        if(level+1 < RealSize)
        {
            VectorStorage::Read(id);
            
            #define Deleter(nodetype) \
                unsigned b = (bytecode.size()-1) / sizeof(nodetype); \
                for(unsigned a=0; a<b; ++a) \
                { \
                    const nodetype* ptr = (const nodetype*)bytecode.GetReadPtr(1+a*sizeof(nodetype),0); \
                    DeleteArray(id + 1 + ptr->delta, level+1); \
                    if(a+1 < b) VectorStorage::Read(id); \
                }
            
            if(!bytecode.empty())
            {
                switch(bytecode[0])
                {
                    default:
                    case 1: { Deleter(RadixTreeNode<1>); break; }
                    case 2: { Deleter(RadixTreeNode<2>); break; }
                    case 3: { Deleter(RadixTreeNode<3>); break; }
                    case 4: { Deleter(RadixTreeNode<4>); break; }
                }
            }
        }
        VectorStorage::Free1();
        /* leafs don't need special handling. */
    }

    void AddBack(VectorStorage::IDtype id, unsigned bytepos, unsigned level,
                 const unsigned char* Data)
    {
        if(level+1 >= RealSize)
        {
            RadixTreeLeaf<PayloadSize> node;
            node.byte = KeySeq[level];
            if(PayloadSize) std::memcpy(&node.data, Data, PayloadSize);
            bytecode.replace(bytepos, 0, (const unsigned char*)&node, sizeof(node));
            
            ++Size;
            DumpKey("Added ", Data);
            Dump("After adding something");
            return;
        }
        /* The intermediate nodes */
        VectorStorage::IDtype NewID = VectorStorage::Create();
        unsigned Delta = NewID-id-1;
        unsigned char nbytes = CalcLinkBytes(Delta);
        
        if(bytecode.empty())
        {
            /* Easy! We get to choose our own data size */
            bytecode.replace(0, 0, &nbytes, 1);
        }
        else
        {
            unsigned char old_nbytes = bytecode[0];
            if(old_nbytes > nbytes)
            {
                /* Pick the existing size */
                nbytes = old_nbytes;
            }
            else if(old_nbytes < nbytes)
            {
                unsigned old_size = 1+old_nbytes;
                unsigned new_size = 1+nbytes;

                /* Need to resize the existing data! */
                unsigned Size = (bytecode.size()-1) / old_size;
                unsigned char NewBuf[1 + (Size+1)*new_size];
                
                /* Scale the bytepos after resizing. */
                unsigned ElemPos = bytepos / old_size;
                bytepos = ElemPos * new_size;
                
                NewBuf[0] = nbytes;
                
                /* While resizing the existing data, we also poke the
                 * new entry into the structure. Saves some work.
                 */
                
                const unsigned char* SrcData = bytecode.GetReadPtr(1, bytecode.size()-1);
                
                /*
                if(nbytes >= 3)
                std::cerr << "RadixTreeGMP-Vector Resize from "
                          << (unsigned)old_nbytes << " to " << (unsigned)nbytes << "\n";
                */
                
                #define Copy(srcType, destType) \
                    unsigned a=0; \
                    for(; a<ElemPos; ++a) \
                    { \
                        ((destType*)(NewBuf+1))[a] = ((const srcType*)(SrcData))[a]; \
                    } \
                    destType& target = ((destType*)(NewBuf+1))[a]; \
                    target.byte  = KeySeq[level]; \
                    target.delta = Delta; \
                    for(; a<Size; ++a) \
                    { \
                        ((destType*)(NewBuf+1))[a+1] = ((const srcType*)(SrcData))[a]; \
                    } \
                
                switch(old_nbytes)
                {
                    default:
                    case 1:
                    {
                        typedef RadixTreeNode<1> srcType;
                        switch(nbytes)
                        {
                            default:
                            case 2: { Copy(srcType, RadixTreeNode<2>); break; }
                            case 3: { Copy(srcType, RadixTreeNode<3>); break; }
                            case 4: { Copy(srcType, RadixTreeNode<4>); break; }
                        }
                        break;
                    }
                    case 2:
                    {
                        typedef RadixTreeNode<2> srcType;
                        switch(nbytes)
                        {
                            default:
                            case 3: { Copy(srcType, RadixTreeNode<3>); break; }
                            case 4: { Copy(srcType, RadixTreeNode<4>); break; }
                        }
                        break;
                    }
                    case 3:
                    {
                        typedef RadixTreeNode<3> srcType;
                        Copy(srcType, RadixTreeNode<4>);
                        /* - there's no other choice
                        switch(nbytes)
                        {
                            case 4: { Copy(srcType, RadixTreeNode<4>); break; }
                        }
                        */
                        break;
                    }
                    case 4:
                        /* It wouldn't be growing if it was "4" */
                        abort();
                        break;
                }
                #undef Copy
                
                bytecode.replace(0,bytecode.size(), NewBuf, sizeof(NewBuf));
                goto DidAdd;
            }
        }
        
        #define DoAdd() \
            node.byte  = KeySeq[level]; \
            node.delta = Delta; \
            bytecode.replace(bytepos+1, 0, (const unsigned char*)&node, sizeof(node));
        switch(nbytes)
        {
            case 1: { RadixTreeNode<1> node; DoAdd(); break; }
            case 2: { RadixTreeNode<2> node; DoAdd(); break; }
            case 3: { RadixTreeNode<3> node; DoAdd(); break; }
            case 4: { RadixTreeNode<4> node; DoAdd(); break; }
        }
        #undef DoAdd
    DidAdd:
        VectorStorage::Read(NewID);
        AddBack(NewID, 0, level+1, Data);
    }

private:
    static unsigned char KeySeq[RealSize];
    
    static void Encode(BigInt& key_enc)
    {
        static BigInt rem_enc;
        for(unsigned a=0; a<RealSize; ++a)
        {
            unsigned remainder = key_enc.DivMod2Exp(key_enc, rem_enc, 8);
            KeySeq[RealSize-a-1] = remainder;
        }
        if(key_enc)
        {
            abort();
        }
    }
    static void DumpKey(const char* msg, const unsigned char* Data)
    {
#if 0
        fprintf(stderr, "%s ", msg);
        for(unsigned a=0; a<RealSize; ++a) fprintf(stderr, "%02X ", KeySeq[a]);
        if(Data)
        {
            for(unsigned a=0; a<PayloadSize; ++a) fprintf(stderr, "|%02X", Data[a]);
        }
        fprintf(stderr, "\n");
        std::fflush(stderr);
#endif
    }
    void Dump(const char* message)
    {
#if 0
        fprintf(stderr, "%s:\n", message);
        Dump(RootID, 2, 0);
#endif
    }
    void Dump(VectorStorage::IDtype id, unsigned indent, unsigned level)
    {
#if 0
        if(level+1 < RealSize)
        {
            VectorStorage::Read(id);
            
            if(bytecode.empty())
            {
                fprintf(stderr, "%*sEmpty node\n", indent,"");
            }
            else
            {
                fprintf(stderr, "%*s[%02X]\n", indent,"", bytecode[0]);
                #define Loop(nodetype) \
                    unsigned b = (bytecode.size()-1) / sizeof(nodetype); \
                    for(unsigned a=0; a<b; ++a) \
                    { \
                        nodetype* ptr = (nodetype*)bytecode.GetReadPtr(1+a*sizeof(nodetype),0); \
                        fprintf(stderr, "%*s%02X\n", indent,"", ptr->byte); \
                        Dump(id + 1 + ptr->delta, indent+1, level+1); \
                        VectorStorage::Read(id); \
                    }
                    
                switch(bytecode[0])
                {
                    default:
                    case 1: { Loop(RadixTreeNode<1>); break; }
                    case 2: { Loop(RadixTreeNode<2>); break; }
                    case 3: { Loop(RadixTreeNode<3>); break; }
                    case 4: { Loop(RadixTreeNode<4>); break; }
                }
                #undef Loop
            }
        }
        else
        {
            VectorStorage::Read(id);
            
            typedef RadixTreeLeaf<PayloadSize> node;
            unsigned b = bytecode.size() / sizeof(node);
            for(unsigned a=0; a<b; ++a)
            {
                node* ptr = (node*)bytecode.GetReadPtr(a*sizeof(node),0);
                fprintf(stderr, "%*s%02X ", indent,"", ptr->byte);
                for(unsigned c=0; c<PayloadSize; ++c)
                    fprintf(stderr, "|%02X", ((unsigned char*)&ptr->data)[c]);
                fprintf(stderr, "\n");
            }
        }
#endif
    }
};
template<unsigned RealSize, unsigned PayloadSize>
unsigned char RadixTreeGMP<RealSize,PayloadSize>::KeySeq[];

#undef bytecode
