#ifndef bqtFileWindowHH
#define bqtFileWindowHH

#ifndef _LARGEFILE64_SOURCE
 #define _LARGEFILE64_SOURCE
 #include <unistd.h>
 #include <fcntl.h>
 #include <sys/stat.h>
#endif
#include <sys/mman.h>

#include <iostream>
#include <cstdlib>

template<unsigned WindowSize>
struct FileWindow
{
    int fd;
    
    off64_t Begin;
    unsigned char* Data;
    bool mapped;
    
    FileWindow(): fd(-1), mapped(false)
    {
    }
    
    void Init(int f)
    {
        if(mapped) { munmap(Data, WindowSize); mapped=false; }
        
        fd = f;
        mapped=false;
    }
    ~FileWindow()
    {
        if(mapped) munmap(Data, WindowSize);
    }
    
    void Detach()
    {
        if(mapped) { munmap(Data, WindowSize); mapped=false; }
    }
    
    bool Overlap(off64_t bBegin) const
    {
        if(!mapped) return false;
        
        off64_t end = Begin   + WindowSize;
        off64_t bend = bBegin + WindowSize;
        return (Begin < bend && bBegin < end);
    }
    
    void Load(off64_t ptr)
    {
        if(ptr % getpagesize())
        {
            std::abort();
        }
        
        Begin=ptr;
        if(mapped) munmap(Data, WindowSize);
        
        //std::cerr << "Load(" << ptr << ")\n";
        
        void* p = mmap64(0, WindowSize, PROT_READ|PROT_WRITE, MAP_SHARED, fd, Begin);
        
        if(p == (void*)-1 || p == (void*)0)
        {
            std::cerr << "mmap64 " << Begin << "\n";
            std::perror("mmap64");
            std::abort();
        }
        else
            Data = (unsigned char*)p;

        mapped=true;
    }
    
    static int CreateTmpFile()
    {
        static unsigned counter=0;
        char Buf[4096];
        sprintf(Buf, "/tmp/rope-%u-%u", WindowSize, counter++);
        int fd = open(Buf, O_CREAT | O_RDWR | O_LARGEFILE | O_TRUNC, 0666);
        if(fd < 0)
        {
            perror(Buf);
        }
        unlink(Buf);
        
        return fd;
    }

    FileWindow(const FileWindow& b) : fd(b.fd), Begin(0), mapped(false)
    {
    }
    FileWindow& operator=(const FileWindow& b)
    {
        Detach();
        fd = b.fd;
        mapped = false;
        return *this;
    }
private:
    bool Overlap(unsigned bBegin) const;
    void Load(unsigned ptr);
};

template<unsigned WindowSize,
         unsigned NumWindows,
         typename DataType=char,
         unsigned MaxSize=0>
class TempFileBuffer
{
public:
    typedef FileWindow<WindowSize> Window;

    static const unsigned ElemCount = MaxSize / sizeof(DataType);
    static const off64_t NBytes = ElemCount * (off64_t)sizeof(DataType);

    TempFileBuffer(): fd(Window::CreateTmpFile()), FileLimit(0)
    {
        ReinitWindows();
    }
    ~TempFileBuffer()
    {
        Detach();
        close(fd);
    }
    
    inline off64_t GetCap() const { return FileLimit; }
    void clear()
    {
        Detach();
        ftruncate64(fd, FileLimit=0);
    }
    
    DataType* GetElemWindow(const off64_t ElemNo, unsigned amount=1) const
    {
        return (DataType*)GetWindow(ElemNo * sizeof(DataType),
                                    amount * sizeof(DataType));
    }
    
    unsigned char* GetWindow(const off64_t begin, unsigned size) const
    {
        const off64_t end = begin + size;
        
        if(end > FileLimit)
        {
            if(NBytes > 0 && FileLimit < NBytes) ftruncate64(fd, FileLimit = NBytes);
            if(end > FileLimit)
            {
                off64_t align    = 0x10000;
                off64_t newlimit = (end + align-1) & ~(align-1);
                FileLimit = newlimit;
                ftruncate64(fd, FileLimit);
            }
        }

        /* Find if there's a window that works */
        /* Try most recently used windows first. */
        for(unsigned b=0; b<NumWindows; ++b)
        {
            const unsigned windownum = age[b];
            Window& w = Windows[windownum];
            if(!w.mapped) continue;
            
            off64_t wbegin = w.Begin;
            off64_t wend   = w.Begin + (off64_t)WindowSize;
            
            if(begin >= wbegin && end <= wend)
            {
                MoveFirst(windownum, b);
                return &w.Data[begin - wbegin];
            }
        }
        
        /* Pick a window randomly */
        unsigned windownum = OldestWindow();
        Window& w = Windows[windownum];
        
        /* Find a new base for it */
        off64_t new_begin = begin;
        new_begin -= (WindowSize - size) / 2;
        if(new_begin < 0) new_begin=0;
        
        new_begin &= ~0xFFFULL;
        if(new_begin + WindowSize < end)
        {
            new_begin=begin;
        }
        
        new_begin &= ~0xFFFULL;
        if(new_begin + WindowSize < end)
        {
            new_begin=begin;
        }

        if(new_begin + WindowSize < end)
        {
            std::abort();
        }

/*
        // According to Warp, this doesn't seem to be needed.
        
        for(unsigned b=0; b<NumWindows; ++b)
            if(b != a && Windows[b].Overlap(new_begin))
                Windows[b].Detach();
*/
        w.Load(new_begin);
        return &w.Data[begin - w.Begin];
    }
    
    void Detach()
    {
        /* Free the memory maps, but keep the file intact */
        for(unsigned n=0; n<NumWindows; ++n) Windows[n].Detach();
    }
private:
    /* Copying is not allowed. If you need to include this structure
     * within a STL container, you must use boost::shared_ptr.
     */
    TempFileBuffer(const TempFileBuffer& b);
    TempFileBuffer& operator=(const TempFileBuffer& b);
private:
    int fd;
    mutable off64_t FileLimit;
    mutable Window Windows[NumWindows];
    mutable unsigned char age[NumWindows];
    
    inline void ReinitWindows()
    {
        for(unsigned n=0; n<NumWindows; ++n)
        {
            Windows[n].Init(fd);
            age[n] = n;
        }
    }
    inline void Touch(unsigned windownum) const
    {
        for(unsigned a=0; a<NumWindows; ++a)
        {
            if(age[a] == windownum)
            {
                MoveFirst(windownum, a);
                return;
            }
        }
    }
    inline unsigned OldestWindow() const
    {
        unsigned a = NumWindows-1;
        unsigned windownum = age[a];
        MoveFirst(windownum, a);
        return windownum;
    }
    inline void MoveFirst(unsigned windownum, unsigned oldpos) const
    {
        // move this window to the beginning of the list
        for(unsigned a=oldpos; a > 0; --a) age[a] = age[a-1];
        age[0] = windownum;
    }
};

#endif
