#include "history.hh"
#include "mapdata.hh"
#include "timings.hh"
#include "levelstate.hh"

#include <sstream>
#include <list>

void History::Dump(std::ostream& out) const
{
    out << "// Solution:\n";
    unsigned size = Moves.size();
    for(unsigned a=0; a<size; ++a)
    {
        const Move& move = Moves[a];
        out << "// - ";
        move.Dump();
        out << "\n";
    }
    out << std::flush;
}

const std::string History::GetJs() const
{
    std::stringstream result;
    
    bool Grab = false;
    bool First = true;
    
    unsigned pushes = 0;
    unsigned steps = 0;
    unsigned kicks = 0;
    
    unsigned timer = 0;
    int prevdx=0, prevdy=-1;
    
    unsigned px = maps[levelno].px - OffsetWidth;
    unsigned py = maps[levelno].py - OffsetHeight;
    
    LevelState state;
    state.CreateFromLevel();
    state.RebuildBlockCaches();

    unsigned goal_delay = 0;
    
    /* Put the moves into a std::list so that they
     * can be handled from first to last quickly.
     */
    std::list<Move> moves;
    for(MoveTree::reverse_const_iterator
        i = Moves.rbegin(); i != Moves.rend(); ++i)
    {
        moves.push_front(*i);
    }
    
    for(std::list<Move>::const_iterator
        i = moves.begin(); i != moves.end(); ++i)
    {
        const unsigned movepx = i->getpx();
        const unsigned movepy = i->getpy();
        
        result << "\n/* " << timer << ": ";
        i->Dump(result); result << "*/";
        
        /* Create the required events to walk to the next location */
        if(movepx != px || movepy != py)
        {
            WalkingRoute route;
            state.FindWalkingRouteTo(movepx, movepy, route);

            Grab=false;
            
            for(unsigned a=0; a<route.size(); ++a)
            {
                unsigned newpx = route[a] % MaxWidth;
                unsigned newpy = route[a] / MaxWidth;
                
                unsigned length = 1;
                prevdx = newpx-px;
                prevdy = newpy-py;
                while(a+length < route.size())
                {
                    unsigned nextpx = route[a+length] % MaxWidth;
                    unsigned nextpy = route[a+length] / MaxWidth;
                    int nextdx = nextpx-newpx;
                    int nextdy = nextpy-newpy;
                    
                    if(nextdx != prevdx || nextdy != prevdy) break;

                    newpx = nextpx;
                    newpy = nextpy;
                    ++length;
                }
                a += length-1;
                
                px=newpx;
                py=newpy;
                
                if(First)First=false;else result << ",\n";
                int param = (py+OffsetHeight)*19+(px+OffsetWidth);
                result << "{t:" << timer
                       << ",y:'w'"
                          ",p:" << param
                       << "}";
                unsigned dura = length * IntWalkingSpeed;
                //result << "/*length " << length << ", duration " << dura << "*/";
                
                timer += dura;
                
                goal_delay = 0;
                steps += length;
            }
            
            // normalize the direction vector
            if(prevdx<0)prevdx/=-prevdx; else if(prevdx>0)prevdx/=prevdx;
            if(prevdy<0)prevdy/=-prevdy; else if(prevdy>0)prevdy/=prevdy;
            
            state.WalkTo(movepx, movepy);
        }
        
        /* Face to the direction if necessary */
        const int movedx = i->getdx();
        const int movedy = i->getdy();

        if(movedx != prevdx || movedy != prevdy)
        {
            if(First)First=false;else result << ",\n";
            int param = (movedy * 19 + movedx);
            result << "{t:" << timer
                   << ",y:'f'"
                      ",p:" << param
                   << "}";
            
            prevdx = movedx;
            prevdy = movedy;
            //timer += IntFacingTime;
        }


        if(i->type == Move::WalkToPush)
        {
            /* Grab if necessary */
            if(!Grab)
            {
                if(First)First=false;else result << ",\n";
                result << "{t:" << timer
                       << ",y:'g'"
                       << "}";
                Grab=true;
                if(timer > 0) timer += IntGrabbingTime;
            }
            
            /* Create the pushing event */
            if(First)First=false;else result << ",\n";
            result << "{t:" << timer
                   << ",y:'p'"
                   << "}";
            timer += IntPushingSpeed;
            
            px += movedx;
            py += movedy;
            goal_delay = 0;
            
            ++steps;
            ++pushes;
            state.Push(movedx,movedy);
        }
        else /*if(movetype == Move::WalkToKick)*/
        {
            Grab = false;
            
            /* Create the kick preparation event */
            if(First)First=false;else result << ",\n";
            result << "{t:" << timer
                   << ",y:'K'"
                   << "}";
            if(timer > 0) timer += IntKickingTime;
            result << "/*timer " << timer << " after " << IntKickingTime << " */";

            /* Create the kick event */
            if(First)First=false;else result << ",\n";
            result << "{t:" << timer
                   << ",y:'k'"
                   << "}";
            
            unsigned length = state.CalculateKick(movedx,movedy);
            state.Kick(movedx,movedy, length);
            //pushes += length;
            kicks += 1;
            
            goal_delay = RoundUp(length*KickingSpeed);
        }
    }
    
    timer += goal_delay;
    
    std::string movie = "[\n" + result.str() + "\n]";
    
    std::stringstream data;

    data << "{stats:{steps:" << steps
               << ",pushes:" << pushes
               << ",kicks:" << kicks
               << ",undoes:0,aborts:99999"
               << ",timer:" << timer
               << ",owner:'BisqBot',exists:1}"
         << ",movie:" << movie << "}";
    
    return "perfect_data=" + data.str() + ";\n";
}
