#ifndef bqtIntArrayHH
#define bqtIntArrayHH

//#define INTARRAY_RANGECHECKING
#include "bigint.hh"
#include "templates.hh"

#include <iterator>

template<unsigned OuterLimit, unsigned MaxCount>
class GMPCompressedIntArrayRep
{
private:
    enum { MaxBits  = BytesUsedBy<OuterLimit,MaxCount>::MaxBits };
    enum { MaxBytes = BytesUsedBy<OuterLimit,MaxCount>::MaxBytes };
    
    typedef typename OptimalStorageType<MaxBytes>::Type ValueType;

    enum { ValueBits = sizeof(ValueType)*8 };
    enum { MaxUnits =  (MaxBits+ValueBits-1)/ValueBits };
    
public:
    typedef ValueType RepType;
    enum { RepSize = MaxUnits, RepMaxValue = RepType(~0) };
    RepType Rep[RepSize];
private:
    static inline BigInt& Exponent(unsigned index)
    {
        static BigInt pow;
        pow.Pow(OuterLimit, index);
        return pow;
    }
    BigInt& Make() const
    {
        static BigInt val;
        val = 0;
        for(unsigned a=0; a<MaxUnits; ++a)
        {
            if(a > 0) val <<= ValueBits;
            val += Rep[a];
        }
        return val;
    }
    void Unmake(BigInt& val)
    {
        BigInt rem;
        for(unsigned b=MaxUnits, a=0; a<MaxUnits; ++a)
        {
            Rep[--b] = val.DivMod2Exp(val, rem, ValueBits);
        }
    }
public:
    unsigned Get_nth(unsigned index) const
    {
#ifdef INTARRAY_RANGECHECKING
        if(index >= MaxCount) throw std::bad_alloc();
#endif
        BigInt& val = Make();
        if(index > 0)
        {
            /* val = val / (base^index) */
            val /= Exponent(index);
        }
        /* result = val % base */
        return val % OuterLimit;
    }
    
    void Set_nth(unsigned index, unsigned newvalue)
    {
#ifdef INTARRAY_RANGECHECKING
        if(index >= MaxCount) throw std::bad_alloc();
        if(newvalue >= OuterLimit)
        {
            abort();
        }
#endif
        BigInt& val = Make();
        
        /* div = base^index
         * rem = val % div
         * val = val / div
         */

        BigInt rem;
        BigInt& div = Exponent(index);
        val.DivMod(val, rem, div);
        
        unsigned oldvalue = val % OuterLimit;
        
        /* val = val-oldvalue+newvalue */
        if(newvalue > oldvalue)
            val += newvalue-oldvalue;
        else if(newvalue < oldvalue)
            val -= oldvalue-newvalue;
        
        /* val = (val*div)+rem */
        val *= div;
        val += rem;
        
        Unmake(val);
    }
    
    template<typename IntType>
    void Set_all(IntType data[MaxCount])
    {
        static BigInt val;
        val = 0;
        for(unsigned a=0; a<MaxCount; ++a)
        {
            if(a>0) val *= OuterLimit;
            val += data[MaxCount-a-1];
        }
        Unmake(val);
    }
    
    template<typename IntType>
    void Get_all(IntType data[MaxCount]) const
    {
        static BigInt rem;
        
        BigInt& val = Make();

        for(unsigned a=0; a<MaxCount; ++a)
        {
            data[a] = val.DivMod(val,rem, OuterLimit);
        }
    }
}     __attribute__((packed));

template<unsigned OuterLimit, unsigned MaxCount>
class CompressedIntArray
{
public:
    typedef GMPCompressedIntArrayRep<OuterLimit, MaxCount> RepType;
    RepType Rep;
private:
    template<typename BasePtr>
    struct storage_proxy
    {
        storage_proxy(BasePtr r, unsigned p) : ref(r),pos(p) { }
        storage_proxy(const storage_proxy& b) : ref(b.ref),pos(b.pos) { }
        storage_proxy() { }
    public:
        storage_proxy& operator=(const storage_proxy& b)
        {
            ref->Rep.Set_nth(pos,b);
            return *this; 
        }
        storage_proxy& operator= (unsigned v) { ref->Rep.Set_nth(pos,v); return *this; }
        storage_proxy& operator+= (int v) { ref->Rep.Set_nth(pos,ref->Rep.Get_nth(pos)+v); return *this; }
        storage_proxy& operator-= (int v) { ref->Rep.Set_nth(pos,ref->Rep.Get_nth(pos)-v); return *this; }
        operator unsigned () const { return ref->Rep.Get_nth(pos); }
        
        BasePtr ref;
        unsigned pos;
    };
    
    template<typename Ptr,typename Ref,typename BasePtr>
    struct iterator_base
    {
        typedef Ptr pointer;
        typedef Ref reference;
        typedef int difference_type;
        typedef unsigned value_type;
        typedef std::random_access_iterator_tag iterator_category;
    public:
        iterator_base() { }
        iterator_base(BasePtr r, unsigned p) : proxy(r, p) { }
        iterator_base(const iterator_base& b) : proxy(b.proxy) { }
        iterator_base& operator=(const iterator_base& b)
            { proxy.ref=b.proxy.ref; proxy.pos=b.proxy.pos; return *this; }
        
        inline bool operator== (const iterator_base& b) const { return proxy.pos==b.proxy.pos; }
        inline bool operator!= (const iterator_base& b) const { return proxy.pos!=b.proxy.pos; }
        inline bool operator<  (const iterator_base& b) const { return proxy.pos< b.proxy.pos; }
        inline bool operator>  (const iterator_base& b) const { return proxy.pos> b.proxy.pos; }
        inline bool operator<= (const iterator_base& b) const { return proxy.pos<=b.proxy.pos; }
        inline bool operator>= (const iterator_base& b) const { return proxy.pos>=b.proxy.pos; }
        
        iterator_base& operator++ () { ++proxy.pos; return *this; }
        iterator_base& operator-- () { --proxy.pos; return *this; }
        iterator_base operator-- (int) { iterator_base tmp(proxy.ref, proxy.pos); --proxy.pos; return tmp; }
        iterator_base operator++ (int) { iterator_base tmp(proxy.ref, proxy.pos); ++proxy.pos; return tmp; }
        iterator_base operator+ (unsigned o) const { return iterator_base(proxy.ref, proxy.pos+o); }
        iterator_base operator- (unsigned o) const { return iterator_base(proxy.ref, proxy.pos-o); }
        iterator_base& operator+= (unsigned o) { proxy.pos+=o; return *this; }
        iterator_base& operator-= (unsigned o) { proxy.pos-=o; return *this; }
        difference_type operator- (const iterator_base& b) const
            { return proxy.pos-b.proxy.pos; }
        
        const value_type operator* () const { return proxy; }
        storage_proxy<BasePtr>& operator* () { return proxy; }
    private:
        storage_proxy<BasePtr> proxy;
    };
public:
    typedef iterator_base<const void*,unsigned,                     const CompressedIntArray*> const_iterator;
    typedef iterator_base<      void*,storage_proxy<CompressedIntArray*>, CompressedIntArray*> iterator;
    
    iterator begin() { return iterator(this, 0); }
    iterator end()   { return iterator(this, MaxCount); }
    const_iterator begin() const { return const_iterator(this, 0); }
    const_iterator end()   const { return const_iterator(this, MaxCount); }
    
    const unsigned operator[] (unsigned n) const { return Rep.Get_nth(n); }
    storage_proxy<CompressedIntArray*> operator[] (unsigned n) { return *(begin()+n); }
public:
    template<typename IntType>
    void Set_all(IntType data[MaxCount])
    {
        Rep.Set_all(data);
    }
    
    template<typename IntType>
    void Get_all(IntType data[MaxCount]) const
    {
        Rep.Get_all(data);
    }
}     __attribute__((packed));

#endif
