#include "NewRadixSet.hh"

#include <algorithm>

class SmallCompleter
{
public:
    SmallCompleter(const LevelState& lev)
    {
        initial_state = lev;
        initial_state.RebuildBlockCaches();
        seenmap.clear();
    }
    
    struct CheckFinished
    {
        CheckFinished() : result(false) { }
        
        void operator() (const LevelState& state)
        {
            if(state.IsCompleted())
            {
                result=true;
                throw 0;
            }
        }
        
        bool result;
    };
    bool IsCompleteable()
    {
        typedef CheckFinished CheckType;
        CheckType check;
        checkfunptr = (void*)&check;
        
        try { TryMoves<CheckType> (initial_state); } catch(int) { }
        
        return check.result;
    }
    
    template<unsigned BlockCount>
    struct CheckPushable
    {
        CheckPushable(unsigned poslist[BlockCount]) : result(NonCoordinate)
        {
            for(unsigned a=0; a<BlockCount; ++a)
                targetpos[a] = poslist[a];
            std::sort(targetpos, targetpos+BlockCount);
        }
        
        void operator() (const LevelState& state)
        {
            for(unsigned a=0; a<BlockCount; ++a)
                if(state.GetBlockPos(a) != targetpos[a])
                    return;
            result=state.GetPlayerPos();
            throw 0;
        }
        
        unsigned targetpos[BlockCount];
        CoordinateType result;
    };

    template<unsigned BlockCount>
    CoordinateType IsPushableInto(unsigned targetposlist[BlockCount])
    {
        typedef CheckPushable<BlockCount> CheckType;
        CheckType check(targetposlist);
        checkfunptr = (void*)&check;
    
        try { TryMoves<CheckType>(initial_state); } catch(int) { }
        
        return check.result;
    }
    
    template<unsigned Depth>
    struct IgnoreMarker
    {
        void operator() (const LevelState& state)
        {
            unsigned plrpos = state.GetPlayerPos();
            unsigned blockposlist[Depth];
            for(unsigned a=0; a<Depth; ++a)
                blockposlist[a] = state.GetBlockPos(a);

            DeadZones.RegisterMultiIgnore<Depth> (plrpos, blockposlist);
        }
    };

    template<unsigned Depth>
    void MarkEverythingIgnorable()
    {
        typedef IgnoreMarker<Depth> CheckType;
        CheckType check;
        checkfunptr = (void*)&check;
        TryMoves<CheckType>(initial_state);
    }

private:
    template<typename FuncType>
    void TryMoves(const LevelState& state)
    {
        typedef FavorablePointList targetlist;
        
        /* checkfunptr points to an object of type "FuncType".
         * Cast the pointer into the right type, and call the
         * provided function.
         *
         * checkfunptr is a member to SmallCompleter to avoid
         * having to pass it as a parameter to each recursion cycle.
         */
        ((FuncType*)(checkfunptr))->operator() (state);
        
        targetlist targets;
        state.FindWalkingDestinations(targets);
        
        for(targetlist::const_iterator i = targets.begin(); i != targets.end(); ++i)
        {
            const CoordinateType Point = *i;
            #define TryDir(offs) { int offset = offs; \
                if(state.CanPushFrom(Point, offset)) \
                { \
                    LevelState NewState = state; \
                    NewState.PushFromNoMove(Point, offset); \
                    if(!NewState.CheckBlunder(Point+offset*2,offset)) \
                    { \
                      NewState.WalkTo(Point+offset); \
                      if(seenmap.add(NewState.GetData())) \
                      { \
                        TryMoves<FuncType>(NewState); \
                      } \
                    } \
                    state.RebuildBlockCaches(); \
                } \
              }
            TryDir(-1);
            TryDir(1);
            TryDir(-MaxWidth);
            TryDir(MaxWidth);
            #undef TryDir
        }
    }

    /* Smallcompleter MUST USE NON-COMPACTED coordinates, because
     * non-compacted coordinates are only built after DeadZones
     * are calculated, and Smallcompleter is used during DeadZones
     * calculation.
     */    
private:
    LevelState initial_state;
    void* checkfunptr;
    
    typedef NewRadixSet<CoordinateType,MaxDim,MaxObject> seenmaptype;
    seenmaptype seenmap;
};
