#include "storage.hh"
#include "templates.hh"

class VectorStorage
{
private:
    typedef unsigned short LengthType;
    typedef MemoryStorage<unsigned char> DataStorageType;

public:
    class VecType
    {
    private:
        unsigned char* Buf;
        LengthType Length;
        bool owned, changed;
    private:
        /* No copying. */
        VecType(const VecType&);
        void operator=(const VecType&);
    public:
        VecType(): Buf(0),Length(0),owned(false),changed(false) { }
        ~VecType() { if(owned && Buf) std::free(Buf); }

        inline bool was_changed() const { return changed; }
        inline void reset_change() { changed=false; }
    
        inline const unsigned char operator[] (unsigned ind) const
        {
            return Buf[ind];
        }
        inline void initial_assign_readonly
            (const unsigned char* begin, const unsigned char* end)
        {
            if(owned && Buf) std::free(Buf);
            Buf    = const_cast<unsigned char*>(begin);
            Length = end-begin;
            owned  = false;
            changed= false;
        }
        inline void assign(const unsigned char* begin, const unsigned char* end)
        {
            unsigned NewLen = end-begin;
            if(!owned) { Buf = (unsigned char*)std::malloc(Length = NewLen); owned=true; }
            if(Length != NewLen)
            {
                if(Buf) std::free(Buf);
                Buf = (unsigned char*) std::malloc(Length = NewLen);
            }
            std::memcpy(Buf, begin, Length);
            changed=true;
        }
        inline bool empty() const { return Length == 0; }
        inline unsigned size() const { return Length; }
        inline void clear()
        {
            if(owned && Buf) std::free(Buf);
            Buf=0;
            Length=0;
            owned=false;
            changed=true;
        }
        void replace(unsigned beginpos, unsigned orig_size,
                     const unsigned char* source, unsigned new_size);
        inline const unsigned char* GetReadPtr(unsigned pos, unsigned length) const
        {
            if(!Length) return (const unsigned char*)"";
            return Buf+pos;
        }
    };

public:
    /* The storage introduced by Read().
     * It acts like a vector, and it masquerades as the
     * actual storage of the SimpleVec item, but in fact
     * it's just a relay.
     */
    static VecType Buf;

private:
    /* The storage where all data is stored in.
     * It is governed by the StupidVec items and clear().
     */
    static DataStorageType storage;

private:
    class StupidVec
    {
        unsigned long Ptr;
        LengthType Length, AllocLength;
    public:
        StupidVec() : Ptr(0),Length(0),AllocLength(0) { }
        /* Warning: destructor is never called. */
        
        void Reassign(const VecType& vec);
        const unsigned char* GetBuf() const
        {
            return storage.GetReadPtr(Ptr, Length);
        }
        unsigned GetLength() const { return Length; }
    } __attribute__((packed));
    
public:
    typedef unsigned long IDtype;
private:
    static IDtype BufOwner;
    static IDtype ID;
    static unsigned used;
    
    /* The storage of pointers to StupidVec items.
     * Indexed by the "IDs" which are the public
     * interface to this allocator.
     * It is governed by Create(), Free1() and clear().
     */
    typedef MemoryStorage<StupidVec> PointerStorageType;
    static PointerStorageType ptrs;
public:
    static IDtype Create()
    {
        ++used;
        
        StupidVec* ptr = ptrs.GetWritePtr(ID);
        // Construct the item in the buffer
        new(ptr) StupidVec;
        return ID++;
    }
    static void clear()
    {
        ID=0;
        ptrs.clear();
        storage.clear();
        BufOwner=0; Buf.clear();
        used=0;
    }
    static void Free1()
    {
        --used;
        if(!used) clear();
    }
    
    /* References returned by this call are guaranteed to be
     * valid only until the next call of Read(), Create() or Free1().
     */
    static VecType& Read(IDtype n)
    {
        if(n != BufOwner)
        {
            if(Buf.was_changed())
            {
                ptrs.GetWritePtr(BufOwner)->Reassign(Buf);
            }
            const StupidVec& v = *ptrs.GetReadPtr(BufOwner=n);
            
            const unsigned char* Source = v.GetBuf();
            Buf.initial_assign_readonly(Source, Source+v.GetLength());
        }
        return Buf;
    }
};

