#include "intarray.hh"
#include "simplevec.hh"
#include "coordinates.hh"
#include "settings.hh"
#include "mapdata.hh"
#include "bigint.hh"

#include <bitset>
#include <iostream>

extern unsigned DeadCounter1, DeadCounter0;
extern unsigned MultiDeadCounter[USE_MULTI_DEAD+1];
extern bool ShowDead;
extern bool PlayingNormally;

/* LevelState contains LevelStateData contains LevelIdentifier.
 * Then there's also CompressedLevel.
 * All clear?
 */

struct FavorablePointWithCost
{
    CoordinateType Point;
    unsigned short Cost;
    
    inline FavorablePointWithCost(unsigned p,unsigned c) : Point(p), Cost(c) { }
    inline FavorablePointWithCost() { }
    
    inline bool operator< (const FavorablePointWithCost& b) const
    {
        return Cost < b.Cost;
    }
};
typedef SimpleVec<CoordinateType, MaxDim> WalkingRoute;
typedef SimpleVec<CoordinateType, 1+4*(MaxObject-1)> FavorablePointList;

/* Favorable positions may include current position and each 4 sides of each block. */
typedef SimpleVec<FavorablePointWithCost, 1+4*(MaxObject-1)> FavorablePointListWithCost;

typedef CompressedIntArray<MaxDim+1, MaxObject> CompressedLevel;

extern char WallMap[MaxDim]; /* Array of wall and goal bits */

#if 1
struct ObstacleMapType
{
public:
    inline void reset() { std::fill_n(data,MaxDim,false); }
    inline void set(unsigned n)
    {
        //if(n >= MaxDim) std::abort();
        data[n]=true;
    }
    inline void reset(unsigned n)
    {
        //if(n >= MaxDim) std::abort();
        data[n]=false;
    }
    inline bool test(unsigned n) const
    {
        //if(n >= MaxDim) std::abort();
        return data[n];
    }
private:
    bool data[MaxDim];
};
#else
typedef std::bitset<MaxDim> ObstacleMapType;
#endif
extern ObstacleMapType DefaultObstacleMap;
extern ObstacleMapType CurrentObstacleMap;
extern std::bitset<FloorAmount> CurrentBlockMap;

struct LevelIdentifier
{
public:
    CoordinateType ID[MaxObject];
public:
    const CompressedLevel Compress() const
    {
        CompressedLevel result;
        result.Set_all(ID);
        return result;
    }
    void Uncompress(const CompressedLevel& data)
    {
        data.Get_all(ID);
    }
    BigInt& GetCompactData() const;

    void Dump() const
    {
        std::cerr << "Identifier:";
        for(unsigned a=0; a<MaxObject; ++a)
            std::cerr << ' ' << (int)ID[a];
        std::cerr << "          \n";
    }
    
    void Reset()
    {
        CoordinateType* const ID_begin = ID;
        CoordinateType* const ID_end   = ID+MaxObject;
        std::fill(ID_begin, ID_end, NonCoordinate);
    }
    
    void RedefineBlock(unsigned pos0, unsigned pos1) REGPARM
    {
        //if(pos1 == pos0) return;
        
        CoordinateType* ID_begin = ID+1;
        CoordinateType* ID_end   = ID+MaxObject;
        CoordinateType* src  = std::lower_bound(ID_begin,ID_end,(CoordinateType)pos0);
        if(src >= ID_end || *src != pos0) return;
        
        CoordinateType* ID_up1 = ID_begin;
        CoordinateType* ID_up2 = ID_end;
        if(pos1 > pos0) ID_up1 = src; else ID_up2 = src+1;
        CoordinateType* dest = std::upper_bound(ID_up1,ID_up2, (CoordinateType)pos1);
        if(src < dest)
        {
            std::copy(src+1, dest, src);
            --dest;
        }
        else if(src > dest)
        {
            std::copy_backward(dest,src,src+1);
        }
        *dest = pos1;
    }
    void SetPlayer(unsigned pos)
    {
        ID[0] = pos;
    }
    unsigned GetPlayer() const
    {
        return ID[0];
    }
    inline unsigned GetBlockPos(unsigned blockno) const
    {
        return ID[blockno+1];
    }
    
    bool HasBlock(CoordinateType pos) const REGPARM
    {
        const CoordinateType* const ID_begin = ID;
        const CoordinateType* const ID_end   = ID+MaxObject;
        return std::binary_search(ID_begin+1, ID_end, pos);
    }
    /*
    unsigned IdentifyBlock(CoordinateType pos) const
    {
        const CoordinateType* const ID_begin = ID;
        const CoordinateType* const ID_end   = ID+MaxObject;
        const CoordinateType* ptr ( std::lower_bound(ID_begin+1, ID_end, pos) );
        if(ptr >= ID_end) return 0;
        if(*ptr == pos) return 0;
        return ptr - ID_begin;
    }
    */
} __attribute__((packed));

struct LevelStateData
{
public:
    const CompressedLevel Compress() const { return State.Compress(); }
    const CoordinateType*const GetData() const { return State.ID; }
    BigInt& GetCompactData() const { return State.GetCompactData(); }
private:
    friend class LevelState;
    LevelIdentifier State;
    
    void BuildFromLevel();
    
    void BuildFrom(const CompressedLevel& data)
    {
        State.Uncompress(data);
    }
    
    void MovePlayer(unsigned x2,unsigned y2)
    {
        MovePlayer(y2*MaxWidth+x2);
    }
    void MovePlayer(unsigned pos1)
    {
        State.SetPlayer(pos1);
    }
    void MoveBlock(unsigned x1,unsigned y1, unsigned x2,unsigned y2)
    {
        unsigned pos0=y1*MaxWidth+x1;
        unsigned pos1=y2*MaxWidth+x2;
        MoveBlock(pos0, pos1);
    }
    void MoveBlock(unsigned pos0, unsigned pos1)
    {
        State.RedefineBlock(pos0,pos1);
    }
    static inline unsigned at(unsigned x,unsigned y)
    {
        return WallMap[y*MaxWidth+x];
    }
    static inline unsigned at(unsigned pos)
    {
        return WallMap[pos];
    }
    unsigned GetPX() const
    {
        return GetPlayer() % MaxWidth;
    }
    unsigned GetPY() const
    {
        return GetPlayer() / MaxWidth;
    }
    CoordinateType GetPlayer() const { return State.GetPlayer(); }

    inline bool IsGoal(unsigned pos) const { return at(pos) & MapBits::Goal; }
    inline bool IsWall(unsigned pos) const { return at(pos) & MapBits::Wall; }

    void Dump(unsigned focus=NonCoordinate) const;
} __attribute__((packed));

struct LevelState
{
public:
    const CompressedLevel Compress() const { return data.Compress(); }
    const CoordinateType*const GetData() const { return data.GetData(); }
    BigInt& GetCompactData() const { return data.GetCompactData(); }
    
    LevelState()
    {
    }
    
    LevelState(const LevelState& b) : data(b.data)
    {
    }
    
    LevelState& operator=(const LevelState& b)
    {
        data = b.data;
        return *this;
    }
    
    void CreateFromLevel()
    {
        data.BuildFromLevel();
        /*RebuildBlockCaches();
         * - disabled, because the call is not allowed from BuildFloorCoordinates()
         */
    }
    
    void CreateFrom(const CompressedLevel& com)
    {
        data.BuildFrom(com);
        RebuildBlockCaches();
    }
    
    void RebuildBlockCaches() const
    {
        CurrentBlockMap.reset();
        
        CurrentObstacleMap = DefaultObstacleMap;
        for(unsigned n=1; n<MaxObject; ++n)
        {
            unsigned pos = GetBlockPos(n-1);
            if(unlikely(pos == NonCoordinate)) break;
            CurrentBlockMap.set(FloorCoordinateTranslation[pos]);
            CurrentObstacleMap.set(pos);
        }
    }

    template<unsigned BlockCount>
    void SetNBlockSituation(unsigned playerpos, const unsigned blockposlist[BlockCount])
    {
        data.State.Reset();
        data.MovePlayer(playerpos);
        for(unsigned a=0; a<BlockCount; ++a)
            data.MoveBlock(NonCoordinate, blockposlist[a]);
        RebuildBlockCaches();
    }

    inline unsigned GetPX() const { return data.GetPX(); }
    inline unsigned GetPY() const { return data.GetPY(); }    
    inline unsigned GetBlockPos(unsigned n) const { return data.State.GetBlockPos(n); }
    inline unsigned GetPlayerPos() const { return data.GetPlayer(); }
    
    void Dump(unsigned focus=NonCoordinate) const { data.Dump(focus); }
    
    bool IsCompleted() const
    {
        /* For each block, see if they're in a goal. */
        for(unsigned n=1; n<MaxObject; ++n)
        {
            unsigned pos = GetBlockPos(n-1);
            if(/*unlikely*/(pos == NonCoordinate)) break;
            if(/*likely*/(!data.IsGoal(pos))) return false;
        }
        return true;
    }
    
    static inline bool PushForbidden(unsigned blockpos, int offset)/*const*/
    {
#if CHOOSE_LEVEL==15
        /* Special rule in 4-1: Don't push right on the goal area. */
        if(offset == 1)
        {
            unsigned x = blockpos%MaxWidth;
            unsigned y = blockpos/MaxWidth;
            if(x <= 7 && y <= 5) return true;
        }
        GARGL
#endif
#if CHOOSE_LEVEL==11
        /* Special rule in 3-2: Don't push right on the goal area. */
        if(offset == 1)
        {
            unsigned x = blockpos%MaxWidth;
            unsigned y = blockpos/MaxWidth;
            if(x >= 3 && x <= 5 && y <= 5) return true;
        }
#endif
        return false;
    }
    
    inline bool CanPush(int offset) const { return CanPushFrom(GetPlayerPos(), offset); }
    bool CanPushFrom(unsigned playerpos, int offset) const
    {
        unsigned blockpos = playerpos + offset;
        unsigned frontpos = blockpos + offset;
        return IsBlock(blockpos)
            && !PushForbidden(blockpos, offset)
            && !CurrentObstacleMap.test(frontpos);
    }
    unsigned CalculateKick(int dx,int dy) const { return CalculateKick(dy*MaxWidth+dx); }
    unsigned CalculateKick(int offset) const { return CalculateKickFrom(GetPlayerPos(), offset); }
    unsigned CalculateKickFrom(unsigned playerpos, int offset) const
    {
        unsigned blockpos = playerpos + offset; // location of the block
        
        if(!IsBlock(blockpos))
        {
            //std::cerr << "No block at " << blockpos << " (from " << playerpos << ")\n";
            return 0;
        }
        if(PushForbidden(blockpos, offset)) return 0;
        unsigned length = 0;
        /* No bounds checking! The field should have walls! */
        while(!CurrentObstacleMap.test(blockpos+offset))
        {
            blockpos += offset;
            ++length;
        }
        //std::cerr << "Got block at " << blockpos << ", length " << length << "\n";
        return length;
    }
    
    /* Walking route handling */
    void FindWalkingDestinations(FavorablePointListWithCost& targets) const;
    void FindWalkingDestinations(FavorablePointList& targets) const;
    void FindWalkingRouteTo(unsigned nx,unsigned ny, WalkingRoute& result) const;
    bool TestWalkingRoute(const unsigned origin, const unsigned goal) const;
    SimpleVec<CoordinateType,MaxDim> BuildWalkingMaps() const;
    
    bool IsWall(unsigned pos) const { return data.IsWall(pos); }
    bool IsBlock(unsigned pos) const
    {
        unsigned coord = FloorCoordinateTranslation[pos];
        return coord < FloorAmount && CurrentBlockMap.test(coord);
    }
    
    void Walk(int dx,int dy)             { WalkTo(data.GetPX()+dx, data.GetPY()+dy); }
    void WalkTo(CoordinateType pos)      { data.MovePlayer(pos); }
    void WalkTo(unsigned nx,unsigned ny) { data.MovePlayer(nx,ny); }
    unsigned Push(int dx,int dy)         { return Push(dy*MaxWidth+dx); }
    unsigned Push(int offset)
    {
        const unsigned pos = GetPlayerPos();
        PushFromNoMove(pos, offset);
        WalkTo(pos+offset);
        return pos+offset+offset;
    }
    inline void Kick(int dx,int dy, unsigned distance)
    {
        Kick(dy*MaxWidth+dx, distance);
    }
    inline unsigned Kick(int offset, unsigned distance)
    {
        return KickFrom(GetPlayerPos(), offset, distance);
    }
    inline void PushFromNoMove(unsigned pos, int offset)
    {
        unsigned oldpos = pos    + offset;
        unsigned newpos = oldpos + offset;
        MoveBlockFrom(oldpos, newpos);
    }
    void MoveBlockFrom(unsigned oldpos, unsigned newpos) REGPARM
    {
        CurrentObstacleMap.reset(oldpos);
        CurrentObstacleMap.set(newpos);
        
        /*
        if(oldpos >= MaxDim || newpos >= MaxDim
        || FloorCoordinateTranslation[oldpos] >= FloorAmount
        || FloorCoordinateTranslation[newpos] >= FloorAmount)
        {
            Dump(oldpos);
            abort();
        }
        */
        CurrentBlockMap.reset(FloorCoordinateTranslation[oldpos]);
        CurrentBlockMap.set(FloorCoordinateTranslation[newpos]);
        data.MoveBlock(oldpos, newpos);
    }
    unsigned KickFrom(unsigned pos, int offset, unsigned distance) REGPARM
    {
        unsigned oldpos = pos+offset, newpos = oldpos+offset*distance;
        CurrentObstacleMap.reset(oldpos);
        CurrentObstacleMap.set(newpos);
        CurrentBlockMap.reset(FloorCoordinateTranslation[oldpos]);
        CurrentBlockMap.set(FloorCoordinateTranslation[newpos]);
        data.MoveBlock(oldpos, newpos);
        return newpos;
    }
    
    /* CheckBlunder.
     * Block position and the direction it was moven are given.
     * Player is assumed to face next to the block in the direction
     * where block was moved from.
     */
    bool CheckBlunder(unsigned blockpos, int offset) const REGPARM;
private:
    LevelStateData data;
} __attribute__((packed));

/*
namespace __gnu_cxx
{
    template<> struct hash<CompressedLevel>
    {
        inline size_t operator() (const CompressedLevel& lev) const
        {
            return lev.hash();
        }
    };
} 
*/
