#include "filewindow.hh"

#include "deadzones.hh"
#include "levelstate.hh"
#include "smallcompleter.hh"
#include "bigint.hh"
#include "binom.hh"

#include <sstream>
#include <string>
#include <cmath> /* For std::fmod */
#include <set>

#include <sys/time.h>

//#include "deadcompleter.hh"

static inline bool DeadFind(const DeadZones::DeadBlkSet& a, unsigned plrpos)
{
    return a.plrpos < plrpos;
}
static inline bool DeadCompBlockSet(const DeadZones::BlockSet& a, const DeadZones::BlockSet& b)
{
    unsigned ac = a.count();
    unsigned bc = b.count();
    /*if(ac != bc) */return ac < bc;
    /*return a.blockmap < b.blockmap;*/
}
static inline bool DeadComp(const DeadZones::DeadBlkSet& a, const DeadZones::DeadBlkSet& b)
{
    if(a.plrpos != b.plrpos) return a.plrpos < b.plrpos;
    return DeadCompBlockSet(a.blockmap, b.blockmap);
}

template<unsigned b>
static inline bool DeadCount(const DeadZones::DeadBlkSet& a)
{
    return a.blockmap.count() == b;
}

template<unsigned b>
static inline bool DeadCountBlockSet(const DeadZones::BlockSet& a)
{
    return a.count() == b;
}

static const std::string GetDeadDataFileName(unsigned Depth)
{
    std::stringstream s;
    s << "dead-" << Depth << "-" << levelno << ".dat";
    return s.str();
}

template<unsigned Depth>
void DeadZones::RegisterMultiDead(unsigned plrpos, const unsigned blockposlist[Depth],
                                  bool Sort)
{
    DeadBlkSet blks;
    blks.plrpos = plrpos;
    for(unsigned a=0; a<Depth; ++a)
        blks.blockmap.set(FloorCoordinateTranslation[blockposlist[a]]);
    
    if(Sort)
    {
        DeadBlkSetVec::iterator j = std::upper_bound
           (MultiDeadFlags.WithPlayer.begin(),
            MultiDeadFlags.WithPlayer.end(), blks, DeadComp);
        MultiDeadFlags.WithPlayer.insert(j, blks);
    }
    else
        MultiDeadFlags.WithPlayer.push_back(blks);
    
    for(unsigned a=0; a<Depth; ++a)
    {
        unsigned bpos = blockposlist[a];
        
        DeadFlagData& target = MultiDeadFlagsByBlock[bpos];
        
        if(Sort)
        {
            DeadBlkSetVec::iterator j = std::upper_bound
               (target.WithPlayer.begin(), target.WithPlayer.end(), blks, DeadComp);
            target.WithPlayer.insert(j, blks);
        }
        else
            target.WithPlayer.push_back(blks);
    }
    
    /* Don't write to .Set here. .Set is only for combinations
     * that are blunder regardless of player's position.
     */
}

template<unsigned Depth>
void DeadZones::RegisterMultiIgnore(unsigned plrpos, const unsigned blockposlist[Depth],
                                    bool Sort)
{
    /* Almost the same as RegisterMultiDead. Instead, write "Ignore" data */
    DeadBlkSet blks;
    blks.plrpos = plrpos;
    for(unsigned a=0; a<Depth; ++a)
        blks.blockmap.set(FloorCoordinateTranslation[blockposlist[a]]);
    
    if(Sort)
    {
        DeadBlkSetVec::iterator j = std::upper_bound
           (MultiIgnoreFlags.WithPlayer.begin(),
            MultiIgnoreFlags.WithPlayer.end(), blks, DeadComp);
        MultiIgnoreFlags.WithPlayer.insert(j, blks);
    }
    else
        MultiIgnoreFlags.WithPlayer.push_back(blks);
    
    for(unsigned a=0; a<Depth; ++a)
    {
        unsigned bpos = blockposlist[a];
        
        DeadFlagData& target = MultiIgnoreFlagsByBlock[bpos];
        if(Sort)
        {
            DeadBlkSetVec::iterator j = std::upper_bound
               (target.WithPlayer.begin(), target.WithPlayer.end(), blks, DeadComp);
            target.WithPlayer.insert(j, blks);
        }
        else
            target.WithPlayer.push_back(blks);
    }
}

template<unsigned Depth>
void DeadZones::RegisterAlwaysMultiDead(const unsigned blockposlist[Depth])
{
    BlockSet blks;
    for(unsigned a=0; a<Depth; ++a)
        blks.set(FloorCoordinateTranslation[blockposlist[a]]);

    MultiDeadFlags.Set.push_back(blks);
    
    for(unsigned a=0; a<Depth; ++a)
    {
        unsigned bpos = blockposlist[a];
        
        DeadFlagData& target = MultiDeadFlagsByBlock[bpos];
        target.Set.push_back(blks);
    }
    
    /* Don't write to .Set here. .Set is only for combinations
     * that are blunder regardless of player's position.
     */
}

#define DigitMagic 0xF11F2EE2

template<unsigned Depth>
void DeadZones::SaveDeadData(const unsigned char digits[Depth])
{
    const std::string fn = GetDeadDataFileName(Depth);
    
    FILE* fp = fopen(fn.c_str(), "wb");

    std::cerr << "// Saving " << fn << "...\n";
    
    putw(DigitMagic, fp);
    fwrite(digits, 1, Depth, fp);
    
    /* Entries:
         -5 -5 x0 y0 x1 y1 x2 y2    = MultiDeadFlags without player position
         px py x0 y0 x1 y1 x2 y2    = MultiDeadFlags with    player position
      -6 px py x0 y0 x1 y1 x2 y2    = MultiIgnoreFlags with  player position
    */
    
    {const BlockSetVec& data = MultiDeadFlags.Set;
    for(unsigned a=0; a<data.size(); ++a)
    {
        const BlockSet& blks = data[a];
        if(blks.count() != Depth) continue;
        
        //printf("%s\n", blks.to_string().c_str());
        
        putw(-5,fp);putw(-5,fp);

        unsigned ind = blks._Find_first();
        for(unsigned b=1; b<=Depth; ++b)
        {
            unsigned coord = FloorCoordinateUntranslation[ind];
            
            //printf("%u,%u[%u] ", coord%MaxWidth,coord/MaxWidth,ind);
            
            unsigned x = coord % MaxWidth + OffsetWidth;
            unsigned y = coord / MaxWidth + OffsetHeight;
            putw(x,fp);putw(y,fp);
            
            ind = blks._Find_next(ind);
        }
        //printf("\n");
    }}

    {const DeadBlkSetVec& data = MultiDeadFlags.WithPlayer;
    for(unsigned a=0; a<data.size(); ++a)
    {
        const DeadBlkSet& blks = data[a];
        if(blks.blockmap.count() != Depth) continue;
        
        //printf("%s\n", blks.blockmap.to_string().c_str());
        
        {unsigned coord = blks.plrpos;
        unsigned x = coord % MaxWidth + OffsetWidth;
        unsigned y = coord / MaxWidth + OffsetHeight;
        putw(x,fp);putw(y,fp);}

        unsigned ind = blks.blockmap._Find_first();
        for(unsigned b=1; b<=Depth; ++b)
        {
            unsigned coord = FloorCoordinateUntranslation[ind];
            
            //printf("%u,%u[%u] ", coord%MaxWidth,coord/MaxWidth,ind);
            
            unsigned x = coord % MaxWidth + OffsetWidth;
            unsigned y = coord / MaxWidth + OffsetHeight;
            putw(x,fp);putw(y,fp);
            
            ind = blks.blockmap._Find_next(ind);
        }
        //printf("\n");
    }}
    
    putw(-6, fp);

    {const DeadBlkSetVec& data = MultiIgnoreFlags.WithPlayer;
    for(unsigned a=0; a<data.size(); ++a)
    {
        const DeadBlkSet& blks = data[a];
        if(blks.blockmap.count() != Depth) continue;
        
        //printf("%s\n", blks.blockmap.to_string().c_str());
        
        {unsigned coord = blks.plrpos;
        unsigned x = coord % MaxWidth + OffsetWidth;
        unsigned y = coord / MaxWidth + OffsetHeight;
        putw(x,fp);putw(y,fp);}

        unsigned ind = blks.blockmap._Find_first();
        for(unsigned b=1; b<=Depth; ++b)
        {
            unsigned coord = FloorCoordinateUntranslation[ind];
            
            //printf("%u,%u[%u] ", coord%MaxWidth,coord/MaxWidth,ind);
            
            unsigned x = coord % MaxWidth + OffsetWidth;
            unsigned y = coord / MaxWidth + OffsetHeight;
            putw(x,fp);putw(y,fp);
            
            ind = blks.blockmap._Find_next(ind);
        }
        //printf("\n");
    }}

    fclose(fp);
}

template<unsigned Depth>
void DeadZones::LoadDeadData(unsigned char digits[Depth])
{
    const std::string fn = GetDeadDataFileName(Depth);
    
    FILE* fp = fopen(fn.c_str(), "rb");
    if(!fp) return;

    std::cerr << "// Loading " << fn << "...\n";
    
    unsigned magic = getw(fp);
    if(magic == DigitMagic)
    {
        fread(digits, 1, Depth, fp);
    }
    else
    {
        if((int)magic != EOF) std::fill(digits,digits+Depth,0xFF);
        rewind(fp);
    }
    
    bool ignore=false;
    for(unsigned a=0;; ++a)
    {
        int plrpos=0;
        unsigned Coords[Depth];
        for(unsigned b=0; b<=Depth; ++b)
        {
            int x = getw(fp);
            
            if(x == -6)
            {
                ignore=true;
                x = getw(fp);
            }
            
            int y = getw(fp);
            if(x==EOF) goto Done;
            
            if(b==0 && x==-5)
                plrpos = -5;
            else
            {
                unsigned coord = (y-OffsetHeight)*MaxWidth + (x-OffsetWidth);
                if(b==0)
                    plrpos = coord;
                else
                {
                    Coords[b-1] = coord;
                }
            }
        }
        if(plrpos < 0)
            RegisterAlwaysMultiDead<Depth>(Coords);
        else if(!ignore)
            RegisterMultiDead<Depth>(plrpos, Coords, false);
        else
            RegisterMultiIgnore<Depth>(plrpos, Coords, false);
    }
Done:
    fclose(fp);
    
    std::sort(MultiDeadFlags.WithPlayer.begin(), MultiDeadFlags.WithPlayer.end(), DeadComp);
    std::sort(MultiIgnoreFlags.WithPlayer.begin(), MultiIgnoreFlags.WithPlayer.end(), DeadComp);
    
    std::map<CoordinateType/*normal block coordinate*/, DeadFlagData>::iterator i;
    for(i=MultiIgnoreFlagsByBlock.begin();
        i!=MultiIgnoreFlagsByBlock.end();
        ++i)
    {
        std::sort(i->second.WithPlayer.begin(), i->second.WithPlayer.end(), DeadComp);
    }
    for(i=MultiDeadFlagsByBlock.begin();
        i!=MultiDeadFlagsByBlock.end();
        ++i)
    {
        std::sort(i->second.WithPlayer.begin(), i->second.WithPlayer.end(), DeadComp);
    }
}

template<unsigned Depth>
void DeadZones::DeadMultiBlockBuilder
    (const unsigned blockposlist[Depth],
     unsigned &ndead,
     unsigned &nalldead,
     unsigned &nignore)
{
    if(MaxObject-1 < Depth) return; // Everything would be dead otherwise.
    
    LevelState level;
    level.SetNBlockSituation<Depth>(0, blockposlist);
    // ^Also assumed to build the Block coordinate data

    const SimpleVec<CoordinateType,MaxDim> routes = level.BuildWalkingMaps();
    /* Now, routes is an array,
     * which contains a value for each point of the room.
     * Points which are reachable by walking are painted
     * the same colour (number). Each disconnected region
     * has its own colour.
     * To test whether path A-B is clear, you only need
     * to check if A and B have the same value in this array.
     */

    if(CheckMultiBlockDeadBlock()) return;

    enum MultiDeadVerdict { unknown='\0', notdead='\1', dead='\2', ignore='\3' };
    MultiDeadVerdict verdict[MaxDim];
    std::memset(&verdict,unknown,sizeof(verdict));
    
    std::set<CoordinateType> playerpositions;
    for(unsigned a=0; a<Depth; ++a)
    {
        const unsigned blockpos = blockposlist[a]; 
        #define TestPlrDir(dir) \
          { \
            unsigned plrpos = blockpos + (dir); \
            if(!CurrentObstacleMap.test(plrpos)) \
            { \
                unsigned route = routes[plrpos]; \
                if(verdict[route] == unknown \
                && playerpositions.find(plrpos) == playerpositions.end()) \
                { \
                    if(CheckMultiBlockDead(plrpos) \
                    || CheckMultiBlockIgnore(plrpos)) \
                    { \
                        verdict[route] = ignore; \
                    } \
                    else \
                        playerpositions.insert(plrpos); \
                } \
            } \
          }
        TestPlrDir(-1);
        TestPlrDir( 1);
        TestPlrDir(-MaxWidth);
        TestPlrDir( MaxWidth);
        #undef TestPlrDir
    }
    
    if(playerpositions.empty()) return;

    bool all_are_dead = true;
    
#if 1
    /*
     * Because the combinations are processed from top-left to bottom-right,
     * we can do the following test to reduce the number of calculations.
     *
     * For each block, check if it can be moved to the left or to the up.
     * If it can, check if that new position (which already has been processed)
     * was dead. If it wasn't dead, this one can't be either because a dead
     * state can not become not-dead.
     */
    for(unsigned a=0; a<Depth; ++a)
    {
        const unsigned blockpos = blockposlist[a];
        #define TestPush(dir) \
          { \
            /* dir = direction to push the block towards */ \
            const unsigned plrpos    = blockpos - (dir); \
            const unsigned targetpos = blockpos + (dir); \
            const unsigned route = routes[plrpos]; \
            if(verdict[route] == unknown) \
            { \
                if(!CurrentObstacleMap.test(plrpos) \
                && !CurrentObstacleMap.test(targetpos)) \
                { \
                    __label__ skipundo; \
                    const unsigned newplrpos = blockpos; \
                    level.MoveBlockFrom(blockpos, targetpos); /* Move the block */ \
                    if(!level.CheckBlunder(targetpos,dir))/*
                    if(!CheckDead(targetpos) \
                    && !CheckMultiBlockDead(newplrpos))*/ /* If this position was not dead */ \
                    { \
                        verdict[route] = notdead; /* This one is not either */ \
                        all_are_dead = false; \
                    } \
                    else if(false) \
                    { \
                        /* Tarkista, voiko sen blokin tynt takaisin
                         * thn positioon ja pelaajan palauttaa siihen
                         * mist lhti. Jos voi, tmkin on blunder. */ \
                        /* Disabled because this check doesn't save anything */ \
                        unsigned NewPushPos = targetpos + (dir); \
                        if(level.TestWalkingRoute(newplrpos, NewPushPos)) \
                        { \
                            level.MoveBlockFrom(targetpos, blockpos); \
                            if(level.TestWalkingRoute(targetpos, plrpos)) \
                            { \
                                std::cerr << "\nHIT\n"; \
                                verdict[route] = dead; \
                            } \
                            goto skipundo; \
                        } \
                    } \
                    level.MoveBlockFrom(targetpos, blockpos); /* Undo the move */ \
                skipundo: ; \
                } \
            } \
          }
        TestPush(-1);
        TestPush(-MaxWidth);
        #undef TestPush
    }
#endif
    
    SimpleVec<CoordinateType, Depth*4> found_dead;
    
    for(typename std::set<CoordinateType>::const_iterator
        i = playerpositions.begin(); i != playerpositions.end(); ++i)
    {
        const CoordinateType plrpos = *i;
        const unsigned route = routes[plrpos];
        switch(verdict[route])
        {
            case ignore:  continue;
            case notdead: continue;
            case dead:    found_dead.push_back(plrpos); continue;
            case unknown: break;
        }
        level.SetNBlockSituation<Depth>(plrpos, blockposlist);
        if(SmallCompleter(level).IsCompleteable())
        {
            verdict[route] = notdead;
            all_are_dead = false;
        }
        else
        {
            verdict[route] = dead;
            found_dead.push_back(plrpos);
        }
    }
    
    unsigned nfound_dead = found_dead.size();
    if(all_are_dead)
    {
        if(nfound_dead > 0)
        {
            /* All positions were dead. */
            RegisterAlwaysMultiDead<Depth> (blockposlist);
            ++nalldead;
        }
    }
    else
    {
        /* Not all positions were dead. Mark those which were. */

        for(unsigned a=0; a<nfound_dead; ++a)
        {
            const unsigned plrpos = found_dead[a];
            RegisterMultiDead<Depth>(plrpos, blockposlist);
            ++ndead;
        }
        
        for(unsigned route=0; route<MaxDim; ++route)
        {
            if(verdict[route] != dead) continue;
            for(unsigned plrpos=0; plrpos<MaxDim; ++plrpos)
                if(routes[plrpos] == route)
                {
                    RegisterMultiIgnore<Depth>(plrpos, blockposlist);
                    ++nignore;
                    //SmallCompleter(level).MarkEverythingIgnorable<Depth> ();
                }
        }
    }
}

double DeadZones::MultiDeadCalculatePosition
    (const BigInt& total,
     const unsigned char digits[],
     unsigned Depth) const
{
    BigInt position;
    position=0;
    unsigned Max=UndeadMax;//-Depth;
    for(unsigned Last=0, nbox=0; nbox<Depth; ++nbox)
    {
        unsigned Coord = digits[nbox]+nbox;
        position += Binom::Choose(Max - Last,  Depth-nbox);
        position -= Binom::Choose(Max - Coord, Depth-nbox);
        Last=Coord+1;
    }
    return position.GetDouble() / total.GetDouble();
}

static const double GetTime()
{
    struct timeval tv;
    gettimeofday(&tv,NULL);
    return tv.tv_sec + tv.tv_usec*1e-6;
}

static const std::string FormatTime(double d)
{
    char Buf[64];
    std::sprintf(Buf, "%u:%02u:%02.0f",
        (unsigned)(d/3600),
        ((unsigned)(d/60))%60,
        std::fmod(d,60.0)
                );
    return Buf;
}

static const char dotchars[2] = {'\'', '.'};

template<unsigned Depth>
void DeadZones::DeadMultiBlockBuilderLoop(unsigned char digits[Depth])
{
    std::cerr << "Collecting Dead" << Depth << " data...\n";

    BigInt total;
    total.Choose(UndeadMax, Depth);
    
    const unsigned maxdigit = UndeadMax-Depth;
    unsigned statuscounter=1+1; /* +1 so that timers won't display NAN */
    unsigned dotcounter=1;
    
    unsigned dumped_lines = MaxHeight;
    unsigned dumped_width = MaxWidth+4;
    
    unsigned interval = 5000 / Depth;
    
    for(unsigned d=0; d<Depth; ++d)
        if(digits[d] >= maxdigit) /* Matches also 0xFF set by LoadDeadData() */
        {
            return;
        }
    
    const double begin_pos  = MultiDeadCalculatePosition(total, digits, Depth);
    const double begin_time = GetTime();
    
    unsigned dotcharcounter=0;
    
    unsigned ndependdead = std::count_if(MultiDeadFlags.WithPlayer.begin(),
                                         MultiDeadFlags.WithPlayer.end(),
                                         DeadCount<Depth>);

    unsigned nalwaysdead = std::count_if(MultiDeadFlags.Set.begin(),
                                         MultiDeadFlags.Set.end(),
                                         DeadCountBlockSet<Depth>);

    unsigned nignore = std::count_if(MultiIgnoreFlags.WithPlayer.begin(),
                                     MultiIgnoreFlags.WithPlayer.end(),
                                     DeadCount<Depth>);
    
    for(;;)
    {
        if(TerminateSignalled)
        {
            break;
        }
        
        unsigned blockposlist[Depth];
    
        for(unsigned d=0; d<Depth; ++d)
            blockposlist[d] = UndeadList[digits[d]+d];
        
        if(!--dotcounter)
        {
            unsigned width = 65-dumped_width;
            dotcounter=interval/width;
            std::cerr << dotchars[dotcharcounter] << std::flush;
        }
        
        if(!--statuscounter)
        {
            statuscounter = interval;
            
            dotcharcounter = (dotcharcounter+1) % sizeof(dotchars);
            
            const double position = MultiDeadCalculatePosition(total, digits, Depth);
            const double timepos  = GetTime();
            
            const double TimeUsed = (timepos-begin_time);
            const double Speed = TimeUsed / (position-begin_pos); // seconds per completion
            const double EstimatedSofar = Speed * position;
            const double EstimatedTotal = 1.0 * Speed;
            const double EstimatedLeft = (1.0 - position) * Speed;
            
            std::cerr << "]  \r" << std::flush;
            
            LevelState level;
            level.SetNBlockSituation<Depth>(0, blockposlist);
            level.Dump();
            
            #define YPosition(y) std::cerr << "[" << y << "A" << std::flush;
            #define XPosition() std::cerr << "\r[" << dumped_width << "C" << std::flush;
            
            YPosition(dumped_lines);
            XPosition();

            std::cerr << "\n";
            
            XPosition();
            std::cerr << "Currently testing combination: [";
            for(unsigned b=0; b<Depth; ++b)
            {
                if(b)std::cerr << ':';
                std::cerr << (unsigned)(digits[b]+b);
                //std::cerr << '-'<<blockposlist[b];
            }
            std::cerr << '/' << maxdigit << "]    \n";
            XPosition();
            
            gmp_fprintf(stderr,
                "Dead%u progress: %.2f%% of %Zu configurations\n",
                Depth, position*100.0, total.GetMpz());
            std::fflush(stderr);
            
            XPosition();
            std::cerr << "Stats: always=" << nalwaysdead  << ", cond=" << ndependdead
                                          << ", ignore=" << nignore << "\n";
            
            XPosition();
            std::cerr << "Time used: " << FormatTime(TimeUsed) << std::endl;
            
            XPosition();
            std::cerr << "Est."
                         " so far " << FormatTime(EstimatedSofar)
                      << ", left " << FormatTime(EstimatedLeft)
                      << ", total " << FormatTime(EstimatedTotal)
                      << "    \n";
            YPosition(6);
            XPosition();
            
            std::cerr << "[";
            
            #undef YPosition
            #undef XPosition
        }
        
        DeadMultiBlockBuilder<Depth> (blockposlist, 
                                      ndependdead,
                                      nalwaysdead,
                                      nignore);
        
        unsigned d=Depth;
        while(d-- > 0) if(++digits[d] < maxdigit) break;
        
        if(d == (unsigned)-1) break;
        
        while(++d < Depth) digits[d] = digits[d-1];
    }
    std::cerr << "\33[" << dumped_lines << "B\r\n" << std::flush;
}

template<unsigned Depth, typename container>
static void MultiDeadCombinationHelperBase(BigInt& amount, const container& m)
{
    BigInt tmp;
    tmp.Choose(CompactBlockCoordinateLimit, ActualMaxObject-1-Depth);
    amount -= tmp * (unsigned)std::count_if(m.begin(), m.end(), DeadCount<Depth>);
}
template<unsigned Depth>
struct MultiDeadCombinationHelper
{
    template<typename container>
    static void Subtract(BigInt& amount, const container& m)
    {
        MultiDeadCombinationHelperBase<Depth>(amount, m);
        MultiDeadCombinationHelper<Depth+1>::Subtract(amount, m);
    }
};
template<>
struct MultiDeadCombinationHelper<USE_MULTI_DEAD>
{
    template<typename container>
    static void Subtract(BigInt& amount, const container& m)
    {
        MultiDeadCombinationHelperBase<USE_MULTI_DEAD>(amount, m);
    }
};


void DeadZones::CountCombinations()
{
    std::cout << std::flush;
    std::cerr << std::flush;
    
    /* Calculate the number of different possible box configurations */
    BigInt tmp;
    tmp.Choose(CompactBlockCoordinateLimit, ActualMaxObject-1);
    
#if CHOOSE_LEVEL == 60
    gmp_fprintf(stderr,
        "// Number of possible different block combinations: %Zu\n",
        tmp.GetMpz());
#endif
    gmp_fprintf(stdout,
        "// Number of possible different block combinations: %Zu\n",
        tmp.GetMpz());
    
    /* Multiply by number of different player positions */
    //tmp *= (CompactPlayerCoordinateLimit - (ActualMaxObject-1));
    tmp *= 1+4*(ActualMaxObject-1);
    
#if CHOOSE_LEVEL == 60
    gmp_fprintf(stderr, "// When including the player, the number is %Zu.\n", tmp.GetMpz());
#endif
    gmp_fprintf(stdout, "// When including the player, the number is %Zu.\n", tmp.GetMpz());
    
#if USE_MULTI_DEAD >= 1
    DeadBlkSetVec& m = MultiDeadFlags.WithPlayer;
    MultiDeadCombinationHelper<1>::Subtract(tmp, m);
#endif

#if CHOOSE_LEVEL == 60
    gmp_fprintf(stderr, "// Excluding the multidead, it is %Zu.\n", tmp.GetMpz());
#endif
    gmp_fprintf(stdout, "// Excluding the multidead, it is %Zu.\n", tmp.GetMpz());

    std::fflush(stderr);
    std::fflush(stdout);
    
    std::cerr << std::endl;
}

void DeadZones::FindMargins(void) const
{
    unsigned nobjects=1; //player
    for(unsigned pos=0; pos<19*16; ++pos)
        if(maps[levelno].mapdata[pos]&4)
            ++nobjects;

    unsigned topmargin=0;
    for(unsigned y=0; y<16; ++y,++topmargin)
        for(unsigned x=0; x<19; ++x)
            if(!(maps[levelno].mapdata[y*19+x]&5))goto d1;
  d1:
    unsigned bottommargin=0;
    for(unsigned y=16; y-->0; ++bottommargin)
        for(unsigned x=0; x<19; ++x)
            if(!(maps[levelno].mapdata[y*19+x]&5))goto d2;
  d2:
    unsigned leftmargin=0;
    for(unsigned x=0; x<19; ++x,++leftmargin)
        for(unsigned y=0; y<16; ++y)
            if(!(maps[levelno].mapdata[y*19+x]&5))goto d3;
  d3:
    unsigned rightmargin=0;
    for(unsigned x=19; x-->0; ++rightmargin)
        for(unsigned y=0; y<16; ++y)
            if(!(maps[levelno].mapdata[y*19+x]&5))goto d4;
  d4:
    /**/
    if(topmargin>0)--topmargin;
    if(bottommargin>0)--bottommargin;
    if(leftmargin>0)--leftmargin;
    if(rightmargin>0)--rightmargin;
    /**/
    unsigned vert = 16-(bottommargin+topmargin);
    unsigned horiz= 19-(leftmargin+rightmargin);
    
    ActualMaxObject = nobjects;
    
#if CHOOSE_LEVEL == 60
    std::cerr <<
        "#if CHOOSE_LEVEL==" << levelno << "\n"
        "// LEVEL " << (levelno/5+1) << "-" << (levelno%5+1) << " :\n"
        "static const unsigned MaxWidth=" << horiz << ", OffsetWidth=" << leftmargin << ";\n"
        "static const unsigned MaxHeight=" << vert << ", OffsetHeight=" << topmargin << ";\n"
        "static const unsigned MaxObject=" << nobjects << ";\n";
#else
    std::cout << "/* Suggested margins: "
              << horiz << "x" << vert
              << "+" << leftmargin << "+" << topmargin
              << " -- " << nobjects << " objects */"
              << std::endl;
    std::cout << "/* Current margins: "
              << MaxWidth << "x" << MaxHeight
              << "+" << OffsetWidth << "+" << OffsetHeight
              << " -- " << MaxObject << " objects */"
              << std::endl;
    

    if(nobjects != MaxObject
    || horiz > MaxWidth
    || vert > MaxHeight
    || topmargin < OffsetHeight
    || leftmargin < OffsetWidth)
    {
        std::abort();
    }
#endif

    BuildCompactCoordinates();
}

/* CheckPoint is a helper function for CreateDeadFlags1(). */
bool DeadZones::CheckPoint(unsigned x,unsigned y,
                           int wallx,int wally,
                           int dx1,int dy1,
                           int dx2,int dy2)
{
    if(at(x,y) & MapBits::Wall) return true;
    if(at(x,y) & MapBits::Goal) return false;
    
    /* If there is no wall directly in front, it's not dead */
    if(!(at(x+wallx, y+wally) & MapBits::Wall)) return false;
    /* If there is a wall directly on either side, it's dead (corner) */

    if(at(x+dx1, y+dy1) & MapBits::Wall) return true;
    if(at(x+dx2, y+dy2) & MapBits::Wall) return true;
    
    for(unsigned px=x,py=y; ; px+=dx1, py+=dy1)
    {
        char a = at(px,py);
        // goal seen? not dead
        if(a & MapBits::Goal) return false;
        // wall seen? maybe dead
        if(a & MapBits::Wall) break;
        // path seen? not dead            
        if(!(at(px+wallx, py+wally) & MapBits::Wall)) return false;
    }
    for(unsigned px=x+dx2,py=y+dy2; ; px+=dx2, py+=dy2)
    {
        char a = at(px,py);
        // goal seen? not dead
        if(a & MapBits::Goal) return false;
        // wall seen? maybe dead
        if(a & MapBits::Wall) break;
        // path seen? not dead            
        if(!(at(px+wallx, py+wally) & MapBits::Wall)) return false;
    }
    // dead!
    return true;
}

void DeadZones::CreateDeadFlags1()
{
    BuildFloorCoordinates();

    /* First pass: Find corners and walls */
    for(unsigned y=0; y<MaxHeight; ++y)
        for(unsigned x=0; x<MaxWidth; ++x)
            DeadFlags[y*MaxWidth+x]
             = CheckPoint(x,y, 0,-1, -1,0, 1,0)  // wall up,   check sides
            || CheckPoint(x,y, 0, 1, -1,0, 1,0)  // wall down, check sides
            || CheckPoint(x,y, -1,0, 0,-1, 0,1)  // wall left, check up&down
            || CheckPoint(x,y,  1,0, 0,-1, 0,1)  // wall right, check up&down
              ;

    LevelState state;
    state.CreateFromLevel();
    state.RebuildBlockCaches();
    const unsigned playerpos = state.GetPY() * MaxWidth + state.GetPX();
    
#if CHOOSE_LEVEL != 60
    std::cerr << "Collecting Dead1 data...\n";
#endif
    for(unsigned testpos=0; testpos<MaxDim; ++testpos)
    {
        if(at(testpos) & MapBits::Wall) { DeadFlags[testpos] = true; continue; }
        if(at(testpos) & MapBits::Goal) { DeadFlags[testpos] = false; continue; }
        
        LevelState tmp = state;
        
        //std::cerr << "testpos = " << testpos << std::endl;
        try
        {
            /* Find if there's a block that can be pushed into the given position */
            for(unsigned n=1; n<MaxObject; ++n)
            {
                unsigned pos = state.GetBlockPos(n-1);
                if(pos == MaxDim) continue;
                
                //std::printf("n=%u, now player is at %u\n", n, playerpos);
                unsigned blockposlist[1] = {pos};
                tmp.SetNBlockSituation<1>(playerpos, blockposlist);
                //std::cerr << "testpos now = " << testpos << std::endl;
                
                SmallCompleter com(tmp);
                
                unsigned coordlist[1] = {testpos};
                CoordinateType result = com.IsPushableInto<1>(coordlist);
                if(result != NonCoordinate)
                {
                    //std::cerr << n << " works\n";
                    throw result;
                }
            }
            
            /* If a block can't be pushed to that location,
             * mark it dead to save in the coordinate space.
             */
            DeadFlags[testpos] = true;
        }
        catch(CoordinateType testplayerpos)
        {
            /* Now try if the block in question can be shoved into one of the
             * goals when the player is on the given position
             */
            //std::cerr << "Trying " << (unsigned)testplayerpos << " at " << testpos << "\n";
            unsigned blockposlist[1] = {testpos};
            tmp.SetNBlockSituation<1>(testplayerpos, blockposlist);
            //tmp.Dump();
            if(!SmallCompleter(tmp).IsCompleteable())
            {
                //std::cerr << "-> Dead\n";
                DeadFlags[testpos] = true;
            }
        }
    }

#if 0
    std::string dead = "[";
    for(unsigned y=0; y<MaxHeight; ++y)
        for(unsigned x=0; x<MaxWidth; ++x)
        {
            if(x||y) dead += ',';
            dead += ('0' + DeadFlags[y*MaxWidth+x]);
        }
    dead += ']';
    std::cout << dead.c_str() << std::endl;
#endif
    
    if(levelno == 53)
    {
        /* WARP'S IDEA */
        static const CoordinateType deads[] =
        {
            163,182,201,220,239,149
        };
        static const unsigned nextradead = sizeof(deads)/sizeof(*deads);
        for(unsigned n=0; n<nextradead; ++n)
        {
            unsigned c = deads[n];
            unsigned x = c%19 + 0;
            unsigned y = c/19 + 1;
            DeadFlags[(y-OffsetHeight)*MaxWidth+(x-OffsetWidth)] = true;
        }
    }

    unsigned ndead=0;
    for(unsigned pos=0; pos<MaxDim; ++pos)
        if(DeadFlags[pos]) ++ndead;
    
    std::cout << "// " << (unsigned)MaxDim << " points, "
                       << (MaxDim-ndead) << " not dead\n";

    UndeadMax=0;
    for(unsigned pos=0; pos<MaxDim; ++pos)
    {
        if(CheckDead(pos)) continue;
        UndeadList[UndeadMax++] = pos;
    }
    
    std::cout << "// UndeadMax=" << UndeadMax
              << ", CompactBlockCoordinateLimit=" << CompactBlockCoordinateLimit
              << "\n";

    std::cout << std::flush;
}

template<unsigned Depth>
void DeadZones::CreateMultiDeadFlags ()
{
    unsigned char digits[Depth] = {0};
    LoadDeadData<Depth>(digits);
    DeadMultiBlockBuilderLoop<Depth> (digits);
    SaveDeadData<Depth>(digits);
    
    unsigned ndependdead = std::count_if(MultiDeadFlags.WithPlayer.begin(),
                                         MultiDeadFlags.WithPlayer.end(),
                                         DeadCount<Depth>);

    unsigned nalwaysdead = std::count_if(MultiDeadFlags.Set.begin(),
                                         MultiDeadFlags.Set.end(),
                                         DeadCountBlockSet<Depth>);

    unsigned nignore = std::count_if(MultiIgnoreFlags.WithPlayer.begin(),
                                     MultiIgnoreFlags.WithPlayer.end(),
                                     DeadCount<Depth>);

    /* Calculate the number of different possible box configurations */
    BigInt total;
    total.Choose(CompactBlockCoordinateLimit, ActualMaxObject-1);
    
    BigInt nothercombinations;
    nothercombinations.Choose(CompactBlockCoordinateLimit-Depth, ActualMaxObject-1-Depth);

    const BigInt nalways_kill = nothercombinations * nalwaysdead;

    char nalways_buf[64];
    std::sprintf(nalways_buf, "%.1f",
       nalways_kill.GetDouble() * 100.0 / total.GetDouble()
                );

    BigInt ndepend_kill = nothercombinations * ndependdead;

    total *= (ActualMaxObject-1)*4; // approximate number of different player positions.
    
    char ndepend_buf[64];
    std::sprintf(ndepend_buf, "%.1f",
       ndepend_kill.GetDouble() * 100.0 / total.GetDouble()
                );

    std::cout << "// Dead" << Depth << " stats:\n";
    std::cout << "//    " << nalwaysdead << " always dead (coverage = " << nalways_buf << "%)\n";
    std::cout << "//    " << ndependdead << " conditional dead (coverage >= " << ndepend_buf << "%)\n";
    std::cout << "//    " << nignore << " to be ignored in consequent passes.\n" << std::flush;
}

template<unsigned Depth>
struct MultiDeadMakingHelper
{
    static void Do()
    {
        DeadZones.CreateMultiDeadFlags<Depth>();
        MultiDeadMakingHelper<Depth+1>::Do();
    }
};
template<>
struct MultiDeadMakingHelper<USE_MULTI_DEAD>
{
    static void Do()
    {
        DeadZones.CreateMultiDeadFlags<USE_MULTI_DEAD>();
    }
};

void DeadZones::CreateAllMultiDeadFlags()
{
#if USE_MULTI_DEAD >= 1
    MultiDeadMakingHelper<1>::Do();
#endif
}

static bool CheckMultiBlockDeadInBlockSet(const DeadZones::BlockSet& com)
{
    const std::bitset<FloorAmount>& cur = CurrentBlockMap;
    
  #if 0
    if((cur&com)!=com) return false;
  #else
    enum { wbits = std::numeric_limits<unsigned long>::digits };
    enum { max   = (FloorAmount+wbits-1)/wbits };
    /* Access to std::bitset's private data. */
    const unsigned long* p1 = (const unsigned long*)&cur;
    const unsigned long* p2 = (const unsigned long*)&com;
    for(unsigned a=0; a<max; ++a)
    {
        if( (p1[a] & p2[a]) != p2[a]) return false;
    }
  #endif
    ++MultiDeadCounter[com.count()];
    return true;
}

static bool CheckMultiBlockDeadIn(const DeadZones::BlockSetVec& container)
{
    return
        std::find_if(container.begin(), container.end(), CheckMultiBlockDeadInBlockSet)
            != container.end();
}

static bool CheckMultiBlockDeadIn(unsigned plrpos,
                                  const DeadZones::DeadBlkSetVec& container)
{
    for(DeadZones::DeadBlkSetVec::const_iterator
        i = std::lower_bound(container.begin(), container.end(),
                             plrpos, DeadFind);
        i != container.end() && i->plrpos == plrpos;
        ++i)
    {
        if(CheckMultiBlockDeadInBlockSet(i->blockmap)) return true;
    }
    return false;
}

bool DeadZones::CheckMultiBlockDead(unsigned plrpos) const
{
    return
        CheckMultiBlockDeadIn(MultiDeadFlags.Set) ||
        CheckMultiBlockDeadIn(plrpos, MultiDeadFlags.WithPlayer);
}

bool DeadZones::CheckMultiBlockIgnore(unsigned plrpos) const
{
    return
        CheckMultiBlockDeadIn(plrpos, MultiIgnoreFlags.WithPlayer);
}

bool DeadZones::CheckMultiBlockDead(unsigned plrpos, unsigned lastblockpos) const
{
    std::map<CoordinateType, DeadFlagData>::const_iterator
        i = MultiDeadFlagsByBlock.find(lastblockpos);
    
    if(i == MultiDeadFlagsByBlock.end()) return false;
    
    return
        CheckMultiBlockDeadIn(i->second.Set) || 
        CheckMultiBlockDeadIn(plrpos, i->second.WithPlayer);
}

bool DeadZones::CheckMultiBlockDeadBlock() const
{
    return CheckMultiBlockDeadIn(MultiDeadFlags.Set);
}

bool DeadZones::CheckMultiBlockDeadBlock(unsigned lastblockpos) const
{
    std::map<CoordinateType, DeadFlagData>::const_iterator
        i = MultiDeadFlagsByBlock.find(lastblockpos);
    
    if(i == MultiDeadFlagsByBlock.end()) return false;
    
    return CheckMultiBlockDeadIn(i->second.Set);
}

bool DeadZones::CheckMultiBlockIgnore(unsigned plrpos, unsigned lastblockpos) const
{
    std::map<CoordinateType, DeadFlagData>::const_iterator
        i = MultiIgnoreFlagsByBlock.find(lastblockpos);
    
    if(i == MultiIgnoreFlagsByBlock.end()) return false;
    return CheckMultiBlockDeadIn(plrpos, i->second.WithPlayer);
}

void DeadZones::BuildFromLevel()
{
    CreateDeadFlags1();

    FindMargins();
    
    CreateAllMultiDeadFlags();
    
    CountCombinations();
#if CHOOSE_LEVEL != 60
    Dump();
#endif
}

void DeadZones::Dump() const
{
    for(unsigned x=0,pos=0; pos<MaxDim; ++pos)
    {
        switch(at(pos)&3)
        {
            case MapBits::Empty:
                std::cerr << (CheckDead(pos) ? 'D' : ' ');
                break;
            case MapBits::Wall:
                std::cerr << '#';
                break;
            case MapBits::Goal:
                std::cerr << '*';
                break;
        }
        if(++x==MaxWidth){std::cerr << "\n";x=0;}
    }
}

class DeadZones DeadZones;
