#include "NewRadixSet.hh"

#include <algorithm>
#include <vector>

class SmallCompleter
{
public:
    SmallCompleter(const LevelState& lev)
    {
        pino.reserve(65536);
        pino.push_back(lev);
    }
private:
    struct CheckFinished
    {
        CheckFinished() : result(false) { }
        
        void operator() (const LevelState& state)
        {
            if(state.IsCompleted())
            {
                result=true;
                throw 0;
            }
        }
        
        bool result;
    };
public:
    bool IsCompleteable()
    {
        typedef CheckFinished CheckType;
        CheckType check;
        checkfunptr = (void*)&check;
        
        try { TryMoves<CheckType> (); } catch(int) { }
        
        return check.result;
    }
private:
    template<unsigned BlockCount>
    struct CheckPushable
    {
        CheckPushable(unsigned poslist[BlockCount]) : result(NonCoordinate)
        {
            for(unsigned a=0; a<BlockCount; ++a)
                targetpos[a] = poslist[a];
            std::sort(targetpos, targetpos+BlockCount);
        }
        
        void operator() (const LevelState& state)
        {
            for(unsigned a=0; a<BlockCount; ++a)
                if(state.GetBlockPos(a) != targetpos[a])
                    return;
            result=state.GetPlayerPos();
            throw 0;
        }
        
        unsigned targetpos[BlockCount];
        CoordinateType result;
    };
public:
    template<unsigned BlockCount>
    CoordinateType IsPushableInto(unsigned targetposlist[BlockCount])
    {
        typedef CheckPushable<BlockCount> CheckType;
        CheckType check(targetposlist);
        checkfunptr = (void*)&check;
    
        try { TryMoves<CheckType>(); } catch(int) { }
        
        return check.result;
    }
private:
    template<unsigned Depth>
    struct IgnoreMarker
    {
        void operator() (const LevelState& state)
        {
            unsigned plrpos = state.GetPlayerPos();
            unsigned blockposlist[Depth];
            for(unsigned a=0; a<Depth; ++a)
                blockposlist[a] = state.GetBlockPos(a);

            DeadZones.RegisterMultiIgnore<Depth> (plrpos, blockposlist);
        }
    };
public:
    template<unsigned Depth>
    void MarkEverythingIgnorable()
    {
        typedef IgnoreMarker<Depth> CheckType;
        CheckType check;
        checkfunptr = (void*)&check;
        TryMoves<CheckType>();
    }

private:
    template<typename FuncType>
    void TryMoves()
    {
        while(!pino.empty())
        {
            const LevelState s = pino.back();
            pino.pop_back();
            s.RebuildBlockCaches();
            TryState<FuncType> (s);
        }
    }
    
    template<typename FuncType>
    void TryState(const LevelState& state)
    {
        typedef FavorablePointList targetlist;
        
        /* checkfunptr points to an object of type "FuncType".
         * Cast the pointer into the right type, and call the
         * provided function.
         *
         * checkfunptr is a member to SmallCompleter to avoid
         * having to pass it as a parameter to each recursion cycle.
         */
        ((FuncType*)(checkfunptr))->operator() (state);
        
        targetlist targets;
        state.FindWalkingDestinations(targets);
        
        for(targetlist::const_iterator i = targets.begin(); i != targets.end(); ++i)
        {
            const CoordinateType Point = *i;
            TryDir<FuncType>(state, Point, -1);
            TryDir<FuncType>(state, Point, 1);
            TryDir<FuncType>(state, Point, -MaxWidth);
            TryDir<FuncType>(state, Point, MaxWidth);
        }
    }
    
    template<typename FuncType>
    void TryDir(const LevelState& state, const CoordinateType Point, int offset) /*REGPARM*/
    {
        if(state.CanPushFrom(Point, offset))
        {
            LevelState NewState = state;
            NewState.PushFromNoMove(Point, offset);
            if(!NewState.CheckBlunder(Point + offset*2, offset))
            {
                NewState.WalkTo(Point+offset);
                if(seenmap.add(NewState.GetData()))
                {
                    pino.push_back(NewState);
                }
            }
            state.RebuildBlockCaches();
        }
    }

    /* Smallcompleter MUST USE NON-COMPACTED coordinates, because
     * non-compacted coordinates are only built after DeadZones
     * are calculated, and Smallcompleter is used during DeadZones
     * calculation.
     *
     * However, floor coordinates must have been calculated.
     */    
private:
    std::vector<LevelState> pino;
    void* checkfunptr;
    
    typedef NewRadixSet<CoordinateType,MaxDim,MaxObject> seenmaptype;
    seenmaptype seenmap;
};
