#include "NewRadixMap.hh"

#undef USE_STANDARD_DEQUE

#ifdef USE_STANDARD_DEQUE
# include <deque>
#else
# include "deque.hh"
#endif

#include "history.hh"
#include "timings.hh"

class Completer
{
private:
    struct Active
    {
        CompressedLevel lev;
        History hist;

        Active(const CompressedLevel& l, const History& h) : lev(l), hist(h) { }
    } __attribute__((packed));

    typedef NewRadixMapGMP<MaxCompactCoordinateData, unsigned short,
                           CompactCoordinateBitsPerUnit> experience_type;
  #ifdef USE_STANDARD_DEQUE
    typedef std::deque<Active> active_list_type;
  #else
    typedef Deque<Active> active_list_type;
  #endif

public:
    Completer(LevelState& lev)
    {
        std::cout << "// Sizeof(Active) = " << sizeof(Active) << std::endl;
        
        active_list.clear();
        
        //experience.clear();
        best.Moves.push_back(Move(Move::WalkToPush,1,1));
        
        History tmp;
        Activate(lev, tmp, 0);
        
        bestduration = 0xFFFF;
        
        LevelState state = lev;
    }
    
    void Activate(const LevelState& lev, History& hist, const unsigned short time)
    {
        hist.time = time;
        Active act(lev.Compress(), hist);
        
    #if 1
        // for Deque or std::deque, without sorting
        active_list.push_back(act);
    #else
     #if 0
        // for Deque, sorting
        const size_t begin  = 0;
        const size_t end    = active_list.size();
        const size_t middle = end/2;
        
        size_t smaller=begin, guess=end;
        /*
        size_t smaller = 0;
        size_t guess   = 0;
        for(size_t jump=1; guess<middle; jump*=2)
        {
            unsigned short act_time = active_list[guess].hist.time;
            if(act_time <= time) smaller=guess;
            if(act_time >= time) break;
            guess += jump;
        }
        if(guess >= middle) guess = end;
        */
        typedef size_t it;

        it i_begin = smaller;
        it i_end   = guess;
     #else
        // for std::deque, sorting
        typedef active_list_type::iterator it;
        it i_begin = active_list.begin();
        it i_end   = active_list.end();
     #endif
        
        size_t len = i_end - i_begin;
        while(len > 0)
        {
            size_t half = len >> 1;
            it i_middle   = i_begin + half;
         #if 0
            unsigned short act_time = active_list[i_middle].hist.time;
         #else
            unsigned short act_time = i_middle->hist.time;
         #endif
            if(act_time < time)
            {
                i_begin = i_middle;
                ++i_begin;
                len = len - half - 1;
            }
            else
                len = half;
        }
        active_list.insert(i_begin, act);
    #endif
        experience.ensure(lev.GetCompactData(), time);
    }
    
    bool Iteration()
    {
        unsigned ignored=0;
        
        DeadCounter0=0;
        std::fill_n(MultiDeadCounter,USE_MULTI_DEAD+1,0);
        
        TimeIgnored=0;
        NTargets=0;
        
        /*
          Front-side handling vs back-side handling.
          
          Front-side handling:
            Apa. bredth-first
            Benefits:
              Eliminates useless paths as early as possible
            Harms:
              Long offsets in the Moves structure, no compression
              The active lists always grow, they are never shortened
          
          Back-side handling:
            Aka. depth-first
            Benefits:
              The Moves structure compresses really well (small offsets)
              The active lists are kept really short
              Most moves actually are blunders or lead to a timeout
            Harms:
              Nothing _useful_ seems to ever happen
        
          Apparently, back-side handling is a lot slower
          than front-side handling. Front-side handling
          leads faster to the cancellation of useless paths.
        */
        ssize_t balance = active_list.size();
        for(unsigned counter=0; counter<20000; ++counter)
        {
            if(TerminateSignalled) break;
            if(!active_list.empty())
            {
#ifdef USE_STANDARD_DEQUE
                // for std::deque
                const Active act = *active_list.begin();
                active_list.pop_front();
#else
                // for Deque
                const Active act = active_list.pop_front();
#endif
                
                LevelState level;
                level.BuildFrom(act.lev);
                level.RebuildBlockCaches();
                
                unsigned short time;
                if(!experience.find(level.GetCompactData(), time))
                {
                    std::cerr << "No experience for\n";
                    level.Dump();
                    abort();
                }
                if(act.hist.time == time) { TryMoves(level, act.hist); } else ++ignored;
            }
            if(active_list.empty())
            {
                break;
            }
        }
        
        size_t act = active_list.size();
        balance = act - balance;
        
        std::cout << "// A:" << act
                  << ", S:" << experience.size()
                  << ", ign " << ignored
                  << ", Tign " << TimeIgnored
                  << ", targ " << NTargets
                  << ", bal " << balance
                  << ", dead=[";
        for(unsigned n=1; n<=USE_MULTI_DEAD; ++n)
            std::cout << "," << MultiDeadCounter[n];
        std::cout << ",(" << DeadCounter0 << ")]"
                  << std::endl;
    #if 0
        active_list.Report(std::cout);
    #endif
    #if 0
        MoveContainer.Report(std::cout);
    #endif
        std::cout << std::flush;

        return act==0;
    }
    void DumpBest() const
    {
        if(bestduration == 0xFFFF)
            std::cout << "/* NO SOLUTION FOUND */\n";
        else
            best.Dump();
    }
    const std::string GetJs() const
    {
        if(bestduration == 0xFFFF) return "";
        return best.GetJs();
    }
private:
    void TryMoves(const LevelState& state, const History& hist)
    {
        if(hist.time >= MaxDuration)
        {
            ++TimeIgnored;
            return;
        }
        
        typedef FavorablePointListWithCost targetlist;
        targetlist targets;
        
        ssize_t CurPoint = state.GetPlayerPos();

        state.FindWalkingDestinations(targets);
        
        NTargets += targets.size();
        
        const bool WasGrabbing = hist.WasGrabbing();

        for(targetlist::const_iterator i = targets.begin(); i != targets.end(); ++i)
        {
            const ssize_t Point      = i->Point;
            const unsigned short Dura = hist.time + IntWalkingSpeed * i->Cost;
            
            if(Dura > MaxDuration) continue;
            
            const bool Walk = Point != CurPoint;            
            int PushCost = IntPushingSpeed;
            if(Walk || (hist.Started() && !WasGrabbing)) PushCost += IntGrabbingTime;
            int KickCost = 0;
            if(Walk || hist.Started()) KickCost = IntKickingTime;
            
            if(i->Dir & 4) TryDir(state,hist, -1,        Point, Dura, PushCost, KickCost);
            if(i->Dir & 8) TryDir(state,hist, 1,         Point, Dura, PushCost, KickCost);
            if(i->Dir & 1) TryDir(state,hist, -MaxWidth, Point, Dura, PushCost, KickCost);
            if(i->Dir & 2) TryDir(state,hist, MaxWidth,  Point, Dura, PushCost, KickCost);
        }
    }
    
    void TryDir(const LevelState& state,
                const History& hist,
                int offset,
                const ssize_t Point,
                const unsigned short Dura,
                const int PushCost,
                const int KickCost) REGPARM
    {
        ssize_t KickLength = state.CalculateKickFrom(Point, offset);
        if(KickLength > 0)
        {
            if(unlikely(KickLength > 90000))
            {
                throw "Eh, what is this KickLength";
            }
            LevelState NewState = state;
            NewState.PushFromNoMove(Point, offset);
            if(!NewState.CheckBlunder(Point+offset*2,offset))
            {
                unsigned short NewDura = Dura;
               
                NewDura += PushCost;
                if(NewDura <= MaxDuration)
                {
                    NewState.WalkTo(Point+offset);
                    TryMove(NewState,hist, Point, offset, NewDura, 0);
                }else ++TimeIgnored;
               
                NewDura += KickCost-PushCost;
               
                if(NewDura <= MaxDuration)
                {
                    state.RebuildBlockCaches();
                    NewState = state;
                    NewState.KickFrom(Point, offset, KickLength);
                    if(!NewState.CheckBlunder(Point + offset*(KickLength+1),offset))
                    {
                      NewState.WalkTo(Point);
                      TryMove(NewState,hist, Point, offset, NewDura, KickLength);
                    }
                }
                else
                    ++TimeIgnored;
            }
            state.RebuildBlockCaches();
        }
    }
    
    void TryMove(const LevelState& NewState, const History& hist,
                 ssize_t NewPos,
                 int PushDir,
                 unsigned NewDura,
                 unsigned KickLength /* 0 = push */
                ) REGPARM
    {
        if(NewState.IsCompleted())
        {
            if(KickLength > 0)
            {
                NewDura += RoundUp(KickingSpeed * KickLength);
            }
            if(NewDura < bestduration)
            {
                Move move(KickLength ? Move::WalkToKick : Move::WalkToPush,
                          NewPos, PushDir);
#ifdef TIMESTAMP_MOVES
                move.timestamp = hist.time;
#endif
            
                History result(hist, move);
                
                best         = result;
                bestduration = NewDura;

                std::cout << "// Completed in " << NewDura
                          << "!\n"
                          << std::flush;
                hist.Dump();
                
                /*if(NewDura < MaxDuration) MaxDuration = NewDura;*/
            }
            else
            {
                std::cout << "// Completed in " << NewDura << "...\n";
            }
            return;
        }
        
        unsigned short OldDura = 0;
        if(!experience.find(NewState.GetCompactData(), OldDura)
        || OldDura > NewDura)
        {
            Move move(KickLength ? Move::WalkToKick : Move::WalkToPush,
                      NewPos, PushDir);
#ifdef TIMESTAMP_MOVES
            move.timestamp = hist.time;
#endif
            
            History NewHist(hist, move);
            Activate(NewState,NewHist, NewDura);
        }
    }
private:
    experience_type experience;

    History best;
    size_t NTargets;
    unsigned TimeIgnored;
    unsigned bestduration;
    
    static active_list_type active_list;
};
