#ifndef __STDC_CONSTANT_MACROS
#define __STDC_CONSTANT_MACROS /* for UINT16_C etc */
#endif

#include "filewindow.hh"

#include "deadzones.hh"
#include "levelstate.hh"
#include "smallcompleter.hh"
#include "bigint.hh"
#include "binom.hh"

#include <sstream>
#include <string>
#include <cmath> /* For std::fmod */
#include <set>

#include <sys/time.h>

//#include "deadcompleter.hh"

static inline bool DeadFind(const DeadZones::DeadBlkSet& a, ssize_t plrpos)
{
    return (ssize_t)a.plrpos < plrpos;
}
static inline bool DeadCompBlockSet(const DeadZones::BlockSet& a, const DeadZones::BlockSet& b)
{
    size_t ac = a.count();
    size_t bc = b.count();
    /*if(ac != bc) */return ac < bc;
    /*return a.blockmap < b.blockmap;*/
}
static inline bool DeadComp(const DeadZones::DeadBlkSet& a, const DeadZones::DeadBlkSet& b)
{
    if(a.plrpos != b.plrpos) return a.plrpos < b.plrpos;
    return DeadCompBlockSet(a.blockmap, b.blockmap);
}

template<size_t b>
static inline bool DeadCount(const DeadZones::DeadBlkSet& a)
{
    return a.blockmap.count() == b;
}

template<size_t b>
static inline bool DeadCountBlockSet(const DeadZones::BlockSet& a)
{
    return a.count() == b;
}

static const std::string GetDeadDataFileName(size_t Depth)
{
    std::stringstream s;
    s << "dead-" << Depth << "-" << levelno << ".dat";
    return s.str();
}

template<size_t Depth>
void DeadZones::RegisterMultiDead(size_t plrpos, const ssize_t blockposlist[Depth],
                                  bool Sort)
{
    DeadBlkSet blks;
    blks.plrpos = plrpos;
    for(size_t a=0; a<Depth; ++a)
        blks.blockmap.set(FloorCoordinateTranslation[blockposlist[a]]);
    
    if(Sort)
    {
        DeadBlkSetVec::iterator j = std::upper_bound
           (MultiDeadFlags.WithPlayer.begin(),
            MultiDeadFlags.WithPlayer.end(), blks, DeadComp);
        MultiDeadFlags.WithPlayer.insert(j, blks);
    }
    else
        MultiDeadFlags.WithPlayer.push_back(blks);
    
    for(size_t a=0; a<Depth; ++a)
    {
        size_t bpos = blockposlist[a];
        
        DeadFlagData& target = MultiDeadFlagsByBlock[bpos];
        
        if(Sort)
        {
            DeadBlkSetVec::iterator j = std::upper_bound
               (target.WithPlayer.begin(), target.WithPlayer.end(), blks, DeadComp);
            target.WithPlayer.insert(j, blks);
        }
        else
            target.WithPlayer.push_back(blks);
    }
    
    /* Don't write to .Set here. .Set is only for combinations
     * that are blunder regardless of player's position.
     */
}

template<size_t Depth>
void DeadZones::RegisterMultiIgnore(size_t plrpos, const ssize_t blockposlist[Depth],
                                    bool Sort)
{
    /* Almost the same as RegisterMultiDead. Instead, write "Ignore" data */
    DeadBlkSet blks;
    blks.plrpos = plrpos;
    for(size_t a=0; a<Depth; ++a)
        blks.blockmap.set(FloorCoordinateTranslation[blockposlist[a]]);
    
    if(Sort)
    {
        DeadBlkSetVec::iterator j = std::upper_bound
           (MultiIgnoreFlags.WithPlayer.begin(),
            MultiIgnoreFlags.WithPlayer.end(), blks, DeadComp);
        MultiIgnoreFlags.WithPlayer.insert(j, blks);
    }
    else
        MultiIgnoreFlags.WithPlayer.push_back(blks);
    
    for(size_t a=0; a<Depth; ++a)
    {
        size_t bpos = blockposlist[a];
        
        DeadFlagData& target = MultiIgnoreFlagsByBlock[bpos];
        if(Sort)
        {
            DeadBlkSetVec::iterator j = std::upper_bound
               (target.WithPlayer.begin(), target.WithPlayer.end(), blks, DeadComp);
            target.WithPlayer.insert(j, blks);
        }
        else
            target.WithPlayer.push_back(blks);
    }
}

template<size_t Depth>
void DeadZones::RegisterAlwaysMultiDead(const ssize_t blockposlist[Depth])
{
    BlockSet blks;
    for(size_t a=0; a<Depth; ++a)
    {
        size_t blockpos = blockposlist[a];
        if(blockpos > (size_t)MaxDim) std::cerr << "// eh, " << blockpos << "?\n";
        
        size_t trans = FloorCoordinateTranslation[blockpos];
        blks.set(trans);
    }

    MultiDeadFlags.Set.push_back(blks);
    
    for(size_t a=0; a<Depth; ++a)
    {
        size_t bpos = blockposlist[a];
        
        DeadFlagData& target = MultiDeadFlagsByBlock[bpos];
        target.Set.push_back(blks);
    }
    
    /* Don't write to .Set here. .Set is only for combinations
     * that are blunder regardless of player's position.
     */
}

#define DigitMagic 0xF11F2EE2

template<size_t Depth>
void DeadZones::SaveDeadData(const unsigned char digits[Depth])
{
    const std::string fn = GetDeadDataFileName(Depth);
    
    FILE* fp = fopen(fn.c_str(), "rb+");
    if(!fp) fp = fopen(fn.c_str(), "wb");

    std::cerr << "// Saving " << fn << "...\n";
    
    rewind(fp);
    putw(DigitMagic, fp);
    fwrite(digits, 1, Depth, fp);
    
    /* Entries:
         -5 -5 x0 y0 x1 y1 x2 y2    = MultiDeadFlags without player position
         px py x0 y0 x1 y1 x2 y2    = MultiDeadFlags with    player position
      -6 px py x0 y0 x1 y1 x2 y2    = MultiIgnoreFlags with  player position
    */
    
    {const BlockSetVec& data = MultiDeadFlags.Set;
    for(size_t a=0; a<data.size(); ++a)
    {
        const BlockSet& blks = data[a];
        if(blks.count() != Depth) continue;
        
        //printf("%s\n", blks.to_string().c_str());
        
        putw(-5,fp);putw(-5,fp);

        size_t ind = blks._Find_first();
        for(size_t b=1; b<=Depth; ++b)
        {
            size_t coord = FloorCoordinateUntranslation[ind];
            
            //printf("%u,%u[%u] ", coord%MaxWidth,coord/MaxWidth,ind);
            
            size_t x = coord % MaxWidth + OffsetWidth;
            size_t y = coord / MaxWidth + OffsetHeight;
            putw(x,fp);putw(y,fp);
            
            ind = blks._Find_next(ind);
        }
        //printf("\n");
    }}

    {const DeadBlkSetVec& data = MultiDeadFlags.WithPlayer;
    for(size_t a=0; a<data.size(); ++a)
    {
        const DeadBlkSet& blks = data[a];
        if(blks.blockmap.count() != Depth) continue;
        
        //printf("%s\n", blks.blockmap.to_string().c_str());
        
        {size_t coord = blks.plrpos;
        size_t x = coord % MaxWidth + OffsetWidth;
        size_t y = coord / MaxWidth + OffsetHeight;
        putw(x,fp);putw(y,fp);}

        size_t ind = blks.blockmap._Find_first();
        for(size_t b=1; b<=Depth; ++b)
        {
            size_t coord = FloorCoordinateUntranslation[ind];
            
            //printf("%u,%u[%u] ", coord%MaxWidth,coord/MaxWidth,ind);
            
            size_t x = coord % MaxWidth + OffsetWidth;
            size_t y = coord / MaxWidth + OffsetHeight;
            putw(x,fp);putw(y,fp);
            
            ind = blks.blockmap._Find_next(ind);
        }
        //printf("\n");
    }}
    
    putw(-6, fp);

    {const DeadBlkSetVec& data = MultiIgnoreFlags.WithPlayer;
    for(size_t a=0; a<data.size(); ++a)
    {
        const DeadBlkSet& blks = data[a];
        if(blks.blockmap.count() != Depth) continue;
        
        //printf("%s\n", blks.blockmap.to_string().c_str());
        
        {size_t coord = blks.plrpos;
        size_t x = coord % MaxWidth + OffsetWidth;
        size_t y = coord / MaxWidth + OffsetHeight;
        putw(x,fp);putw(y,fp);}

        size_t ind = blks.blockmap._Find_first();
        for(size_t b=1; b<=Depth; ++b)
        {
            size_t coord = FloorCoordinateUntranslation[ind];
            
            //printf("%u,%u[%u] ", coord%MaxWidth,coord/MaxWidth,ind);
            
            size_t x = coord % MaxWidth + OffsetWidth;
            size_t y = coord / MaxWidth + OffsetHeight;
            putw(x,fp);putw(y,fp);
            
            ind = blks.blockmap._Find_next(ind);
        }
        //printf("\n");
    }}

    ftruncate(fileno(fp), ftell(fp));
    fclose(fp);
}

template<size_t Depth>
void DeadZones::LoadDeadData(unsigned char digits[Depth])
{
    const std::string fn = GetDeadDataFileName(Depth);
    
    FILE* fp = fopen(fn.c_str(), "rb");
    if(!fp) return;

    std::cerr << "// Loading " << fn << "...\n";
    
    uint_least32_t magic = getw(fp);
    if(magic == DigitMagic)
    {
        fread(digits, 1, Depth, fp);
    }
    else
    {
        fprintf(stderr, "magic wasn't digitmagic, got 0x%X instead\n", magic);
        if((int)magic != EOF) std::fill(digits,digits+Depth,0xFF);
        rewind(fp);
    }
    
    bool ignore=false;
    for(size_t a=0;; ++a)
    {
        int plrpos=0;
        ssize_t Coords[Depth];
        for(size_t b=0; b<=Depth; ++b) // 0 = plrpos or flags
        {
            int x = getw(fp);
            
            if(x == -6)
            {
                ignore=true;
                x = getw(fp);
            }
            
            int y = getw(fp);
            if(x==EOF) goto Done;
            
            //fprintf(stderr, "got %d,%d, ignore=%d\n", x,y, ignore);
            
            if(b == 0 && x == -5)
            {
                plrpos = -5;
                // y is probably -5 as well, but ignore it
            }
            else
            {
                ssize_t coord = (y-OffsetHeight)*MaxWidth + (x-OffsetWidth);
                if(b==0)
                    plrpos = coord;
                else
                {
                    Coords[b-1] = coord;
                }
            }
        }
        if(plrpos < 0)
            RegisterAlwaysMultiDead<Depth>(Coords);
        else if(!ignore)
            RegisterMultiDead<Depth>(plrpos, Coords, false);
        else
            RegisterMultiIgnore<Depth>(plrpos, Coords, false);
    }
Done:
    fclose(fp);
    
    std::sort(MultiDeadFlags.WithPlayer.begin(), MultiDeadFlags.WithPlayer.end(), DeadComp);
    std::sort(MultiIgnoreFlags.WithPlayer.begin(), MultiIgnoreFlags.WithPlayer.end(), DeadComp);
    
    std::map<CoordinateType/*normal block coordinate*/, DeadFlagData>::iterator i;
    for(i=MultiIgnoreFlagsByBlock.begin();
        i!=MultiIgnoreFlagsByBlock.end();
        ++i)
    {
        std::sort(i->second.WithPlayer.begin(), i->second.WithPlayer.end(), DeadComp);
    }
    for(i=MultiDeadFlagsByBlock.begin();
        i!=MultiDeadFlagsByBlock.end();
        ++i)
    {
        std::sort(i->second.WithPlayer.begin(), i->second.WithPlayer.end(), DeadComp);
    }
}

template<size_t Depth>
void DeadZones::DeadMultiBlockBuilder
    (const ssize_t blockposlist[Depth],
     size_t &ndead,
     size_t &nalldead,
     size_t &nignore)
{
    if(MaxObject-1 < Depth) return; // Everything would be dead otherwise.
    
    LevelState level;
    level.SetNBlockSituation<Depth>(0, blockposlist);
    // ^Also assumed to build the Block coordinate data

    const SimpleVec<CoordinateType,MaxDim> routes = level.BuildWalkingMaps();
    /* Now, routes is an array,
     * which contains a value for each point of the room.
     * Points which are reachable by walking are painted
     * the same colour (number). Each disconnected region
     * has its own colour.
     * To test whether path A-B is clear, you only need
     * to check if A and B have the same value in this array.
     */

    if(CheckMultiBlockDeadBlock()) return;

    enum MultiDeadVerdict { unknown='\0', notdead='\1', dead='\2', ignore='\3' };
    MultiDeadVerdict verdict[MaxDim];
    std::memset(&verdict,unknown,sizeof(verdict));
    
    std::set<CoordinateType> playerpositions;
    for(size_t a=0; a<Depth; ++a)
    {
        const ssize_t blockpos = blockposlist[a]; 
        #define TestPlrDir(dir) \
          { \
            size_t plrpos = blockpos + (dir); \
            if(!CurrentObstacleMap.test(plrpos)) \
            { \
                size_t route = routes[plrpos]; \
                if(verdict[route] == unknown \
                && playerpositions.find(plrpos) == playerpositions.end()) \
                { \
                    if(CheckMultiBlockDead(plrpos)/* \
                    || CheckMultiBlockIgnore(plrpos)*/) \
                    { \
                        verdict[route] = ignore; \
                    } \
                    else \
                        playerpositions.insert(plrpos); \
                } \
            } \
          }
        TestPlrDir(-1);
        TestPlrDir( 1);
        TestPlrDir(-MaxWidth);
        TestPlrDir( MaxWidth);
        #undef TestPlrDir
    }
    
    if(playerpositions.empty()) return;

    bool all_are_dead = true;
    
#if 1
    /*
     * Because the combinations are processed from top-left to bottom-right,
     * we can do the following test to reduce the number of calculations.
     *
     * For each block, check if it can be moved to the left or to the up.
     * If it can, check if that new position (which already has been processed)
     * was dead. If it wasn't dead, this one can't be either because a dead
     * state can not become not-dead.
     */
    for(size_t a=0; a<Depth; ++a)
    {
        const ssize_t blockpos = blockposlist[a];
        #define TestPush(dir) \
          { \
            /* dir = direction to push the block towards */ \
            const ssize_t plrpos    = blockpos - (dir); \
            const ssize_t targetpos = blockpos + (dir); \
            const size_t route = routes[plrpos]; \
            if(verdict[route] == unknown) \
            { \
                if(!CurrentObstacleMap.test(plrpos) \
                && !CurrentObstacleMap.test(targetpos)) \
                { \
                    __label__ skipundo; \
                    const ssize_t newplrpos = blockpos; \
                    level.MoveBlockFrom(blockpos, targetpos); /* Move the block */ \
                    if(!level.CheckBlunder(targetpos,dir))/*
                    if(!CheckDead(targetpos) \
                    && !CheckMultiBlockDead(newplrpos))*/ /* If this position was not dead */ \
                    { \
                        verdict[route] = notdead; /* This one is not either */ \
                        all_are_dead = false; \
                    } \
                    else if(false) \
                    { \
                        /* Tarkista, voiko sen blokin tynt takaisin
                         * thn positioon ja pelaajan palauttaa siihen
                         * mist lhti. Jos voi, tmkin on blunder. */ \
                        /* Disabled because this check doesn't save anything */ \
                        size_t NewPushPos = targetpos + (dir); \
                        if(level.TestWalkingRoute(newplrpos, NewPushPos)) \
                        { \
                            level.MoveBlockFrom(targetpos, blockpos); \
                            if(level.TestWalkingRoute(targetpos, plrpos)) \
                            { \
                                std::cerr << "\nHIT\n"; \
                                verdict[route] = dead; \
                            } \
                            goto skipundo; \
                        } \
                    } \
                    level.MoveBlockFrom(targetpos, blockpos); /* Undo the move */ \
                skipundo: ; \
                } \
            } \
          }
        TestPush(-1);
        TestPush(-MaxWidth);
        #undef TestPush
    }
#endif
    
    SimpleVec<CoordinateType, Depth*4> found_dead;
    
    for(typename std::set<CoordinateType>::const_iterator
        i = playerpositions.begin(); i != playerpositions.end(); ++i)
    {
        const CoordinateType plrpos = *i;
        const size_t route = routes[plrpos];
        switch(verdict[route])
        {
            case ignore:  continue;
            case notdead: continue;
            case dead:    found_dead.push_back(plrpos); continue;
            case unknown: break;
        }
        level.SetNBlockSituation<Depth>(plrpos, blockposlist);
        if(SmallCompleter(level).IsCompleteable())
        {
            verdict[route] = notdead;
            all_are_dead = false;
        }
        else
        {
            verdict[route] = dead;
            found_dead.push_back(plrpos);
        }
    }
    
    size_t nfound_dead = found_dead.size();
    if(all_are_dead)
    {
        if(nfound_dead > 0)
        {
            /* All positions were dead. */
            RegisterAlwaysMultiDead<Depth> (blockposlist);
            ++nalldead;
        }
    }
    else
    {
        /* Not all positions were dead. Mark those which were. */

        for(size_t a=0; a<nfound_dead; ++a)
        {
            const size_t plrpos = found_dead[a];
            RegisterMultiDead<Depth>(plrpos, blockposlist);
            ++ndead;
        }
        
        for(size_t route=0; route< (size_t)MaxDim; ++route)
        {
            if(verdict[route] != dead) continue;
            for(size_t plrpos=0; plrpos< (size_t)MaxDim; ++plrpos)
                if((size_t)routes[plrpos] == route)
                {
                    RegisterMultiIgnore<Depth>(plrpos, blockposlist);
                    ++nignore;
                    //SmallCompleter(level).MarkEverythingIgnorable<Depth> ();
                }
        }
    }
}

double DeadZones::MultiDeadCalculatePosition
    (const BigInt& total,
     const unsigned char digits[],
     size_t Depth) const
{
    BigInt position;
    position=0;
    size_t Max=UndeadMax;//-Depth;
    for(size_t Last=0, nbox=0; nbox<Depth; ++nbox)
    {
        size_t Coord = digits[nbox]+nbox;
        position += Binom::Choose(Max - Last,  Depth-nbox);
        position -= Binom::Choose(Max - Coord, Depth-nbox);
        Last = Coord+1;
    }
    return position.GetDouble() / total.GetDouble();
}

static const double GetTime()
{
    struct timeval tv;
    gettimeofday(&tv,NULL);
    return tv.tv_sec + tv.tv_usec*1e-6;
}

static const std::string FormatTime(double d)
{
    char Buf[64];
    std::sprintf(Buf, "%u:%02u:%02.0f",
        (unsigned)(d/3600),
        ((unsigned)(d/60))%60,
        std::fmod(d,60.0)
                );
    return Buf;
}

static const char dotchars[2] = {'\'', '.'};

template<size_t Depth>
void DeadZones::DeadMultiBlockBuilderLoop(unsigned char digits[Depth])
{
    if(Depth >= MaxObject) return;
    
    std::cerr << "Collecting Dead" << Depth << " data...\n";
    
    BigInt total;
    total.Choose(UndeadMax, Depth);
    
    const size_t maxdigit = UndeadMax-Depth+1;
    size_t statuscounter=1+1; /* +1 so that timers won't display NAN */
    size_t dotcounter=1;
    
    size_t dumped_lines = MaxHeight;
    size_t dumped_width = MaxWidth+4;
    
    size_t interval = 5000 / Depth;
    
    for(size_t d=0; d<Depth; ++d)
        if(digits[d] >= maxdigit) /* Matches also 0xFF set by LoadDeadData() */
        {
            return;
        }
    
    const double begin_pos  = MultiDeadCalculatePosition(total, digits, Depth);
    const double begin_time = GetTime();
    
    size_t dotcharcounter=0;
    
    size_t ndependdead = std::count_if(MultiDeadFlags.WithPlayer.begin(),
                                       MultiDeadFlags.WithPlayer.end(),
                                       DeadCount<Depth>);

    size_t nalwaysdead = std::count_if(MultiDeadFlags.Set.begin(),
                                       MultiDeadFlags.Set.end(),
                                       DeadCountBlockSet<Depth>);

    size_t nignore = std::count_if(MultiIgnoreFlags.WithPlayer.begin(),
                                   MultiIgnoreFlags.WithPlayer.end(),
                                   DeadCount<Depth>);
    
    for(;;)
    {
        if(TerminateSignalled)
        {
            break;
        }
        
        ssize_t blockposlist[Depth];
    
        for(size_t d=0; d<Depth; ++d)
            blockposlist[d] = UndeadList[digits[d]+d];
        
        if(!--dotcounter)
        {
            size_t width = 65-dumped_width;
            dotcounter=interval/width;
            std::cerr << dotchars[dotcharcounter] << std::flush;
        }
        
        if(!--statuscounter)
        {
            statuscounter = interval;
            
            dotcharcounter = (dotcharcounter+1) % sizeof(dotchars);
            
            const double position = MultiDeadCalculatePosition(total, digits, Depth);
            const double timepos  = GetTime();
            
            const double TimeUsed = (timepos-begin_time);
            const double Speed = TimeUsed / (position-begin_pos); // seconds per completion
            const double EstimatedSofar = Speed * position;
            const double EstimatedTotal = 1.0 * Speed;
            const double EstimatedLeft = (1.0 - position) * Speed;
            
            std::cerr << "]  \r" << std::flush;
            
            LevelState level;
            level.SetNBlockSituation<Depth>(0, blockposlist);
            level.Dump();
            
            #define YPosition(y) std::cerr << "[" << y << "A" << std::flush;
            #define XPosition() std::cerr << "\r[" << dumped_width << "C" << std::flush;
            
            YPosition(dumped_lines);
            XPosition();

            std::cerr << "\n";
            
            XPosition();
            std::cerr << "Currently testing combination: [";
            for(size_t b=0; b<Depth; ++b)
            {
                if(b)std::cerr << ':';
                std::cerr << (size_t)(digits[b]+b);
                //std::cerr << '-'<<blockposlist[b];
            }
            std::cerr << '/' << maxdigit << "]    \n";
            XPosition();
            
            gmp_fprintf(stderr,
                "Dead%u progress: %.2f%% of %Zu configurations\n",
                Depth, position*100.0, total.GetMpz());
            std::fflush(stderr);
            
            XPosition();
            std::cerr << "Stats: always=" << nalwaysdead  << ", cond=" << ndependdead
                                          << ", ignore=" << nignore << "\n";
            
            XPosition();
            std::cerr << "Time used: " << FormatTime(TimeUsed) << std::endl;
            
            XPosition();
            std::cerr << "Est."
                         " so far " << FormatTime(EstimatedSofar)
                      << ", left " << FormatTime(EstimatedLeft)
                      << ", total " << FormatTime(EstimatedTotal)
                      << "    \n";
            
            std::cerr << "\n\n\n\n";
            SaveDeadData<Depth>(digits);
            YPosition(11);
            XPosition();
            
            std::cerr << "[";
            
            #undef YPosition
            #undef XPosition
        }
        
        DeadMultiBlockBuilder<Depth> (blockposlist, 
                                      ndependdead,
                                      nalwaysdead,
                                      nignore);
        
        size_t d=Depth;
        while(d-- > 0) if(++digits[d] < maxdigit) break;
        
        if(d == (size_t)-1LL) break;
        
        while(++d < Depth) digits[d] = digits[d-1];
    }
    std::cerr << "\33[" << dumped_lines << "B\r\n" << std::flush;
}

template<size_t Depth, typename container>
static void MultiDeadCombinationHelperBase(BigInt& amount, const container& m)
{
    BigInt tmp;
    tmp.Choose(CompactBlockCoordinateLimit, ActualMaxObject-1-Depth);
    amount -= tmp * (size_t)std::count_if(m.begin(), m.end(), DeadCount<Depth>);
}
template<size_t Depth>
struct MultiDeadCombinationHelper
{
    template<typename container>
    static void Subtract(BigInt& amount, const container& m)
    {
        MultiDeadCombinationHelperBase<Depth>(amount, m);
        MultiDeadCombinationHelper<Depth+1>::Subtract(amount, m);
    }
};
template<>
struct MultiDeadCombinationHelper<USE_MULTI_DEAD>
{
    template<typename container>
    static void Subtract(BigInt& amount, const container& m)
    {
        MultiDeadCombinationHelperBase<USE_MULTI_DEAD>(amount, m);
    }
};


void DeadZones::CountCombinations()
{
    std::cout << std::flush;
    std::cerr << std::flush;
    
    /* Calculate the number of different possible box configurations */
    BigInt tmp;
    tmp.Choose(CompactBlockCoordinateLimit, ActualMaxObject-1);
    
#if CHOOSE_LEVEL == 60
    gmp_fprintf(stderr,
        "// Number of possible different block combinations: %Zu\n",
        tmp.GetMpz());
#endif
    gmp_fprintf(stdout,
        "// Number of possible different block combinations: %Zu\n",
        tmp.GetMpz());
    
    /* Multiply by number of different player positions */
    //tmp *= (CompactPlayerCoordinateLimit - (ActualMaxObject-1));
    tmp *= 1+4*(ActualMaxObject-1);
    
#if CHOOSE_LEVEL == 60
    gmp_fprintf(stderr, "// When including the player, the number is %Zu.\n", tmp.GetMpz());
#endif
    gmp_fprintf(stdout, "// When including the player, the number is %Zu.\n", tmp.GetMpz());
    
    
#if USE_MULTI_DEAD >= 1
    DeadBlkSetVec& m = MultiDeadFlags.WithPlayer;
    MultiDeadCombinationHelper<1>::Subtract(tmp, m);
    /* The macro will also subtract recursively the higher values,
     * so no case for 2, 3, and so on, is required.
     */
#endif

#if CHOOSE_LEVEL == 60
    gmp_fprintf(stderr, "// Excluding the multidead, it is %Zu. // %u-%u\n", tmp.GetMpz(),
        levelno/5+1, levelno%5+1);
#endif
    gmp_fprintf(stdout, "// Excluding the multidead, it is %Zu. // %u-%u\n", tmp.GetMpz(),
        levelno/5+1, levelno%5+1);

    std::fflush(stderr);
    std::fflush(stdout);
    
    std::cerr << std::endl;
}

void DeadZones::FindMargins(void) const
{
    size_t nobjects=1; //player
    for(size_t pos=0; pos<19*16; ++pos)
        if(maps[levelno].mapdata[pos]&4)
            ++nobjects;

    size_t topmargin=0;
    for(size_t y=0; y<16; ++y,++topmargin)
        for(size_t x=0; x<19; ++x)
            if(!(maps[levelno].mapdata[y*19+x]&5))goto d1;
  d1:
    size_t bottommargin=0;
    for(size_t y=16; y-->0; ++bottommargin)
        for(size_t x=0; x<19; ++x)
            if(!(maps[levelno].mapdata[y*19+x]&5))goto d2;
  d2:
    size_t leftmargin=0;
    for(size_t x=0; x<19; ++x,++leftmargin)
        for(size_t y=0; y<16; ++y)
            if(!(maps[levelno].mapdata[y*19+x]&5))goto d3;
  d3:
    size_t rightmargin=0;
    for(size_t x=19; x-->0; ++rightmargin)
        for(size_t y=0; y<16; ++y)
            if(!(maps[levelno].mapdata[y*19+x]&5))goto d4;
  d4:
    /**/
    if(topmargin>0)--topmargin;
    if(bottommargin>0)--bottommargin;
    if(leftmargin>0)--leftmargin;
    if(rightmargin>0)--rightmargin;
    /**/
    size_t vert = 16-(bottommargin+topmargin);
    size_t horiz= 19-(leftmargin+rightmargin);
    
    ActualMaxObject = nobjects;
    
#if CHOOSE_LEVEL == 60
    std::cerr <<
        "#if CHOOSE_LEVEL==" << levelno << "\n"
        "// LEVEL " << (levelno/5+1) << "-" << (levelno%5+1) << " :\n"
        "enum  { MaxWidth=" << horiz << ", OffsetWidth=" << leftmargin << " };\n"
        "enum  { MaxHeight=" << vert << ", OffsetHeight=" << topmargin << " };\n"
        "enum  { MaxObject=" << nobjects << " };\n";
#else
    std::cout << "/* Suggested margins: "
              << horiz << "x" << vert
              << "+" << leftmargin << "+" << topmargin
              << " -- " << nobjects << " objects */"
              << std::endl;
    std::cout << "/* Current margins: "
              << MaxWidth << "x" << MaxHeight
              << "+" << OffsetWidth << "+" << OffsetHeight
              << " -- " << MaxObject << " objects */"
              << std::endl;
    

    if(nobjects != MaxObject
    || horiz > MaxWidth
    || vert > MaxHeight
    || topmargin < OffsetHeight
    || leftmargin < OffsetWidth)
    {
        std::abort();
    }
#endif

    BuildCompactCoordinates();
}

/* CheckPoint is a helper function for CreateDeadFlags1(). */
int DeadZones::CheckPoint(size_t point,
                          int walldir,
                          int sidedir)
{
    LevelState level;
    level.BuildFromLevel();
    level.RebuildBlockCaches();
    
    bool goal = level.IsGoal(point);
    
    if(level.IsWall(point)) return 2;
    
    /* If there is no wall directly in front, it's not dead */
    const bool wall_front = level.IsWall(point+walldir);
    const bool wall_back  = level.IsWall(point-walldir);
    if(!wall_front && !wall_back) return 0;
    
    /* There is a wall directly in the front! */

    /* If there is a wall directly on either side, it's dead (corner) */
    if(level.IsWall(point+sidedir) || level.IsWall(point-sidedir))
        return goal ? 1 : 2;
    
    if(goal) return 0; /* Can be moved, and it's at goal, so not dead */
    
    for(size_t p=point; ; p+=sidedir)
    {
        // goal seen? not dead
        if(level.IsGoal(p)) return 0;
        // wall seen? maybe dead
        if(level.IsWall(p)) break;
        // path seen? not necessarily dead
        if(wall_front && !(level.IsWall(p+walldir))) return 0;
        if(wall_back  && !(level.IsWall(p-walldir))) return 0;
    }
    for(size_t p=point-sidedir; ; p-=sidedir)
    {
        // goal seen? not dead
        if(level.IsGoal(p)) return 0;
        // wall seen? maybe dead
        if(level.IsWall(p)) break;
        // path seen? not necessarily dead            
        if(wall_front && !(level.IsWall(p+walldir))) return 0;
        if(wall_back  && !(level.IsWall(p-walldir))) return 0;
    }
    // dead!
    return 2;
}

void DeadZones::CreateDeadFlags1()
{
    BuildFloorCoordinates();

    /* First pass: Find corners and walls */
    for(size_t testpos=0; testpos< (size_t)MaxDim; ++testpos)
        DeadFlags[testpos]
         = std::max(
           CheckPoint(testpos,  MaxWidth, 1)  // wall up&down, check sides
        ,  CheckPoint(testpos,  1, MaxWidth)  // wall sides, check up&down
         );

/*{
    size_t ndead=0;
    for(size_t pos=0; pos< (size_t)MaxDim; ++pos)
        if(DeadFlags[pos]) ++ndead;
    
    std::cout << "// " << (size_t)MaxDim << " points, "
                       << (MaxDim-ndead) << " not dead\n";
}*/
    LevelState level;
    level.BuildFromLevel();
    const size_t playerpos = level.GetPlayerPos();
    
#if CHOOSE_LEVEL != 60
    std::cerr << "Collecting Dead1 data...\n";
#endif
    for(size_t testpos=0; testpos< (size_t)MaxDim; ++testpos)
    {
        if(DeadFlags[testpos] == 2 || level.IsGoal(testpos)) continue;
        
        level.RebuildBlockCaches();
        if(level.IsWall(testpos)) { DeadFlags[testpos] = 2;/* always dead */ continue; }
        
        const bool goal = level.IsGoal(testpos);
        
        LevelState tmp = level;
        
        //std::cerr << "testpos = " << testpos << std::endl;
        try
        {
            /* Find if there's a block that can be pushed into the given position */
            for(size_t n=1; n< (size_t)MaxObject; ++n)
            {
                size_t pos = level.GetBlockPos(n-1);
                if(pos == (size_t)MaxDim) continue;
                
                //std::printf("n=%u, now player is at %u\n", n, playerpos);
                ssize_t blockposlist[1] = {pos};
                tmp.SetNBlockSituation<1>(playerpos, blockposlist);
                //std::cerr << "testpos now = " << testpos << std::endl;
                
                ssize_t coordlist[1] = {testpos};
                CoordinateType result = SmallCompleter(tmp).IsPushableInto<1>(coordlist);
                if(result != NonCoordinate)
                {
                    //std::cerr << n << " works\n";
                    throw result;
                }
            }
            
            /* If a block can't be pushed to that location even if the field
             * is empty, mark it dead to save in the coordinate space.
             */
            //std::cerr << "Didn't work\n";
            DeadFlags[testpos] = 2; // always dead
        }
        catch(CoordinateType testplayerpos)
        {
            /* Now try if the block in question can be shoved into one of the
             * goals when the player is on the given position
             */
            //std::cerr << "Trying " << (size_t)testplayerpos << " at " << testpos << "\n";
            ssize_t blockposlist[1] = {testpos};
            tmp.SetNBlockSituation<1>(testplayerpos, blockposlist);
            //tmp.Dump();
            if(!SmallCompleter(tmp).IsCompleteable())
            {
                //std::cerr << "-> Dead\n";
                DeadFlags[testpos] = goal ? 1 : 2;
            }
        }
    }

#if 0
    std::string dead = "[";
    for(size_t y=0; y< (size_t)MaxHeight; ++y)
        for(size_t x=0; x< (size_t)MaxWidth; ++x)
        {
            if(x||y) dead += ',';
            dead += ('0' + DeadFlags[y*MaxWidth+x]);
        }
    dead += ']';
    std::cout << dead.c_str() << std::endl;
#endif
    
    if(levelno == 53)
    {
        /* WARP'S IDEA */
        static const CoordinateType deads[] =
        {
            163,182,201,220,239,149
        };
        static const size_t nextradead = sizeof(deads)/sizeof(*deads);
        for(size_t n=0; n<nextradead; ++n)
        {
            size_t c = deads[n];
            size_t x = c%19 + 0;
            size_t y = c/19 + 1;
            DeadFlags[(y-OffsetHeight)*MaxWidth+(x-OffsetWidth)] = 2;
        }
    }

    size_t ndead=0;
    for(size_t pos=0; pos< (size_t)MaxDim; ++pos)
        if(DeadFlags[pos] > 1) ++ndead;
    
    std::cout << "// " << (size_t)MaxDim << " points, "
                       << (MaxDim-ndead) << " not dead\n";

    UndeadMax=0;
    for(size_t pos=0; pos< (size_t)MaxDim; ++pos)
    {
        if(CheckDead(pos)) continue;
        UndeadList[UndeadMax++] = pos;
    }
    
    std::cout << "// UndeadMax=" << UndeadMax
              << ", CompactBlockCoordinateLimit=" << CompactBlockCoordinateLimit
              << "\n";

    std::cout << std::flush;
}

template<size_t Depth>
void DeadZones::CreateMultiDeadFlags ()
{
    unsigned char digits[Depth] = {0};
    LoadDeadData<Depth>(digits);
    DeadMultiBlockBuilderLoop<Depth> (digits);
    SaveDeadData<Depth>(digits);
    
    size_t ndependdead = std::count_if(MultiDeadFlags.WithPlayer.begin(),
                                         MultiDeadFlags.WithPlayer.end(),
                                         DeadCount<Depth>);

    size_t nalwaysdead = std::count_if(MultiDeadFlags.Set.begin(),
                                         MultiDeadFlags.Set.end(),
                                         DeadCountBlockSet<Depth>);

    size_t nignore = std::count_if(MultiIgnoreFlags.WithPlayer.begin(),
                                     MultiIgnoreFlags.WithPlayer.end(),
                                     DeadCount<Depth>);

    /* Calculate the number of different possible box configurations */
    BigInt total;
    total.Choose(CompactBlockCoordinateLimit, ActualMaxObject-1);
    
    BigInt nothercombinations;
    nothercombinations.Choose(CompactBlockCoordinateLimit-Depth, ActualMaxObject-1-Depth);

    const BigInt nalways_kill = nothercombinations * nalwaysdead;

    char nalways_buf[64];
    std::sprintf(nalways_buf, "%.1f",
       nalways_kill.GetDouble() * 100.0 / total.GetDouble()
                );

    BigInt ndepend_kill = nothercombinations * ndependdead;

    total *= (ActualMaxObject-1)*4; // approximate number of different player positions.
    
    char ndepend_buf[64];
    std::sprintf(ndepend_buf, "%.1f",
       ndepend_kill.GetDouble() * 100.0 / total.GetDouble()
                );

    std::cout << "// Dead" << Depth << " stats:\n";
    std::cout << "//    " << nalwaysdead << " always dead (coverage = " << nalways_buf << "%)\n";
    std::cout << "//    " << ndependdead << " conditional dead (coverage >= " << ndepend_buf << "%)\n";
    std::cout << "//    " << nignore << " to be ignored in consequent passes.\n" << std::flush;
}

template<size_t Depth>
struct MultiDeadMakingHelper
{
    static void Do()
    {
        DeadZones.CreateMultiDeadFlags<Depth>();
        MultiDeadMakingHelper<Depth+1>::Do();
    }
};
template<>
struct MultiDeadMakingHelper<USE_MULTI_DEAD>
{
    static void Do()
    {
        DeadZones.CreateMultiDeadFlags<USE_MULTI_DEAD>();
    }
};

void DeadZones::CreateAllMultiDeadFlags()
{
#if USE_MULTI_DEAD >= 1
    MultiDeadMakingHelper<1>::Do();
#endif
}

static bool CheckMultiBlockDeadInBlockSet(const DeadZones::BlockSet& com)
{
    const std::bitset<FloorAmount>& cur = CurrentBlockMap;
    
  #if 0
    if((cur&com)!=com) return false;
  #else
    typedef unsigned long wordtype;
    enum { wbits = std::numeric_limits<wordtype>::digits };
    enum { max   = (FloorAmount+wbits-1)/wbits };
    /* Access to std::bitset's private data. */
    const wordtype* p1 = (const wordtype*)&cur; // current position
    const wordtype* p2 = (const wordtype*)&com; // the dead layout
    for(size_t a=0; a<max; ++a)
    {
        if( (p1[a] & p2[a]) != p2[a]) return false;
    }
  #endif
    ++MultiDeadCounter[com.count()];
    return true;
}

static bool CheckMultiBlockDeadIn(const DeadZones::BlockSetVec& container)
{
    return
        std::find_if(container.begin(), container.end(), CheckMultiBlockDeadInBlockSet)
            != container.end();
}

static bool CheckMultiBlockDeadIn(ssize_t plrpos,
                                  const DeadZones::DeadBlkSetVec& container)
{
    for(DeadZones::DeadBlkSetVec::const_iterator
        i = std::lower_bound(container.begin(), container.end(),
                             plrpos, DeadFind);
        i != container.end() && (ssize_t)i->plrpos == plrpos;
        ++i)
    {
        if(CheckMultiBlockDeadInBlockSet(i->blockmap)) return true;
    }
    return false;
}

bool DeadZones::CheckMultiBlockDead(size_t plrpos) const
{
    return
        CheckMultiBlockDeadIn(MultiDeadFlags.Set) ||
        CheckMultiBlockDeadIn(plrpos, MultiDeadFlags.WithPlayer);
}

bool DeadZones::CheckMultiBlockIgnore(size_t plrpos) const
{
    return
        CheckMultiBlockDeadIn(plrpos, MultiIgnoreFlags.WithPlayer);
}

bool DeadZones::CheckMultiBlockDead(size_t plrpos, size_t lastblockpos) const
{
    std::map<CoordinateType, DeadFlagData>::const_iterator
        i = MultiDeadFlagsByBlock.find(lastblockpos);
    
    if(i == MultiDeadFlagsByBlock.end()) return false;
    
    return
        CheckMultiBlockDeadIn(i->second.Set) || 
        CheckMultiBlockDeadIn(plrpos, i->second.WithPlayer);
}

bool DeadZones::CheckMultiBlockDeadBlock() const
{
    return CheckMultiBlockDeadIn(MultiDeadFlags.Set);
}

bool DeadZones::CheckMultiBlockDeadBlock(size_t lastblockpos) const
{
    std::map<CoordinateType, DeadFlagData>::const_iterator
        i = MultiDeadFlagsByBlock.find(lastblockpos);
    
    if(i == MultiDeadFlagsByBlock.end()) return false;
    
    return CheckMultiBlockDeadIn(i->second.Set);
}

bool DeadZones::CheckMultiBlockIgnore(size_t plrpos, size_t lastblockpos) const
{
    std::map<CoordinateType, DeadFlagData>::const_iterator
        i = MultiIgnoreFlagsByBlock.find(lastblockpos);
    
    if(i == MultiIgnoreFlagsByBlock.end()) return false;
    return CheckMultiBlockDeadIn(plrpos, i->second.WithPlayer);
}

void DeadZones::BuildFromLevel()
{
    CreateDeadFlags1();

    FindMargins();
    
    CreateAllMultiDeadFlags();
    
    CountCombinations();

#if CHOOSE_LEVEL != 60
    /* Delegate some work: Let the LevelState show the dead zone map! */
    LevelState tmp;
    tmp.BuildFromLevel();
    tmp.Dump();
#endif
}

class DeadZones DeadZones;
