#include "binom.hh"
#include "deadzones.hh"
#include "levelstate.hh"

size_t DeadCounter1=0, DeadCounter0=0;
size_t MultiDeadCounter[USE_MULTI_DEAD+1]={0};
bool ShowDead=false;
bool PlayingNormally=false;

ObstacleMapType DefaultObstacleMap;
ObstacleMapType CurrentObstacleMap;
std::bitset<FloorAmount> CurrentBlockMap;

char LevelState::WallMap[MaxDim];

BigInt& LevelIdentifier::GetCompactData() const
{
    static BigInt compact;

    compact = 0;

    const ssize_t MaxCoord = CompactBlockCoordinateLimit;
    const size_t MaxBox   = MaxObject-1;
    for(size_t Last=0, nbox=0; nbox<MaxBox; ++nbox)
    {
        ssize_t Coord = CompactBlockCoordinateTranslation[ID[nbox+1]];
        compact += Binom::Choose(MaxCoord-Last,  MaxBox-nbox);
        compact -= Binom::Choose(MaxCoord-Coord, MaxBox-nbox);
        Last=Coord+1;
    }
    
    /* Since a player can't be in the same position as blocks are,
     * eliminate the block positions from the formula.
     */
    const CoordinateType* b = ID+1, *e = ID+MaxObject;
    ssize_t minus = std::upper_bound(b, e, ID[0]) - b;
    
    compact *= CompactPlayerCoordinateLimit;
    compact += CompactPlayerCoordinateTranslation[ID[0]] - minus;
    
    return compact;
}

void LevelIdentifier::MoveBlock(ssize_t pos0, ssize_t pos1)
{
    //if(pos1 == pos0) return;
    
    CoordinateType* ID_begin = ID+1;
    CoordinateType* ID_end   = ID+MaxObject;
    CoordinateType* src  = std::lower_bound(ID_begin,ID_end,(CoordinateType)pos0);
    if(src >= ID_end || *src != pos0) return;
    
    CoordinateType* ID_up1 = ID_begin;
    CoordinateType* ID_up2 = ID_end;
    if(pos1 > pos0) ID_up1 = src; else ID_up2 = src+1;
    CoordinateType* dest = std::upper_bound(ID_up1,ID_up2, (CoordinateType)pos1);
    if(src < dest)
    {
        std::copy(src+1, dest, src);
        --dest;
    }
    else if(src > dest)
    {
        std::copy_backward(dest,src,src+1);
    }
    *dest = pos1;
}

void LevelState::BuildFromLevel()
{
    const mapdata& map = maps[levelno];
    
    ssize_t px = map.px - OffsetWidth;
    ssize_t py = map.py - OffsetHeight;
    
    Reset();
    LevelIdentifier::SetPlayer(py*MaxWidth+px);
    
    DefaultObstacleMap.reset();
    
    size_t nblocks = 0;
    for(ssize_t y=0; y<MaxHeight; ++y)
        for(ssize_t x=0; x<MaxWidth; ++x)
        {
            int c = map.mapdata[(y+OffsetHeight)*19+(x+OffsetWidth)] & 7;
            if(c & MapBits::Block)
            {
                LevelIdentifier::SetBlock(nblocks++, y*MaxWidth+x);
            }
            
            WallMap[y*MaxWidth + x] = (c & 3);
            if(c & MapBits::Wall)
            {
                DefaultObstacleMap.set(y*MaxWidth + x);
            }
        }
}


void LevelState::Dump(ssize_t focus) const
{
    bool FocusFound=false;
    bool FoundNonObstacleBlock=false;
    size_t nblocks=0;
    for(ssize_t y=0; y<MaxHeight; ++y)
    {
        for(ssize_t x=0; x<MaxWidth; ++x)
        {
            ssize_t pos = y*MaxWidth+x;
            
            //if(CurrentObstacleMap.test(pos)) std::cerr << "[44m";
            
            ssize_t floorpos = FloorCoordinateTranslation[pos];
            if(floorpos < FloorAmount
            && CurrentBlockMap.test(floorpos))
            {
                std::cerr << "[1;31m";
                if(!CurrentObstacleMap.test(pos)) FoundNonObstacleBlock=true;
                nblocks += 1;
            }
            if(pos == focus) std::cerr << "[5m";
            
            if(IsWall(pos))
                std::cerr << '#';
            else if(IsBlock(pos))
            {
                if(pos==focus) { FocusFound=true; }
                std::cerr << '@';
            }
            else if(pos == GetPlayerPos())
                std::cerr << 'P';
            else if(DeadZones.CheckDead(pos))
                std::cerr << "[34md";
            else if(DeadZones.CheckUnmovable(pos))
                std::cerr << "[34m*";
            else if(IsGoal(pos))
                std::cerr << "[32m*";
            else
                std::cerr << ' ';
            
            std::cerr << "[0m";
        }
        std::cerr << "\n";
    }
    
    
    if(focus != NonCoordinate && !FocusFound)
    {
        std::cerr << "ERROR: Focus " << focus << " not found!\n";
    }
    if(FoundNonObstacleBlock) std::cerr << "ERROR: Found non-obstacle block!\n";
    if(PlayingNormally && 
       nblocks != MaxObject-1) std::cerr << "ERROR: Block amount doesn't match!\n";
    
    std::cerr << "\33[2A\33[" << (MaxWidth+4) << "C" << std::flush;
    LevelIdentifier::Dump();
    
    std::cerr << "\r\33[" << (MaxWidth+4) << "C" << std::flush;

    std::fprintf(stderr, "Compacted:");
    BigInt& key_enc = GetCompactData();       
    BigInt rem_enc;
    
    enum { RealSize = MaxCompactCoordinateData };
    CompactCoordinateUnit KeySeq[RealSize];   
    for(size_t a=0; a<RealSize; ++a)
    {
        size_t remainder = key_enc.DivMod2Exp(key_enc, rem_enc, CompactCoordinateBitsPerUnit);
        KeySeq[RealSize-a-1] = remainder;
    }
    for(size_t a=0; a<RealSize; ++a) std::fprintf(stderr, " %02X", KeySeq[a]);
    std::fprintf(stderr, "\n");
    std::fflush(stderr);
    
}


bool LevelState::CheckBlunder(ssize_t blockpos, int offset) const /*REGPARM*/
{
    //if(!PlayingNormally)return false;
    
    ssize_t plrpos = blockpos-offset;

    /* blockpos  = new location of the block */
    if(DeadZones.CheckDead(blockpos))
    {
        /*
        if(ShowDead)
        {
            std::cerr << "--- Blunder 1 ? ---\n";
            ssize_t tmp = GetPlayerPos();
            (const_cast<LevelState*>(this))->WalkTo(plrpos);
            Dump(blockpos);
            (const_cast<LevelState*>(this))->WalkTo(tmp);
        }
        */

        ++MultiDeadCounter[1];
        return true;
    }

    if(DeadZones.CheckMultiBlockDead(plrpos, blockpos))
    {
        if(ShowDead)
        {
            std::cerr << "--- Multiblunder ---\n";
            ssize_t tmp = GetPlayerPos();
            (const_cast<LevelState*>(this))->WalkTo(plrpos);
            Dump(blockpos);
            (const_cast<LevelState*>(this))->WalkTo(tmp);
        }
        return true;
    }
    
#if USE_MULTI_DEAD >= 4
    /*if(PlayingNormally)*/ return false;
#endif

#if false || USE_MULTI_DEAD < 3
    //SimpleVec<int,MaxObject> Movable;
    int Movable[MaxObject];
    Movable[0] = 2;
    
    //SimpleVec<size_t,MaxDim> Identify;
    unsigned char Identify[MaxDim] = { 0 };

    CoordinateType BlockPos[MaxObject];
    for(size_t n1=1; n1<MaxObject; ++n1)
    {
        ssize_t n1pos = GetBlockPos(n1-1);
        BlockPos[n1] = n1pos;
        if(unlikely(n1pos == NonCoordinate)) break;

        //if(n1pos >= MaxDim) { std::cerr << n1pos << std::endl; }
        Identify[n1pos] = n1;
        
        if(PlayingNormally)
        {
            /*
            Optimize:
              - If the block is at goal, it's "maybe"
              - If the block is the given block, it's "maybe"
              - If the block is adjacent to given block, it's "maybe"
              - Otherwise it's "movable=yes", because assume we've not done blunder yet.
              
              - Diagonals are also included in the adjacency test.
                Otherwise the Magic Four is not detected as a blunder.
            */
            if((n1pos < blockpos         -1 || n1pos > blockpos         +1)
            && (n1pos < blockpos-MaxWidth-1 || n1pos > blockpos-MaxWidth+1)
            && (n1pos < blockpos+MaxWidth-1 || n1pos > blockpos+MaxWidth+1)
            && !IsGoal(n1pos))
            {
                Movable[n1] = 2;
            }
            else
                Movable[n1] = 1;
        }
        else
        {
            /* Not playing normally, thus we can't assume
             * that a blunder hasn't been made
             */
            Movable[n1] = 1;
        }
    }/*for n1*/

    
    #define Movability_1dir(frontpos) \
        (IsWall(frontpos) ? 0 : Movable[Identify[frontpos]]) \
    /* If there is no obstacle in that direction, it's movable (objno=0, movable=Yes) */
    
    /* No bounds checking! The field should have walls! */
    #define Movability_4dirs(blockpos) \
        ({ int m1 = Movability_1dir(blockpos+1); \
           int m2 = Movability_1dir(blockpos-1); \
           int m3 = Movability_1dir(blockpos+MaxWidth); \
           int m4 = Movability_1dir(blockpos-MaxWidth); \
           ((m1==0||m2==0) && (m3==0||m4==0)) ? 0 : \
           ((m1==2&&m2==2) || (m3==2&&m4==2)) ? 2 : \
           1; \
        })
    
    size_t rolls;
    for(rolls=1; ; ++rolls)
    {
        bool missing=false, helped=false;
        for(size_t n=1; n<MaxObject; ++n)
        {
            ssize_t npos = BlockPos[n];
            if(unlikely(npos == NonCoordinate)) break;
            
            int CurrentState = Movable[n];
            if(CurrentState != 1)
            {
                continue; // This block already determined
            }
            
            int NewState = Movability_4dirs(npos);
            if(NewState != CurrentState)
            {
                helped=true;
                Movable[n] = NewState;
            }
            if(NewState == 1)
                missing = true;
        }
        
        /* If nothing to solve, or nothing was solved, break */
        if(!missing || !helped)
        {
            break;
        }
    }
    
    for(size_t n=1; n<MaxObject; ++n)
    {
        ssize_t npos = BlockPos[n];
        if(unlikely(npos == NonCoordinate)) break;
    
        int State = Movable[n];
        /* If the block could not be proven to be moveable,
         * it must be stuck.
         */
        if(State < 2)
        {
            /* If the block is not movable, it must be in a goal or it's blunder */
            if(!IsGoal(npos))
            {
             /*   std::cerr << "Blunder, in " << rolls << " rounds:\n";
                Dump(npos);*/
                ++DeadCounter0;
                return true;
            }
            else
            {/*
                std::cerr << "Unmovable, but not blunder:\n";
                Dump(npos);*/
            }
        }
    }
    return false;
#endif
}

void LevelState::FindWalkingDestinations(FavorablePointListWithCost& targets) const
{
    ObstacleMapType UnwantedPoints = CurrentObstacleMap;

    /* Find all positions that could be favorable */
    /* Such positions are next to each block. */
    unsigned char Favorable[MaxDim];
    std::memset(Favorable,0,sizeof(Favorable));
    size_t NFavorable=0;
    
    for(size_t n=1; n<MaxObject; ++n)
    {
        ssize_t pos = GetBlockPos(n-1); // location of a block
        
        if(DeadZones.CheckUnmovable(pos)) continue; // ignore an unmovable block
        
        if(unlikely(pos == NonCoordinate)) break;
        /* No bounds checking! The field should have walls! */
        /* It doesn't matter if the favorable points are inside walls,
         * but we check it anyway in hopes that the !any() condition
         * will match more often.
         */
        #define Try(p, tgt, mask) \
            { if(!UnwantedPoints.test(p) \
              && !DeadZones.CheckDead(tgt) \
              && !(Favorable[p] & mask)) \
                { if(!Favorable[p]) ++NFavorable; Favorable[p] |= mask; } }
        Try(pos-1,        pos+1,        8);
        Try(pos+1,        pos-1,        4);
        Try(pos-MaxWidth, pos+MaxWidth, 2);
        Try(pos+MaxWidth, pos-MaxWidth, 1);
        #undef Try
    }

    SimpleVec<CoordinateType, MaxDim> CurPoints, NextPoints;
    const ssize_t origin = GetPlayerPos();
    CurPoints.push_back(origin);
    if(!Favorable[origin])
    {
        Favorable[origin] = 1+2+4+8; // Consider the origin favorable too
        ++NFavorable;
    }
    UnwantedPoints.set(origin); // But the origin should never be returned into

    for(size_t Cost=0; !CurPoints.empty(); ++Cost)
    {
        NextPoints.clear();
        for(size_t a=CurPoints.size(); a-->0; )
        {
            ssize_t curpoint = CurPoints[a];
            if(Favorable[curpoint])
            {
                unsigned dir = Favorable[curpoint];
                Favorable[curpoint] = 0; --NFavorable;
                targets.push_back(FavorablePointWithCost(curpoint, Cost, dir));
                if(unlikely(!NFavorable)) return;
            }

            #define Try(p) { ssize_t newpoint = p; \
                if(!UnwantedPoints.test(newpoint)) \
                { \
                    UnwantedPoints.set(newpoint); \
                    NextPoints.push_back(newpoint); \
                } }

            /* No bounds checking! The field should have walls! */
            Try(curpoint-MaxWidth);
            Try(curpoint-1);
            Try(curpoint+1);
            Try(curpoint+MaxWidth);
            
            #undef Try
        }
        CurPoints.quickswap(NextPoints);
    }
}

void LevelState::FindWalkingDestinations(FavorablePointList& targets) const
{
    ObstacleMapType UnwantedPoints = CurrentObstacleMap;

    /* Find all positions that could be favorable */
    /* Such positions are next to each block. */
    unsigned char Favorable[MaxDim];
    std::memset(Favorable,0,sizeof(Favorable));
    size_t NFavorable=0;
    
    for(size_t n=1; n<MaxObject; ++n)
    {
        ssize_t pos = GetBlockPos(n-1); // location of a block

        if(DeadZones.CheckUnmovable(pos)) continue; // ignore an unmovable block
        
        if(unlikely(pos == NonCoordinate)) break;
        /* No bounds checking! The field should have walls! */
        /* It doesn't matter if the favorable points are inside walls,
         * but we check it anyway in hopes that the !any() condition
         * will match more often.
         */
        #define Try(p, tgt, mask) \
            { if(!UnwantedPoints.test(p) \
              && !DeadZones.CheckDead(tgt) \
              && !(Favorable[p] & mask)) \
                { if(!Favorable[p]) ++NFavorable; Favorable[p] |= mask; } }
        Try(pos-1,        pos+1,        8);
        Try(pos+1,        pos-1,        4);
        Try(pos-MaxWidth, pos+MaxWidth, 2);
        Try(pos+MaxWidth, pos-MaxWidth, 1);
        #undef Try
    }

    SimpleVec<CoordinateType, MaxDim> CurPoints, NextPoints;
    const ssize_t origin = GetPlayerPos();
    CurPoints.push_back(origin);
    if(!Favorable[origin])
    {
        Favorable[origin] = 1+2+4+8; // Consider the origin favorable too
        ++NFavorable;
    }
    UnwantedPoints.set(origin); // But the origin should never be returned into

    while(!CurPoints.empty())
    {
        NextPoints.clear();
        for(size_t a=CurPoints.size(); a-->0; )
        {
            ssize_t curpoint = CurPoints[a];
            if(Favorable[curpoint])
            {
                unsigned dir = Favorable[curpoint];
                Favorable[curpoint] = 0; --NFavorable;
                targets.push_back(FavorablePoint(curpoint, dir));
                if(unlikely(!NFavorable)) return;
            }

            #define Try(p) { ssize_t newpoint = p; \
                if(!UnwantedPoints.test(newpoint)) \
                { \
                    UnwantedPoints.set(newpoint); \
                    NextPoints.push_back(newpoint); \
                } }

            /* No bounds checking! The field should have walls! */
            Try(curpoint-MaxWidth);
            Try(curpoint-1);
            Try(curpoint+1);
            Try(curpoint+MaxWidth);
            
            #undef Try
        }
        CurPoints.quickswap(NextPoints);
    }
}

void LevelState::FindWalkingRouteTo(ssize_t nx,ssize_t ny, WalkingRoute& result) const
{
    WalkingRoute Routes[MaxDim];
    
    ObstacleMapType UnwantedPoints = CurrentObstacleMap;
    
    #define Reassign(pos, route, newpoint) \
        Routes[pos] = route; Routes[pos].push_back(newpoint);
    
    SimpleVec<CoordinateType, MaxDim> CurPoints, NextPoints;
    const ssize_t origin = GetPlayerPos();
    const ssize_t goal   = ny*MaxWidth+nx;
    CurPoints.push_back(origin);
    
    for(size_t Cost=0; !CurPoints.empty(); ++Cost)
    {
        NextPoints.clear();
        for(size_t a=CurPoints.size(); a-->0; )
        {
            ssize_t curpoint = CurPoints[a];
            const WalkingRoute& curroute = Routes[curpoint];
            
            if(curpoint == goal)
            {
                result = curroute;
                return;
            }

            #define Try(dir) { ssize_t newpoint=curpoint+dir; \
                if(!UnwantedPoints.test(newpoint)) \
                { \
                    UnwantedPoints.set(newpoint); \
                    Reassign(newpoint, curroute, newpoint); \
                    NextPoints.push_back(newpoint); \
                } }

            /* No bounds checking! The field should have walls! */
            Try(-MaxWidth);
            Try(-1);
            Try(1);
            Try(MaxWidth);
            
            #undef Try
        }
        CurPoints.quickswap(NextPoints);
    }
    #undef Reassign
    
    result = Routes[goal];
}

bool LevelState::TestWalkingRoute(const ssize_t origin, const ssize_t goal) const
{
    ObstacleMapType UnwantedPoints = CurrentObstacleMap;
    
    SimpleVec<CoordinateType, MaxDim> CurPoints, NextPoints;
    CurPoints.push_back(origin);
    UnwantedPoints.set(origin);
    
    while(!CurPoints.empty())
    {
        NextPoints.clear();
        for(size_t a=CurPoints.size(); a-->0; )
        {
            ssize_t curpoint = CurPoints[a];
            if(curpoint == goal) return true;

            #define Try(dir) { ssize_t newpoint=curpoint+dir; \
                if(!UnwantedPoints.test(newpoint)) \
                { \
                    UnwantedPoints.set(newpoint); \
                    NextPoints.push_back(newpoint); \
                } }

            /* No bounds checking! The field should have walls! */
            Try(-MaxWidth);
            Try(-1);
            Try(1);
            Try(MaxWidth);
            
            #undef Try
        }
        CurPoints.quickswap(NextPoints);
    }
    return false;
}

SimpleVec<CoordinateType,MaxDim> LevelState::BuildWalkingMaps() const
{
    SimpleVec<CoordinateType,MaxDim> result;
    result.resize(MaxDim);
    std::fill(result.begin(), result.end(), 0);
    
    ObstacleMapType UnwantedPoints = CurrentObstacleMap;
    
    SimpleVec<CoordinateType, MaxDim> CurPoints, NextPoints;
    for(ssize_t plrpos=0; plrpos < MaxDim; ++plrpos)
    {
        if(UnwantedPoints.test(plrpos)) continue;
        
        CurPoints.clear();
        CurPoints.push_back(plrpos);
        UnwantedPoints.set(plrpos);
        
        while(!CurPoints.empty())
        {
            NextPoints.clear();
            for(size_t a=CurPoints.size(); a-->0; )
            {
                ssize_t curpoint = CurPoints[a];
                result[curpoint] = plrpos;

                #define Try(dir) { ssize_t newpoint=curpoint+dir; \
                    if(!UnwantedPoints.test(newpoint)) \
                    { \
                        UnwantedPoints.set(newpoint); \
                        NextPoints.push_back(newpoint); \
                    } }

                /* No bounds checking! The field should have walls! */
                Try(-MaxWidth);
                Try(-1);
                Try(1);
                Try(MaxWidth);
                
                #undef Try
            }
            CurPoints.quickswap(NextPoints);
        }
    }
    return result;
}
