#include "intarray.hh"
#include "simplevec.hh"
#include "coordinates.hh"
#include "settings.hh"
#include "mapdata.hh"
#include "bigint.hh"

#include <bitset>
#include <iostream>

extern size_t DeadCounter0;
extern size_t MultiDeadCounter[USE_MULTI_DEAD+1];
extern bool ShowDead;
extern bool PlayingNormally;

struct FavorablePointWithCost
{
    CoordinateType Point;
    unsigned short Cost;
    unsigned short Dir; // bitmask
    /* 1=up, 2=down, 4=left, 8=right */
    
    inline FavorablePointWithCost(size_t p,unsigned c) : Point(p), Cost(c) { }
    inline FavorablePointWithCost(size_t p,unsigned c,unsigned d) : Point(p), Cost(c), Dir(d) { }
    inline FavorablePointWithCost() { }
    
    inline bool operator< (const FavorablePointWithCost& b) const
    {
        return Cost < b.Cost;
    }
};
struct FavorablePoint
{
    CoordinateType Point;
    unsigned short Dir; // bitmask

    inline FavorablePoint(size_t p, unsigned d) : Point(p), Dir(d) { }
    inline FavorablePoint() { }
    
    inline bool operator< (const FavorablePoint& b) const
    {
        return Point != b.Point ? Point < b.Point : Dir < b.Dir;
    }
};

typedef SimpleVec<CoordinateType, MaxDim> WalkingRoute;
typedef SimpleVec<FavorablePoint,         1+4*(MaxObject-1)> FavorablePointList;

/* Favorable positions may include current position and each 4 sides of each block. */
typedef SimpleVec<FavorablePointWithCost, 1+4*(MaxObject-1)> FavorablePointListWithCost;

typedef CompressedIntArray<MaxDim+1, MaxObject> CompressedLevel;

#if 1
struct ObstacleMapType
{
public:
    inline void reset() { std::fill_n(data,MaxDim,false); }
    inline void set(ssize_t n)
    {
        //if(n >= MaxDim) std::abort();
        data[n]=true;
    }
    inline void reset(ssize_t n)
    {
        //if(n >= MaxDim) std::abort();
        data[n]=false;
    }
    inline bool test(ssize_t n) const
    {
        //if(n >= MaxDim) std::abort();
        return data[n];
    }
private:
    bool data[MaxDim];
};
#else
typedef std::bitset<MaxDim> ObstacleMapType;
#endif
extern ObstacleMapType DefaultObstacleMap;
extern ObstacleMapType CurrentObstacleMap;
extern std::bitset<FloorAmount> CurrentBlockMap;

class LevelIdentifier
{
private:
    CoordinateType ID[MaxObject];
public:
    const CompressedLevel& Compress() const
    {
        static CompressedLevel result;
        result.InitFrom(ID);
        return result;
    }
    void Uncompress(const CompressedLevel& data)
    {
        data.ExtractTo(ID);
    }
    BigInt& GetCompactData() const;
    inline const CoordinateType* GetData() const { return ID; }

    void Dump() const
    {
        std::cerr << "Identifier:";
        for(size_t a=0; a<MaxObject; ++a)
            std::cerr << ' ' << (int)ID[a];
        std::cerr << "          \n";
    }
    
    void Reset()
    {
        CoordinateType* const ID_begin = ID;
        CoordinateType* const ID_end   = ID+MaxObject;
        std::fill(ID_begin, ID_end, NonCoordinate);
    }
    
    void MoveBlock(ssize_t pos0, ssize_t pos1) REGPARM;
    inline void SetBlock(ssize_t n, ssize_t pos) { ID[n+1] = pos; }
    inline void SetPlayer(ssize_t pos) { ID[0] = pos; }
    inline ssize_t GetPlayerPos() const
    {
        return ID[0];
    }
    inline ssize_t GetBlockPos(ssize_t blockno) const
    {
        return ID[blockno+1];
    }
    
    inline bool IsBlock(CoordinateType pos) const REGPARM
    {
        const CoordinateType* const ID_begin = ID;
        const CoordinateType* const ID_end   = ID+MaxObject;
        return std::binary_search(ID_begin+1, ID_end, pos);
    }
    /*
    ssize_t IdentifyBlock(CoordinateType pos) const
    {
        const CoordinateType* const ID_begin = ID;
        const CoordinateType* const ID_end   = ID+MaxObject;
        const CoordinateType* ptr ( std::lower_bound(ID_begin+1, ID_end, pos) );
        if(ptr >= ID_end) return 0;
        if(*ptr != pos) return 0;
        return ptr - ID_begin;
    }
    */
} __attribute__((packed));

class LevelState: private LevelIdentifier
{
private:
    static char WallMap[MaxDim]; /* Array of wall and goal bits */
    /* WallMap is updated by BuildFromLevel() */
    
    static inline ssize_t at(ssize_t x,ssize_t y)
    {
        return WallMap[y*MaxWidth+x];
    }
    static inline ssize_t at(ssize_t pos)
    {
        return WallMap[pos];
    }
public:
    using LevelIdentifier::Compress;
    using LevelIdentifier::GetData;
    using LevelIdentifier::GetCompactData;
    
    using LevelIdentifier::GetBlockPos;
    using LevelIdentifier::GetPlayerPos;
    
    void BuildFromLevel();
    
    inline void BuildFrom(const CompressedLevel& data)
    {
        LevelIdentifier::Uncompress(data);
    }
    
    void Dump(ssize_t focus=NonCoordinate) const;

    inline bool IsGoal(ssize_t pos) const { return at(pos) & MapBits::Goal; }
    inline bool IsWall(ssize_t pos) const { return at(pos) & MapBits::Wall; }
#if 0
    using LevelIdentifier::IsBlock;
#else
    bool IsBlock(ssize_t pos) const
    {
        ssize_t coord = FloorCoordinateTranslation[pos];
        return coord < FloorAmount && CurrentBlockMap.test(coord);
    }
#endif
    
    void RebuildBlockCaches() const
    {
        CurrentBlockMap.reset();
        
        CurrentObstacleMap = DefaultObstacleMap;
        for(size_t n=1; n<MaxObject; ++n)
        {
            ssize_t pos = GetBlockPos(n-1);
            if(unlikely(pos == NonCoordinate)) break;
            CurrentBlockMap.set(FloorCoordinateTranslation[pos]);
            CurrentObstacleMap.set(pos);
        }
    }

    template<size_t BlockCount>
    void SetNBlockSituation(ssize_t playerpos, const ssize_t blockposlist[BlockCount])
    {
        LevelIdentifier::Reset();
        LevelIdentifier::SetPlayer(playerpos);
        for(size_t a=0; a<BlockCount; ++a)
            LevelIdentifier::SetBlock(a, blockposlist[a]);
        RebuildBlockCaches();
    }

    bool IsCompleted() const
    {
        /* For each block, see if they're in a goal. */
        for(size_t n=1; n<MaxObject; ++n)
        {
            CoordinateType pos = GetBlockPos(n-1);
            if(unlikely(pos == NonCoordinate)) break;
            if(likely(!IsGoal(pos))) return false;
        }
        return true;
    }
    
    static inline bool PushForbidden(ssize_t blockpos, int offset)
    {
#if CHOOSE_LEVEL==60
        #define LevelOnly(l) levelno==l
#else
        #define LevelOnly(l) CHOOSE_LEVEL==l
#endif
        /* Special rule in 4-1: Don't push right on the goal area. */
        if(LevelOnly(15) && offset == 1)
        {
            ssize_t x = blockpos%MaxWidth;
            ssize_t y = blockpos/MaxWidth;
            if(x <= 7 && y <= 5) return true;
        }
        /* Special rule in 3-2: Don't push right on the goal area. */
        if(LevelOnly(11) && offset == 1)
        {
            ssize_t x = blockpos%MaxWidth;
            ssize_t y = blockpos/MaxWidth;
            if(x >= 3 && x <= 5 && y <= 5) return true;
        }
        return false;
        #undef LevelOnly
    }
    
    inline bool CanPush(int offset) const { return CanPushFrom(GetPlayerPos(), offset); }
    bool CanPushFrom(ssize_t playerpos, int offset) const
    {
        ssize_t blockpos = playerpos + offset;
        ssize_t frontpos = blockpos + offset;
        return IsBlock(blockpos)
            && !PushForbidden(blockpos, offset)
            && !CurrentObstacleMap.test(frontpos);
    }
    ssize_t CalculateKick(int dx,int dy) const { return CalculateKick(dy*MaxWidth+dx); }
    ssize_t CalculateKick(int offset) const { return CalculateKickFrom(GetPlayerPos(), offset); }
    ssize_t CalculateKickFrom(ssize_t playerpos, int offset) const
    {
        ssize_t blockpos = playerpos + offset; // location of the block
        
        if(!IsBlock(blockpos))
        {
            //std::cerr << "No block at " << blockpos << " (from " << playerpos << ")\n";
            return 0;
        }
        if(PushForbidden(blockpos, offset)) return 0;
        size_t length = 0;
        /* No bounds checking! The field should have walls! */
        while(!CurrentObstacleMap.test(blockpos+offset))
        {
            blockpos += offset;
            ++length;
        }
        //std::cerr << "Got block at " << blockpos << ", length " << length << "\n";
        return length;
    }
    
    /* Walking route handling */
    void FindWalkingDestinations(FavorablePointListWithCost& targets) const;
    void FindWalkingDestinations(FavorablePointList& targets) const;
    void FindWalkingRouteTo(ssize_t nx,ssize_t ny, WalkingRoute& result) const;
    bool TestWalkingRoute(const ssize_t origin, const ssize_t goal) const;
    SimpleVec<CoordinateType,MaxDim> BuildWalkingMaps() const;
    
    void Walk(int dx,int dy)             { WalkTo(GetPlayerPos() + (dy*MaxWidth+dx)); }
    void WalkTo(CoordinateType pos)      { LevelIdentifier::SetPlayer(pos); }
    void WalkTo(ssize_t nx,ssize_t ny) { LevelIdentifier::SetPlayer(ny*MaxWidth+nx); }
    ssize_t Push(int dx,int dy)         { return Push(dy*MaxWidth+dx); }
    ssize_t Push(int offset)
    {
        const ssize_t pos = GetPlayerPos();
        PushFromNoMove(pos, offset);
        WalkTo(pos+offset);
        return pos+offset+offset;
    }
    inline void Kick(int dx,int dy, size_t distance)
    {
        Kick(dy*MaxWidth+dx, distance);
    }
    inline ssize_t Kick(int offset, size_t distance)
    {
        return KickFrom(GetPlayerPos(), offset, distance);
    }
    inline void PushFromNoMove(ssize_t pos, int offset)
    {
        ssize_t oldpos = pos    + offset;
        ssize_t newpos = oldpos + offset;
        MoveBlockFrom(oldpos, newpos);
    }
    void MoveBlockFrom(ssize_t oldpos, ssize_t newpos) REGPARM
    {
        CurrentObstacleMap.reset(oldpos);
        CurrentObstacleMap.set(newpos);
        
        /*
        if(oldpos >= MaxDim || newpos >= MaxDim
        || FloorCoordinateTranslation[oldpos] >= FloorAmount
        || FloorCoordinateTranslation[newpos] >= FloorAmount)
        {
            Dump(oldpos);
            abort();
        }
        */
        CurrentBlockMap.reset(FloorCoordinateTranslation[oldpos]);
        CurrentBlockMap.set(FloorCoordinateTranslation[newpos]);
        LevelIdentifier::MoveBlock(oldpos, newpos);
    }
    ssize_t KickFrom(ssize_t pos, int offset, size_t distance) REGPARM
    {
        ssize_t oldpos = pos+offset, newpos = oldpos+offset*distance;
        CurrentObstacleMap.reset(oldpos);
        CurrentObstacleMap.set(newpos);
        CurrentBlockMap.reset(FloorCoordinateTranslation[oldpos]);
        CurrentBlockMap.set(FloorCoordinateTranslation[newpos]);
        LevelIdentifier::MoveBlock(oldpos, newpos);
        return newpos;
    }
    
    /* CheckBlunder.
     * Block position and the direction it was moven are given.
     * Player is assumed to face next to the block in the direction
     * where block was moved from.
     */
    bool CheckBlunder(ssize_t blockpos, int offset) const REGPARM;
} __attribute__((packed));
