#include "settings.hh"
#include "templates.hh"

#include <iostream>
#include <cstdlib>

//#define TIMESTAMP_MOVES

struct Move
{
    enum types { WalkToKick, WalkToPush };
    enum dirs  { Up,Left,Right,Down     };
    
#ifdef TIMESTAMP_MOVES
    size_t timestamp : 16;
#endif
    CoordinateType  pos:  (Log2Ceil<MaxDim>::Bits+1); // +1 because of signedness
    unsigned   type: Log2Ceil<2>::Bits; // 1
    unsigned   dir:  Log2Ceil<4>::Bits; // 2
    unsigned offsetbytes: Log2Ceil<4>::Bits; // 2
    
    /* Unfortunately g++ doesn't allow zero-width bitfields. */
    
    enum { offset_extra_bits = 8-((Log2Ceil<MaxDim>::Bits+5)&7) };
    unsigned offset_extra: offset_extra_bits;
    /* ^ Number of bits required by the offset that comes after */

public:
    Move(types t, CoordinateType where, int facing) : pos(where),type(t)
    {
        SetFacing(facing);
    }
    
    inline int getdx() const
    {
        static const signed char MoveDirXtable[4] = {0,-1,1,0};
        return MoveDirXtable[dir];
    }
    inline int getdy() const
    {
        static const signed char MoveDirYtable[4] = {-1,0,0,1};
        return MoveDirYtable[dir];
    }
    inline ssize_t getpx() const { return pos%MaxWidth; }
    inline ssize_t getpy() const { return pos/MaxWidth; }
    
    inline int GetFacing() const
    {
        static const int MoveDirTable[4] = {-MaxWidth,-1,1,MaxWidth};
        return MoveDirTable[dir];
    }
    void SetFacing(int facing)
    {
        switch(facing)
        {
            case -1: dir=Left; break;
            case  1: dir=Right; break;
            case -MaxWidth: dir=Up; break;
            case  MaxWidth: dir=Down; break;
            default:
                throw "Spurious facing\n";
        }
    }
    
    void Dump(std::ostream& out = std::cout) const;
}   __attribute__((packed));

class MoveTree
{
private:
    size_t Last;
    static const size_t Root = (size_t)(~0ULL);
public:
    MoveTree() : Last(Root)
    {
    }
    ~MoveTree()
    {
        clear();
    }
    MoveTree(const MoveTree& b): Last(b.Last)
    {
    }
    MoveTree(const MoveTree& b, const Move& t) : Last(b.Last)
    {
        push_back(t);
    }

    bool empty() const { return rbegin() == rend(); }
    const Move& back() const { return *rbegin(); }

    void push_back(const Move& t);
    void clear();
    size_t size(size_t maxsize = ~0ULL) const
    {
        size_t Size = 0;
        for(reverse_const_iterator i = rbegin(); i != rend(); ++i)
        {
            if(++Size >= maxsize) break;
        }
        return Size;
    }
    
    const Move& operator[] (size_t n) const
    {
        n = size() - 1 - n;
        reverse_const_iterator i = rbegin();
        while(n > 0) { ++i; --n; }
        return *i;
    }
    
public:
    struct reverse_const_iterator
    {
        inline reverse_const_iterator(size_t p): ptr(p) { }
        inline bool operator== (const reverse_const_iterator& b) const { return ptr==b.ptr; }
        inline bool operator!= (const reverse_const_iterator& b) const { return ptr!=b.ptr; }

        reverse_const_iterator& operator++ ();
        const Move& operator* ();
    private:
        size_t ptr;
    };
    reverse_const_iterator rbegin() const { return reverse_const_iterator(Last); }
    reverse_const_iterator rend() const { return reverse_const_iterator(Root); }
} __attribute__((packed));
