#include "NewRadixSet.hh"

#include <algorithm>
#include <vector>

class SmallCompleter
{
public:
    SmallCompleter(const LevelState& lev)
    {
        pino.reserve(65536);
        pino.push_back(lev);
    }
private:
    struct WasFinished { };
    static void CheckFinished(const LevelState& state) throw(WasFinished)
    {
        if(state.IsCompleted())
        {
            throw WasFinished();
        }
    }
public:
    bool IsCompleteable()
    {
        try
        {
            TryMoves(CheckFinished);
        }
        catch(WasFinished)
        {
            pino.clear();
            return true;
        }
        return false;
    }
private:
    template<size_t BlockCount>
    struct CheckPushable
    {
        CheckPushable(ssize_t poslist[BlockCount]) throw()
        {
            for(size_t a=0; a<BlockCount; ++a) targetpos[a] = poslist[a];
            std::sort(targetpos, targetpos+BlockCount);
        }
        void operator() (const LevelState& state) throw(CoordinateType)
        {
            for(size_t a=0; a<BlockCount; ++a)
                if(likely(state.GetBlockPos(a) != targetpos[a]))
                    return;
            throw (CoordinateType)state.GetPlayerPos();
        }
        ssize_t targetpos[BlockCount];
    };
public:
    template<size_t BlockCount>
    CoordinateType IsPushableInto(ssize_t targetposlist[BlockCount])
    {
        try
        {
            TryMoves(CheckPushable<BlockCount>(targetposlist));
        }
        catch(CoordinateType coord)
        {
            pino.clear();
            return coord;
        }        
        return NonCoordinate;
    }
private:
    template<size_t Depth>
    static void IgnoreMarker(const LevelState& state)
    {
        size_t plrpos = state.GetPlayerPos();
        ssize_t blockposlist[Depth];
        for(size_t a=0; a<Depth; ++a)
            blockposlist[a] = state.GetBlockPos(a);

        DeadZones.RegisterMultiIgnore<Depth> (plrpos, blockposlist);
    }
public:
    template<size_t Depth>
    void MarkEverythingIgnorable()
    {
        TryMoves(IgnoreMarker<Depth>);
    }

private:
    template<typename FuncType>
    void TryMoves(FuncType check)
    {
        while(!pino.empty())
        {
            const LevelState state = pino.back();
            pino.pop_back();

            state.RebuildBlockCaches();

            check(state);
            TryState(state);
        }
    }
    
    void TryState(const LevelState& state)
    {
        typedef FavorablePointList targetlist;
        targetlist targets;
        state.FindWalkingDestinations(targets);
        
        for(targetlist::const_iterator i = targets.begin(); i != targets.end(); ++i)
        {
            const CoordinateType Point = i->Point;
            
            if(i->Dir & 4) TryDir(state, Point, -1);
            if(i->Dir & 8) TryDir(state, Point, 1);
            if(i->Dir & 1) TryDir(state, Point, -MaxWidth);
            if(i->Dir & 2) TryDir(state, Point, MaxWidth);
        }
    }
    
    void TryDir(const LevelState& state, const CoordinateType Point, int offset) REGPARM
    {
        if(state.CanPushFrom(Point, offset))
        {
            LevelState NewState = state;
            NewState.PushFromNoMove(Point, offset);
            if(!NewState.CheckBlunder(Point + offset*2, offset))
            {
                NewState.WalkTo(Point+offset);
                if(seenmap.add(NewState.GetData()))
                {
                    pino.push_back(NewState);
                }
            }
            state.RebuildBlockCaches();
        }
    }

    /* SmallCompleter MUST USE NON-COMPACTED coordinates, because
     * non-compacted coordinates are only built after DeadZones
     * are calculated, and SmallCompleter is used during DeadZones
     * calculation.
     *
     * However, floor coordinates must have been calculated.
     */    
private:
    std::vector<LevelState> pino;
    
    typedef NewRadixSet<CoordinateType,MaxDim,MaxObject> seenmaptype;
    seenmaptype seenmap;
};
